// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_ThreadImpl.C,v 1.4 1998/07/09 23:53:47 sameer Exp $

#include <HPCxx_Profile.h>
#include <HPCxx_ThreadImpl.h>
#include <HPCxx_Thread.h>
#include <HPCxx_Runnable.h>
#ifdef AIX
#include <sys/sched.h>		// nexus_thread_yield -> sched_yield
#else
#include <sched.h>
#endif
#include <stdio.h>
#include <pthread.h>

static pthread_key_t this_thread_key;

void init_thread_key(){
     pthread_key_create(&this_thread_key, NULL);
}


HPCxx_ThreadImpl::HPCxx_ThreadImpl(void) : runnable_((HPCxx_Runnable *)0)
{
  alive_ = 0;
  suspended_ = 0;
}
 
HPCxx_ThreadImpl::HPCxx_ThreadImpl(const HPCxx_ThreadImpl& thread)
  : thread_(thread.thread_), runnable_(thread.runnable_)
{
  TAU_PROFILE("HPCxx_ThreadImpl::HPCxx_ThreadImpl()", 
    "void (HPCxx_ThreadImpl& )", TAU_HPCXX);
  alive_ = 0;
  suspended_ = 0;
}
 
HPCxx_ThreadImpl::HPCxx_ThreadImpl(HPCxx_Runnable *runnable)
  : runnable_(runnable)
{
  TAU_PROFILE("HPCxx_ThreadImpl::HPCxx_ThreadImpl()", 
    "void (HPCxx_Runnable * )", TAU_HPCXX);
  alive_ = 0;
  suspended_ = 0;
}

HPCxx_ThreadImpl& HPCxx_ThreadImpl::operator=(const HPCxx_ThreadImpl& thread)
{
  TAU_PROFILE("HPCxx_ThreadImpl::operator=()", 
    "HPCxx_ThreadImpl& (HPCxx_ThreadImpl& )", TAU_HPCXX);
  if (this == &thread) return(*this); // self-assignment

  thread_ = thread.thread_;
  // WARNING, unless this clones runnable the potential exists for runnable_
  // to be deleted twice.
  runnable_ = thread.runnable_;

  return(*this);
}

HPCxx_ThreadImpl::~HPCxx_ThreadImpl()
{
  TAU_PROFILE("HPCxx_ThreadImpl::~HPCxx_ThreadImpl()", "void ()", TAU_HPCXX);
  delete runnable_;
}

void HPCxx_ThreadImpl::run()
{
  TAU_PROFILE("HPCxx_ThreadImpl::run()", "void ()", TAU_HPCXX);
  if (runnable_) runnable_->run();
}

void *HPCxx_ThreadImpl::start_(void *rhs)
{
  TAU_REGISTER_THREAD();
  TAU_PROFILE_TIMER(threadtimer, "start_()", "void *(void *)", TAU_HPCXX);
  TAU_PROFILE_START(threadtimer);

  HPCxx_Thread *thread = (HPCxx_Thread *)rhs;
  thread->impl_.alive_ = 1;
  pthread_setspecific(this_thread_key, thread);
  thread->run();
  TAU_PROFILE_STOP(threadtimer); // so profile data can be written to disk

  thread->impl_.joinmutex.lock();
     thread->impl_.alive_ = 2;
     thread->impl_.joincond.broadcast();
  thread->impl_.joinmutex.unlock();
  return((void *)0);
}

void HPCxx_ThreadImpl::start(HPCxx_Thread& thread)
{
  TAU_PROFILE("HPCxx_ThreadImpl::start()", "void (HPCxx_Thread& )", TAU_HPCXX);
  nexus_thread_create(&thread_, (nexus_thread_attr_t *)NULL,
					  (nexus_thread_func_t)start_,
					  (void *)&thread);
}

void HPCxx_ThreadImpl::stop(void *status)
{
  TAU_PROFILE("HPCxx_ThreadImpl::stop()", "void (void *)", TAU_HPCXX);
  nexus_thread_exit(status);
}

void HPCxx_ThreadImpl::yield()
{
  TAU_PROFILE("HPCxx_ThreadImpl::yield()", "void ()", TAU_HPCXX);
  nexus_thread_yield();
}

void HPCxx_ThreadImpl::sleep(long milliseconds, long nanoseconds)
{
  TAU_PROFILE("HPCxx_ThreadImpl::sleep()", "void (long, long)", TAU_HPCXX);
  nexus_usleep(milliseconds*1000 + (long)(nanoseconds/1000 + .5));
}

void HPCxx_ThreadImpl::suspend()
{
  TAU_PROFILE("HPCxx_ThreadImpl::suspend()", "void ()", TAU_HPCXX);
  suspendmutex.lock();
  suspended_ = 1;
  suspendcond.wait(suspendmutex);
  suspended_ = 0;
  suspendmutex.unlock();
}
  
void HPCxx_ThreadImpl::resume()
{
  TAU_PROFILE("HPCxx_ThreadImpl::resume()", "void ()", TAU_HPCXX);
  suspendmutex.lock();
   if(suspended_) suspendcond.signal();
  suspendmutex.unlock();
}
  
bool HPCxx_ThreadImpl::isAlive()
{
  TAU_PROFILE("HPCxx_ThreadImpl::isAlive()", "bool ()", TAU_HPCXX);
  if(alive_ == 1) return(true);
  else return(false);
}
  
HPCxx_Thread *HPCxx_ThreadImpl::currentThread()
{
  TAU_PROFILE("HPCxx_ThreadImpl::currentThread()", "HPCxx_Thread * ()", 
    TAU_HPCXX);
  HPCxx_Thread * thp = 
    (HPCxx_Thread *) pthread_getspecific(this_thread_key);
  return(thp);
}

// Returns the numeric ID of the thread
int HPCxx_ThreadImpl::getThreadID()
{
  TAU_PROFILE("HPCxx_ThreadImpl::getThreadID()", "int ()", TAU_HPCXX);
  //return(nexus_thread_self());  FIXME
  return(nexus_thread_self());  
  //return 1;
}

void HPCxx_ThreadImpl::join(int milliseconds, int nanoseconds)
{
  TAU_PROFILE("HPCxx_ThreadImpl::join()", "void (int, int)", TAU_HPCXX);
  joinmutex.lock();
    if(alive_ < 2) joincond.wait(joinmutex);
  joinmutex.unlock();
}

int HPCxx_ThreadImpl::getPriority()
{
  TAU_PROFILE("HPCxx_ThreadImpl::getPriority()", "int ()", TAU_HPCXX);
  // FIXME - not implemented
  return(0);
}

int HPCxx_ThreadImpl::setPriority(int priority)
{
  TAU_PROFILE("HPCxx_ThreadImpl::setPriority()", "int (int )", TAU_HPCXX);
  // FIXME - not implemented
  return(0);
}
