//-----------------------------------------------------------------------------
// HPCxx_BufferImpl.C - Tulip layer implementation of the HPCxx_Buffer class.
//
// HPC++ Run Time System Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: HPCxx_BufferImpl.C,v 1.1 1998/06/23 00:04:02 ejohnson Exp $
//-----------------------------------------------------------------------------
#include <HPCxx_BufferImpl.h>
#include <hpcxx_rts.h>
#include <assert.h>


//
// inline code
//

//---- reallocIfNecessary -----------------------------------------------------
// Reallocs more memory if the current size won't be able to contain an element
// of the given size.
//-----------------------------------------------------------------------------
inline void HPCxx_BufferImpl::reallocIfNecessary(size_t size)
{
  if (size + offset_ > size_) {
    size_ *= 2;
    if (size + offset_ > size_) size_ = size + offset_;
    buffer_ = (char*) realloc(buffer_, size_);
    assert(buffer_ != NULL);
  }
}


//---- copyToBuffer -----------------------------------------------------------
// Copies data from the given buffer into this buffer.
//
// data -- the data to copy
// size -- the number of bytes to copy
//-----------------------------------------------------------------------------
inline int HPCxx_BufferImpl::copyToBuffer(const void* data, size_t size)
{
  reallocIfNecessary(size);
  memcpy(buffer_ + offset_, data, size);
  offset_ += size;

  return HPCXX_SUCCESS;
}


//---- copyFromBuffer ---------------------------------------------------------
// Copies from this buffer into the given buffer.
//
// data -- receives the data
// size -- the number of bytes to copy
//-----------------------------------------------------------------------------
inline void HPCxx_BufferImpl::copyFromBuffer(void* data, size_t size)
{
  memcpy(data, buffer_ + offset_, size);
  offset_ += size;
}


//---- HPCxx_BufferImpl -------------------------------------------------------
// Constructor.
//
// bufSize -- the size of the buffer to be allocated
// buffer -- a buffer to use rather than allocating one directly
//-----------------------------------------------------------------------------
HPCxx_BufferImpl::HPCxx_BufferImpl(size_t bufSize, void* buffer)
{
  offset_ = 0;
  size_ = bufSize;
  if (buffer != NULL) {
    buffer_ = (char*) buffer;
    doOwnBuffer_ = false;
  } else {
    buffer_ = (char*) hpcxx_malloc(size_);
    doOwnBuffer_ = true;
  }
  assert (buffer_ != NULL);
}


//---- ~HPCxx_BufferImpl ------------------------------------------------------
// Destructor.
//-----------------------------------------------------------------------------
HPCxx_BufferImpl::~HPCxx_BufferImpl()
{
  if (buffer_ && doOwnBuffer_) {
    hpcxx_free(buffer_);
    buffer_ = NULL;
  }
}


//---- buffer -----------------------------------------------------------------
// Returns the data buffer.
//-----------------------------------------------------------------------------
void* HPCxx_BufferImpl::buffer(void)
{
  return (void*) buffer_;
}


//---- copyBuffer -----------------------------------------------------------
// Copies data from the given buffer into this buffer.
//
// buf  -- the buffer to copy from
// size -- the number of bytes to copy
//-----------------------------------------------------------------------------
void HPCxx_BufferImpl::copyBuffer(HPCxx_Buffer& buf, size_t size)
{
  reallocIfNecessary(size);
  memcpy(buffer_ + offset_, buf.impl_.buffer_ + buf.impl_.offset_, size);
  buf.impl_.offset_ += size;
  offset_ += size;
}


//---- reuse ------------------------------------------------------------------
// Resets data buffer so that it can be reused.  Data currently stored in the
// buffer will be overwritten.
//-----------------------------------------------------------------------------
void HPCxx_BufferImpl::reuse(size_t offset)
{
  offset_ = offset;
}


//---- setAllocatedSize -------------------------------------------------------
// Sets the allocated size of the buffer to be at least the given amount.
//-----------------------------------------------------------------------------
void HPCxx_BufferImpl::setAllocatedSize(size_t size)
{
  reallocIfNecessary(size);
}


//---- size -------------------------------------------------------------------
// Returns the currently used size of the data buffer.
//-----------------------------------------------------------------------------
size_t HPCxx_BufferImpl::size(void)
{
  return offset_;
}


//
// packing routines
//

int HPCxx_BufferImpl::pack(const float* data, int count)
{
  return copyToBuffer(data, count*sizeof(float));
}

int HPCxx_BufferImpl::pack(const double* data, int count)
{
  return copyToBuffer(data, count*sizeof(double));
}

int HPCxx_BufferImpl::pack(const short* data, int count)
{
  return copyToBuffer(data, count*sizeof(short));
}

int HPCxx_BufferImpl::pack(const u_short* data, int count)
{
  return copyToBuffer(data, count*sizeof(u_short));
}

int HPCxx_BufferImpl::pack(const int* data, int count)
{
  return copyToBuffer(data, count*sizeof(int));
}

int HPCxx_BufferImpl::pack(const u_int* data, int count)
{
  return copyToBuffer(data, count*sizeof(u_int));
}

int HPCxx_BufferImpl::pack(const long* data, int count)
{
  return copyToBuffer(data, count*sizeof(long));
}

int HPCxx_BufferImpl::pack(const u_long* data, int count)
{
  return copyToBuffer(data, count*sizeof(u_long));
}

int HPCxx_BufferImpl::pack(const char* data, int count)
{
  return copyToBuffer(data, count*sizeof(char));
}

int HPCxx_BufferImpl::pack(const u_char* data, int count)
{
  return copyToBuffer(data, count*sizeof(u_char));
}

//
// unpack routines
//

void HPCxx_BufferImpl::unpack(float* data, int count)
{
  copyFromBuffer(data, count*sizeof(float));
}

void HPCxx_BufferImpl::unpack(double* data, int count)
{
  copyFromBuffer(data, count*sizeof(double));
}

void HPCxx_BufferImpl::unpack(short* data, int count)
{
  copyFromBuffer(data, count*sizeof(short));
}

void HPCxx_BufferImpl::unpack(u_short* data, int count)
{
  copyFromBuffer(data, count*sizeof(u_short));
}

void HPCxx_BufferImpl::unpack(int* data, int count)
{
  copyFromBuffer(data, count*sizeof(int));
}

void HPCxx_BufferImpl::unpack(u_int* data, int count)
{
  copyFromBuffer(data, count*sizeof(u_int));
}

void HPCxx_BufferImpl::unpack(long* data, int count)
{
  copyFromBuffer(data, count*sizeof(long));
}

void HPCxx_BufferImpl::unpack(u_long* data, int count)
{
  copyFromBuffer(data, count*sizeof(u_long));
}

void HPCxx_BufferImpl::unpack(char* data, int count)
{
  copyFromBuffer(data, count*sizeof(char));
}

void HPCxx_BufferImpl::unpack(u_char* data, int count)
{
  copyFromBuffer(data, count*sizeof(u_char));
}
