//-----------------------------------------------------------------------------
// HPCxx_Comm.C -- Category of objects used for communication.  For example,
// Remote Method Invocation (RMI) and global pointer get/put.
//-----------------------------------------------------------------------------

#include "HPCxx_Comm.h"
#include "hpcxx_rtsImpl.h"
#include <iostream.h>


static const int DEFAULT_BUF_SIZE = 128;


HPCxx_Comm::HPCxx_Comm(HPCxx_Mutex* mutex)
           : staticBuf_(DEFAULT_BUF_SIZE)
{
  tag_  = 0;
  unique_  = false;
  waiting_ = false;
  comm_    = MPI_COMM_WORLD;
  mutex_   = mutex;
  buffer_  = &staticBuf_;  
  destination_ = 0;
}


HPCxx_Comm::~HPCxx_Comm(void)
{
  // don't have anything to do yet
}


int HPCxx_Comm::messageID(void) const
{
  return ((tag_ & ID_MASK) >> ID_SHIFT);
}


int HPCxx_Comm::messageSender(void) const
{
  return sender_;
}


int HPCxx_Comm::messageTag(void) const
{
  return tag_;
}


int HPCxx_Comm::messageThread(void) const
{
  return (tag_ & THRD_MASK);
}


int HPCxx_Comm::messageType(void) const
{
  return (tag_ >> TYPE_SHIFT);
}


HPCxx_Buffer& HPCxx_Comm::messageBuffer(void) const
{
  return *buffer_;
}


bool HPCxx_Comm::isUnique(void)
{
  return unique_;
}


//---- setMessageForReturn ----------------------------------------------------
// Switch the message addresses for a return to sender.
//-----------------------------------------------------------------------------
void HPCxx_Comm::setMessageForReturn(void)
{
  int sender = sender_;
  sender_ = destination_;
  destination_ = sender;
}


void HPCxx_Comm::setMessageID(int msgID, bool unique)
{
  tag_ = ( (tag_ & ~ID_MASK) | (msgID << ID_SHIFT) );
  unique_ = unique;
}


void HPCxx_Comm::setMessageType(int msgType)
{
  tag_ = ( (tag_ & ~TYPE_MASK) | (msgType << TYPE_SHIFT) );
}


void HPCxx_Comm::setDestination(const HPCxx_ContextID& dest)
{
  destination_ = dest;
}


void HPCxx_Comm::setSender(const HPCxx_ContextID& sender)
{
  sender_ = sender;
}


void HPCxx_Comm::setTag(int tag)
{
  tag_ = tag;
}


void HPCxx_Comm::setUnique(bool unique)
{
  unique_ = unique;
}


// MPI specific

void HPCxx_Comm::send(void)
{
  mutex_ -> lock();
  MPI_Isend(buffer_->impl_.buffer(), buffer_->impl_.size(), MPI_BYTE,
	    destination_, tag_, comm_, &request_);
  mutex_ -> unlock();
}


void HPCxx_Comm::send(HPCxx_Buffer& buffer)
{
  mutex_ -> lock();
  MPI_Isend(buffer.impl_.buffer(), buffer.impl_.size(), MPI_BYTE,
	    destination_, tag_, comm_, &request_);
  mutex_ -> unlock();
}


void HPCxx_Comm::send(void* buf, size_t size)
{
  mutex_ -> lock();
  MPI_Isend(buf, size, MPI_BYTE, destination_, tag_, comm_, &request_);
  mutex_ -> unlock();
}


void HPCxx_Comm::recv(void* buf, size_t size)
{
  mutex_ -> lock();
  MPI_Irecv(buf, size, MPI_BYTE, destination_, tag_, comm_, &request_);
  mutex_ -> unlock();
}


void HPCxx_Comm::receiveWaitingMessage(MPI_Status& status)
{
  int size;

  tag_ = status.MPI_TAG;
  sender_ = status.MPI_SOURCE;
  destination_ = *hpcxx_currentContext->getContextID();
  MPI_Get_count(&status, MPI_BYTE, &size);
  buffer_->impl_.setAllocatedSize(size);    // make sure buffer is big enough

  mutex_ -> lock();
  MPI_Recv(buffer_->impl_.buffer(),
	   size, MPI_BYTE, sender_, tag_, MPI_COMM_WORLD, &status);
  mutex_ -> unlock();
}


//---- wait -------------------------------------------------------------------
// Waits until the communication operation has completed (or at least
// until buffers can be reused).  The function blocks until the signal
// method on this object is called.
//-----------------------------------------------------------------------------
void HPCxx_Comm::wait(void)
{
  condMutex_.lock();
  waiting_ = true;
  while (waiting_) cond_.wait(condMutex_);
  condMutex_.unlock();
}


//---- signal -----------------------------------------------------------------
// Signals that the communication operation has completed and allows threads
// waiting on completion of the communication to continue execution.
//-----------------------------------------------------------------------------
void HPCxx_Comm::signal(void)
{
  condMutex_.lock();
  if (waiting_) cond_.broadcast();
  waiting_ = false;
  condMutex_.unlock();
}


ostream& operator<<(ostream& s, const HPCxx_Comm& c)
{
  return s << "tag = " << (void*) c.messageTag() << ", sender = " << c.sender_
           << ", dest = " << c.destination_
           << ", size = " << c.messageBuffer().impl_.size()
           << ", type = " << c.messageType() << ", id = " << c.messageID();
//           << ", thread = " << c.messageThread();
}

