//-----------------------------------------------------------------------------
// HPCxx_Comm.h -- Category of objects used for communication.  For example,
// Remote Method Invocation (RMI) and global pointer get/put.
//-----------------------------------------------------------------------------

#ifndef __HPCXX_COMM_H
#define __HPCXX_COMM_H

#include <hpcxx_types.h>
#include <HPCxx_Buffer.h>
#include <HPCxx_Sync.h>
#include <iostream.h>
#include <mpi.h>


//#----------------------------------------------------------------------------
//# Message Types
//#----------------------------------------------------------------------------

#define BarrierFanOutMsg   100 /* Message to notify that work can resume */
#define BarrierFanInMsg    101 /* Fan into one node to sync */
#define BroadcastMsg       110 /* All processors listen and receive... */
#define ReduceMsg          120 /* Reduce a value across all processors */

#define AskForDataMsg      130 /* Send this type to remote node to get data */
#define FetchBlockMsg      140 /* Partial reply to earlier AskForDataMsg */
#define EndOfFetchMsg      141 /* Completed reply to earlier AskForDataMsg */

#define PrefixStoreMsg     150 /* Send this type to prepare remote for store */
#define EagerPutMsg        151 /* Send the data as part of the put message */
#define StoreBlockMsg      160 /* Partial block of store */
#define EndOfStoreMsg      161 /* Final block of a store message */

#define RemoteFunctionMsg  171 /* Send to remote to execute function handler */
#define RemoteMethodMsg    172 /* Send to remote to execute method handler */

#define AckReplyMsg        180 /* Reply from remote that store has completed */

#define MAXMSGS          512 /* Maximum "active" message handles */
#define MAXBUFSIZE      16384 /* Maximum buffer size */


//-----------------------------------------------------------------------------
// A message tag is 4 byte quantity = (msgType, msgType, msgID, threadID)
// where the message type takes up the first 2 bytes.  Masks and shifts
// defined below.
//-----------------------------------------------------------------------------

#define TYPE_MASK	0xffff0000
#define   ID_MASK	0x0000ff00
#define THRD_MASK	0x000000ff

static const int TYPE_SHIFT = 16;
static const int   ID_SHIFT =  8;
static const int THRD_SHIFT =  0;


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
class HPCxx_Comm {

 public:

  HPCxx_Comm(HPCxx_Mutex* mutex);
  ~HPCxx_Comm(void);

  HPCxx_Buffer&	messageBuffer(void) const;
  int		messageID(void) const;
  int		messageSender(void) const;
  int		messageTag(void) const;
  int		messageThread(void) const;
  int		messageType(void) const;

  bool isUnique(void);

  void setDestination(const HPCxx_ContextID& dest);
  void setMessageForReturn(void);
  void setMessageID(int msgID, bool unique = false);
  void setMessageType(int msgType);
  void setSender(const HPCxx_ContextID& sender);
  void setTag(int tag);
  void setUnique(bool unique);
  void signal(void);
  void wait(void);

  // MPI specific

  void send(void);
  void send(HPCxx_Buffer& buffer);
  void send(void* buf, size_t size);
  void recv(void* buf, size_t size);
  void receiveWaitingMessage(MPI_Status& status);

 private:
  int		    tag_;		// tag for the message
  int		    destination_;	// destination of the message
  int		    sender_;		// sender of the message
  bool		    unique_;		// true if this msg id is unique
  HPCxx_Buffer	    staticBuf_;		// fixed size buffer (not always used)
  HPCxx_Buffer*	    buffer_;		// points to buffer for message params
  HPCxx_Cond	    cond_;		// allows waiting on message completion
  HPCxx_Mutex	    condMutex_;		// mutex used with cond_
  bool		    waiting_;		// flag is true while waiting

  // MPI specific

  HPCxx_Mutex*   mutex_;		// used to protect MPI calls
  MPI_Comm	 comm_;			// communication group
  MPI_Request	 request_;		// MPI comm object for competion

  friend class HPCxx_CommCoord;
  friend ostream& operator<<(ostream& s, const HPCxx_Comm&);
};


#endif // __HPCXX_COMM_H
