#include "HPCxx_CommCoord.h"
#include <hpcxx_rts.h>		// for hpcxx_debug_level
#include <HPCxx_Thread.h>
#include <hpcxx_rtsImpl.h>	// for hpcxx_threadID()
#include <iostream.h>

//
// This class is designed to have a separate queue for each kernel level
// thread within a context.  Since the user-level threads are non-preemptive,
// only one user-level thread at a time can access the CommCoord 
// class methods.  The bottom line: no locks are needed to protect the
// communication object queues.
//

static const int NUM_START_COMM_OBJECTS = 32; // initial number of comm objects

static const int MSG_ID_NOT_FOUND    = -1;  // error msg id return
static const int GENERIC_MSG_ID      = 0;   // a generic msg id
static const int UNIQUE_MSG_ID_START = 1;   // starting value of unique msg ids
static const int NUM_UNIQUE_MSG_IDS  = 255; // total number of valid msg ids


//
// Class objects and methods (static)
//

// Create a mutex to protect communication.  This mutex may be replaced.
static HPCxx_Mutex commMutex_g;

HPCxx_Mutex* HPCxx_CommCoord::commMutex_ = &commMutex_g;


void HPCxx_CommCoord::setCommMutex(HPCxx_Mutex* mutex)
{
  commMutex_ = mutex;
}


HPCxx_Mutex* HPCxx_CommCoord::commMutex(void)
{
  return commMutex_;
}


//---- HPCxx_CommCoord --------------------------------------------------------
// Constructor.
//-----------------------------------------------------------------------------
HPCxx_CommCoord::HPCxx_CommCoord(void)
{
  nCommunicators_  = 0;

  // create message id stack

  int msgID = UNIQUE_MSG_ID_START;
  for (int i = 0; i < NUM_UNIQUE_MSG_IDS; i++) {
    msgIDs_.push_back(msgID++);
  }

  // comm objects are created lazily, on demand (see increaseCommStackSize)
}


//---- ~HPCxx_CommCoord -------------------------------------------------------
// Destructor.  Frees all released communication objects.
//-----------------------------------------------------------------------------
HPCxx_CommCoord::~HPCxx_CommCoord(void)
{
  if (readyComms_.size() != nCommunicators_) {
    cerr << "HPCxx_CommCoord::~HPCxx_CommCoord, warning, communications"
	    "coordinator closing down and communication objects still out.\n";
  }
  //
  // free all released (stored in stack) comm objects
  //
  while (readyComms_.size() > 0) {
    HPCxx_Comm* comm = readyComms_.back();
    readyComms_.pop_back();
    delete comm;
  }
  //
  // free all posted comm objects here (is this the best thing to do?)
  //
  while (postedComms_.size() > 0) {
    HPCxx_Comm* comm = postedComms_.back();
    readyComms_.pop_back();
    delete comm;
  }
}


//---- checkStatus ------------------------------------------------------------
// Checks the status of completed communication to make sure that it is
// consistent with what the communication object expected.  This is a place
// to catch errors.
//
// comm   -- the communications object to check against
// status -- the status of the message (tag, sender, size)
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::checkStatus(HPCxx_Comm* comm, MPI_Status* status)
{
  int size;
  int tag = status->MPI_TAG;
  int sender = status->MPI_SOURCE;
  MPI_Get_count(status, MPI_BYTE, &size);

  // check status against communications object, this may not be necessary
  // so may want to do this optionally with a debug compile option

  if (tag != comm->tag_) {
    cerr << "HPCxx_CommCoord::checkStatus, tag[" << tag << "]"
         << " != comm_tag[" << comm->tag_ << "]" << endl;
  }
  if (sender != comm->destination_) {
    cerr << "HPCxx_CommCoord::checkStatus, sender[" << sender << "]"
         << " != comm_destination[" << comm->destination_ << "]" << endl;
  }
}


//---- increaseCommStackSize --------------------------------------------------
// Increases the communication object stack size by doubling.
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::increaseCommStackSize(void)
{
  // double the size of the working set if non-zero, otherwise use default

  int nToAdd = 2 * nCommunicators_;
  if (nToAdd == 0) {
    nToAdd = NUM_START_COMM_OBJECTS;
  }
  
  // create communication objects

  for (int i = 0; i < nToAdd; i++) {
    HPCxx_Comm* comm;
    comm = new HPCxx_Comm(HPCxx_CommCoord::commMutex());
    readyComms_.push_back(comm);
    nCommunicators_ += 1;
  }
}


//---- communicator -----------------------------------------------------------
// Returns a communications object.
//-----------------------------------------------------------------------------
HPCxx_Comm* HPCxx_CommCoord::communicator(const HPCxx_ContextID& dest,
					  int msgType, bool unique)
{
  int msgID;

  // Get unique message id, yield if can't, this must be reentrant at yield pt.
  // Note that threads are assumed to be nonpreemptive so that it can be
  // assumed that only one user level thread will be running this at a time.
  // Although one or more threads may be blocking at the yield pt.

  if (unique) {
    msgID = uniqueMsgID();
    while (msgID == MSG_ID_NOT_FOUND) {
      HPCxx_Thread::yield();
      msgID = uniqueMsgID();
    }
  } else {
    msgID = GENERIC_MSG_ID;
  }

  // create new communicators if necessary
  if (readyComms_.size() == 0) increaseCommStackSize();

  // find a communicator that has completed it's communications

  HPCxx_Comm* comm = readyComms_.back();
  readyComms_.pop_back();

  // set message parameters

  comm->setUnique(unique);
  int tag = HPCxx_CommCoord::tag(msgType, msgID);
  comm->setTag(tag);
  comm->setDestination(dest);

  return comm;
}


//---- poll -------------------------------------------------------------------
// Polls communication queues for events.
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::poll(void)
{
  HPCxx_Comm* comm;
  MPI_Status status;
  int completed, incoming;
  list<HPCxx_Comm*>::iterator i;

  int id = *(hpcxx_currentContext->getContextID());
  
  if (hpcxx_debug_level > 3) {
    cerr << id <<":Entering HPCxx_CommCoord::poll()" << endl;
  }

  //
  // First take care of all posted communications which have completed
  // so that the communication objects can be reused.  These are primarily
  // MPI_Isend which have may not have been received, but at least have 
  // progressed far enough that their internal buffers can be reused.
  //

  if (postedComms_.size() > 0) {
    i = postedComms_.begin();
    commMutex_->lock();
    while (i != postedComms_.end()) {
      comm = *i;
      
      MPI_Test(&comm->request_, &completed, &status);

      if (completed != 0) {
	if (hpcxx_debug_level > 2) {
	  cerr << id << ":HPCxx_CommCoord::poll, posted:\n  " << *comm << endl;
	}
	postedComms_.erase(i++);		// remove from posted list
	release(comm);				// add it back to ready list
      } else {
	i++;
      }
    }
    commMutex_->unlock();
  }

  //
  // Now check pre-posted communications which have received a reply message
  // from a remote call.  These are likely MPI_Irecv calls that were 
  // preposted so that a buffer would be waiting for the remote send
  // when it came in.
  // 

  if (prepostedComms_.size() > 0) {
    commMutex_->lock();
    i = prepostedComms_.begin();
    while (i != prepostedComms_.end()) {
      comm = *i;

      MPI_Test(&comm->request_, &completed, &status);

      if (completed) {
	if (hpcxx_debug_level > 2) {
	  cerr << id << ":HPCxx_CommCoord::poll, preposted:\n "
	       << *comm << endl;
	}
	postedComms_.erase(i++);		// remove from posted list

	int count;
	MPI_Get_count(&status, MPI_BYTE, &count);
	if (hpcxx_debug_level > 2) {
	  cerr << id << ":HPCxx_CommCoord::poll, preposted: count = "
	       << count << endl;
	}

	// a message has arrived, deal with it
	handleMessage(*comm);

	release(comm);				// add it back to ready list
      } else {
	i++;
      }
    }
    commMutex_->unlock();
  }

  //
  // Check for unrequested incoming messages.  These messages won't have
  // been previously posted so we don't have a communications object built for
  // it.  These messages could be coming from any source with any tag.
  //

  commMutex_->lock();
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &incoming, &status);
  commMutex_->unlock();

  while (incoming) {
    if (readyComms_.size() == 0) {
      if (hpcxx_debug_level > 2) {
        cerr << id << ":HPCxx_CommCoord::poll, increasing stack size" << endl;
      }
      increaseCommStackSize();
    }

    HPCxx_Comm* comm = readyComms_.back();
    readyComms_.pop_back();

    if (comm == NULL) {
      cerr << id << ":HPCxx_CommCoord::poll, comm == NULL" << endl;
      break;
    }

    comm->receiveWaitingMessage(status);
    if (hpcxx_debug_level > 2) {
      cerr << id << ":HPCxx_CommCoord::poll incoming:\n  " << *comm << endl;
    }

    // a message has arrived, deal with it
    handleMessage(*comm);

    commMutex_->lock();
    MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD,
	       &incoming, &status);
    commMutex_->unlock();
  }

  if (hpcxx_debug_level > 3) {
    cerr << id << ":Exiting HPCxx_CommCoord::poll()" << endl;
  }
}


//---- handleMessage ----------------------------------------------------------
// Handles an incoming message.
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::handleMessage(HPCxx_Comm& comm)
{
  int msgType = comm.messageType();

  switch (msgType) {
    case AckReplyMsg :
      hpcxx_returnHandler(comm);
      break;
    case AskForDataMsg:
      hpcxx_readHandler(comm);
      break;
    case BarrierFanInMsg:
      hpcxx_barrierFanIn(comm);
      break;
    case BarrierFanOutMsg:
      hpcxx_barrierFanOut(comm);
      break;
    case EagerPutMsg :
      hpcxx_writeHandler(comm);
      break;
    case EndOfFetchMsg:
      comm.signal();
      break;
    case RemoteFunctionMsg :
      hpcxx_invokeGFHandler(comm);
      break;
    case RemoteMethodMsg :
      hpcxx_invokeRMIHandler(comm);
      break;
    default:
      cerr << "HPCxx_CommCoord::handleMessage, unknown message type" << endl;
  }
}


//---- release ----------------------------------------------------------------
// Releases what had been an "in use" communications object.
//
// comm   -- the communications object to release
// status -- the status of the message (tag, sender, size)
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::release(HPCxx_Comm* comm)
{
  // if the message has a unique message id, release id for reuse

  if ( comm->isUnique() ) {
    msgIDs_.push_back( comm->messageID() );
  }

  // reset the message buffer so that it can be reused
  comm->messageBuffer().impl_.reuse();

  readyComms_.push_back(comm);

  // signal any objects that may be waiting on completion
  comm->signal();
}


//---- releaseToPosted --------------------------------------------------------
// Releases a communications object back to the coordinator to be placed
// in the posted queue.
// <ul>
// <li><b>comm</b> -- the communication object to release
// </ul>
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::releaseToPosted(HPCxx_Comm* comm)
{
  postedComms_.push_back(comm);

  // poll to see if posting has completed to clear waits on send
  hpcxx_poll();
}


//---- releaseToPreposted -----------------------------------------------------
// Releases a communications object back to the coordinator to be placed
// in the preposted queue (those objects that have preposted a receive).
// <ul>
// <li><b>comm</b> -- the communication object to release
// </ul>
//-----------------------------------------------------------------------------
void HPCxx_CommCoord::releaseToPreposted(HPCxx_Comm* comm)
{
  prepostedComms_.push_back(comm);
}


//---- tag --------------------------------------------------------------------
// Returns a tag based on message type, message id, and kernel-level thread #
//-----------------------------------------------------------------------------
int HPCxx_CommCoord::tag(int msgType, int msgID)
{
  return ( (((msgType << 8) | msgID) << 8)
	   | (HPCxx_Thread::getThreadID() & THRD_MASK));
}


//---- uniqueMsgID ------------------------------------------------------------
// Returns a unique message id if one exists, MSG_ID_NOT_FOUND otherwise.
//-----------------------------------------------------------------------------
int HPCxx_CommCoord::uniqueMsgID(void)
{
  if ( msgIDs_.empty() ) return MSG_ID_NOT_FOUND;

  int msgID = msgIDs_.back();
  msgIDs_.pop_back();

  return msgID;
}
