#ifndef __HPCXX_COMMCOORD_H_
#define __HPCXX_COMMCOORD_H_

#include "HPCxx_Comm.h"
#include "HPCxx_Mutex.h"
#include <vector.h>
#include <deque.h>
#include <list.h>

class HPCxx_CommCoord {

 public:

  //#---- HPCxx_CommCoord -----------------------------------------------------
  // Constructs a communications coordinator which builds communication objects
  // and coordinates and facilitates communications.
  //#--------------------------------------------------------------------------
  HPCxx_CommCoord(void);

  //#---- ~HPCxx_CommCoord ----------------------------------------------------
  // Destroys this communications coordinator.
  //#--------------------------------------------------------------------------
  ~HPCxx_CommCoord(void);

  //#---- checkStatus ---------------------------------------------------------
  // Checks the status of completed communication to make sure that it is
  // consistent with what the communication object expected.  This is a place
  // to catch errors.
  // <ul>
  // <li><b>comm</b> -- the communications object to check against
  // <li><b>status</b> -- the status of the message (tag, sender, size)
  // </ul>
  //#--------------------------------------------------------------------------
  void checkStatus(HPCxx_Comm* comm, MPI_Status* status);

  //#---- communicator --------------------------------------------------------
  // Returns a communications object.
  // <ul>
  // <li><b>dest</b> -- the destination of the message
  // <li><b>msgType</b> -- the message type to send
  // <li><b>unique</b> -- use a unique message id if true
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_Comm* communicator(const HPCxx_ContextID& dest, int msgType,
			   bool unique = false);

  //#---- handleMessage -------------------------------------------------------
  // Handles an incoming message.
  // <ul>
  // <li><b>status</b> -- the status of the message (tag, sender, size)
  //#--------------------------------------------------------------------------
  void handleMessage(HPCxx_Comm& comm);

  //#---- release -------------------------------------------------------------
  // Releases what had been an "in use" communications object.
  // <ul>
  // <li><b>comm</b> -- the communication object to release
  // </ul>
  //#--------------------------------------------------------------------------
  void release(HPCxx_Comm* comm);

  //#---- releaseToPosted -----------------------------------------------------
  // Releases a communications object back to the coordinator to be placed
  // in the posted queue.
  // <ul>
  // <li><b>comm</b> -- the communication object to release
  // </ul>
  //#--------------------------------------------------------------------------
  void releaseToPosted(HPCxx_Comm* comm);

  //#---- releaseToPreposted --------------------------------------------------
  // Releases a communications object back to the coordinator to be placed
  // in the preposted queue (those objects that have preposted a receive).
  // <ul>
  // <li><b>comm</b> -- the communication object to release
  // </ul>
  //#--------------------------------------------------------------------------
  void releaseToPreposted(HPCxx_Comm* comm);

  //#---- tag -----------------------------------------------------------------
  // Returns a tag based on message type, message id, and kernel-level thread #
  // <ul>
  // <li><b>msgType</b> -- the message type to use in constructing the tag
  // <li><b>msgID</b> -- the message id to use in constructing the tag
  // </ul>
  //#--------------------------------------------------------------------------
  static int tag(int msgType, int msgID);

  //#---- commMutex -----------------------------------------------------------
  // Returns the communication mutex to be used to protect communication
  // critical sections.
  //#--------------------------------------------------------------------------
  static HPCxx_Mutex* commMutex(void);

  //#---- setCommMutex --------------------------------------------------------
  // Sets the communication mutex to be used to protect communication critical
  // sections.
  //#--------------------------------------------------------------------------
  static void setCommMutex(HPCxx_Mutex* mutex);

  //#---- poll ----------------------------------------------------------------
  // Polls communication queues for events.
  //#--------------------------------------------------------------------------
  void poll(void);

 private:

  static HPCxx_Mutex* commMutex_;	// mutex to protect communication

  int nCommunicators_;	// the number of comm objects that have been created

  vector<int> msgIDs_;		   // container for unique message ids
  list<HPCxx_Comm*> readyComms_;   // container for ready (to use) msg objects
  list<HPCxx_Comm*> postedComms_;  // container for posted (in use) msg objects
  list<HPCxx_Comm*> prepostedComms_;  // container for preposted msg objects

  //#--------------------------------------------------------------------------
  // Increases the communication object stack size.
  //#--------------------------------------------------------------------------
  void increaseCommStackSize(void);

  //#--------------------------------------------------------------------------
  // Returns a unique message id if one exists, MSG_ID_NOT_FOUND otherwise.
  //#--------------------------------------------------------------------------
  int  uniqueMsgID(void);
};

#endif // __HPCXX_COMMCOORD_H_
