// HPC++ Run Time Sytem Library (pthread)
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.C,v 1.2 1998/07/02 23:15:24 rasmussn Exp $

#include <iostream.h>
#include <hpcxx_types.h>
#include <hpcxx_rts.h>
#include <hpcxx_rtsImpl.h>
#include <HPCxx_CommCoord.h>
#include <HPCxx_RFITable.h>
#include <mpi.h>
#include <string.h>

//
// Note: in this MPI implementation, an HPCxx_ContextID is an integer
//
static int                nContexts_;		// number of contexts
static int		  currentIdIndex_;	// current context index/id
static HPCxx_ContextID*   context_ids;		// array of context ids
static HPCxx_ContextID**  contextIDs_;		// array of ptrs to context ids

//
// Barrier global variables
//
#define BARRIER_DONE -1

static bool remote_execute_flag_ = false;
static bool fanOutReceived_ = false;		// true if fan out received
static int nBarrierFanIn_ = 0;			// number fan in msgs
static int nChildren_ = 0;			// number of children
static HPCxx_ContextID* parent_ = NULL;		// parent context
static HPCxx_ContextID* leftChild_ = NULL;	// context of left child
static HPCxx_ContextID* rightChild_ = NULL;	// context of right child

// a global pointer to the current context
HPCxx_Context* hpcxx_currentContext;

// the communication coordinator, should eventually be an array
HPCxx_CommCoord* hpcxx_commCoord_;


//---- hpcxx_basename ---------------------------------------------------------
// Needed because of trouble finding basename on some machines.
//-----------------------------------------------------------------------------
static const char* hpcxx_basename(const char* path)
{
  if (path == NULL) return NULL;
  else if ((strrchr(path, '/')) == NULL) return path;
  else return (strrchr(path, '/') + 1);
}


//---- hpcxx_init -------------------------------------------------------------
// Initializes an HPC++ session.  This function should be called before other
// HPC++ calls are made.  When session is over, call hpcxx_exit.
//-----------------------------------------------------------------------------
int hpcxx_init(int& argc, char**& argv, HPCxx_Group*& group)
{
  MPI_Init(&argc, &argv);

  // determine how many contexts there are
  MPI_Comm_size(MPI_COMM_WORLD, &nContexts_);

  // determine the local context id
  MPI_Comm_rank(MPI_COMM_WORLD, &currentIdIndex_);

  // create the communication coordinator
  hpcxx_commCoord_ = new HPCxx_CommCoord();
  
  // create the array of context ids and an array of pointers to them

  context_ids = new HPCxx_ContextID  [nContexts_];
  contextIDs_ = new HPCxx_ContextID* [nContexts_];
  for (int i = 0; i < nContexts_; i++) {
    context_ids[i] = i;
    contextIDs_[i] = &context_ids[i];
  }

  if (currentIdIndex_ == 0) {			// become the manager

    // create this group
    group = new HPCxx_Group(contextIDs_, nContexts_, 
		            HPCXX_GEN_LOCAL_GROUPID, hpcxx_basename(argv[0]));
    group->setLocalIndex(currentIdIndex_);
    hpcxx_currentContext =  new HPCxx_Context( contextIDs_[currentIdIndex_],
					       true, "manager_context" );

  } else {					// become a worker

    // create this group
    group = new HPCxx_Group( contextIDs_, nContexts_,
			     HPCXX_GEN_LOCAL_GROUPID, hpcxx_basename("child"));

    group->setLocalIndex(currentIdIndex_);
    hpcxx_currentContext =  new HPCxx_Context( contextIDs_[currentIdIndex_],
					       true, "worker_context" );
  }

  //
  // initialize barrier variables
  //

  if (currentIdIndex_ != 0) {
    parent_ = contextIDs_[ (currentIdIndex_ - 1)/2 ];
  }
  if (nContexts_ > 2*currentIdIndex_ + 1) {
    leftChild_ = contextIDs_[ 2*currentIdIndex_ + 1 ];
    nChildren_ += 1;
  }
  if (nContexts_ > 2*currentIdIndex_ + 2) {
    rightChild_ = contextIDs_[ 2*currentIdIndex_ + 2 ];
    nChildren_ += 1;
  }

  hpcxx_barrier();
  return HPCXX_SUCCESS;
}


//---- hpcxx_exit -------------------------------------------------------------
// Cleans up at the end of a session.  This function should be called after
// all HPC++ calls have been made.
//-----------------------------------------------------------------------------
int hpcxx_exit(HPCxx_Group*& group)
{
  MPI_Finalize();
  return HPCXX_SUCCESS;
}


//---- hpcxx_poll -------------------------------------------------------------
// Probes for incoming messages.
//-----------------------------------------------------------------------------
void hpcxx_poll()
{
  hpcxx_commCoord_ -> poll();
}

void *hpcxx_malloc(size_t size)
{
  return malloc(size);
}

void hpcxx_free(void *ptr)
{
  if (ptr) free(ptr);
}


//---- hpcxx_getNumContexts----------------------------------------------------
// Returns the number of contexts in this computation.
//-----------------------------------------------------------------------------
int hpcxx_getNumContexts()
{
  return nContexts_;
}


//---- hpcxx_getNumNodes ------------------------------------------------------
// Returns the number of nodes in this computation.
//-----------------------------------------------------------------------------
int hpcxx_getNumNodes()
{
  return nContexts_;		// for now at least
}


// Get a list of all nodes
// User must 'delete []' the null terminated node array
HPCxx_GlobalPtr<HPCxx_Node>* hpcxx_getNodes()
{
  // FIXME - not implemented
  return((HPCxx_GlobalPtr<HPCxx_Node> *)0);
}


//---- hpcxx_msg --------------------------------------------------------------
// Prints out an error in the master context only.
//-----------------------------------------------------------------------------
void hpcxx_msg(const char *msg)
{
  if (currentIdIndex_ == 0) {
    cerr << msg << endl;
  }
}

// Abort the hpcxx computation.  hpcxx_abort() will call hpcxx_exit()
void hpcxx_abort(int)
{
  HPCxx_Group* group = NULL;
  hpcxx_exit(group);
}


//---- hpcxx_contextid_equal --------------------------------------------------
// Returns true if the context ids are equal.
//-----------------------------------------------------------------------------
bool hpcxx_contextid_equal(HPCxx_ContextID id1, HPCxx_ContextID id2)
{
  return (id1 == id2);
}


//---- hpcxx_readHandler ------------------------------------------------------
// MPI Handler for GlobalPointer reads.  A get request has *ARRIVED* from a
// remote processor (AskForDataMsg).  To satisfy the request, this processor
// must package up the requested block of memory into possibly several chunks,
// and send them to the remote node.
//-----------------------------------------------------------------------------
void hpcxx_readHandler(HPCxx_Comm& comm)
{
  int byteCount, byteOffset;
  unsigned long localPtr;
  HPCxx_Buffer& buffer = comm.messageBuffer();

  buffer.unpack(&localPtr, 1);
  buffer.unpack(&byteCount, 1);
  buffer.unpack(&byteOffset, 1);

  // return to sender
  comm.setMessageForReturn();
  comm.setMessageType(EndOfFetchMsg);

  char* ptr = (char*) localPtr + byteOffset;
  comm.send(ptr, byteCount);

  // release the communication object
  hpcxx_commCoordinator().releaseToPosted(&comm);
}


//---- hpcxx_writeHandler -----------------------------------------------------
//-----------------------------------------------------------------------------
void hpcxx_writeHandler(HPCxx_Comm& comm)
{
  int byteCount, byteOffset;
  unsigned long localPtr;
  HPCxx_ContextID caller;
  HPCxx_Buffer& buffer = comm.messageBuffer();

  // unpack the global pointer, the count, offset, and data

  hpcxx_unpack(buffer, &caller, 1);
  buffer.unpack(&localPtr);
  buffer.unpack(&byteCount);
  buffer.unpack(&byteOffset);
  buffer.unpack((char*)(localPtr + byteOffset), byteCount);

  // release the communication object
  hpcxx_commCoordinator().release(&comm);
}


void hpcxx_invokeRMIHandler(HPCxx_Comm& comm)
{
  int funcID;
  unsigned long localPtr;
  HPCxx_ContextID caller;
  HPCxx_Buffer& buffer = comm.messageBuffer();

  // unpack the function id
  buffer.unpack(&funcID, 1);

  // Unpack the global pointer (this is cheating, it should be done by
  // hpcxx_unpack, but we don't know the type yet).  This is done to get
  // the calling context and to advance buffer to the args.

  hpcxx_unpack(buffer, &caller, 1);
  buffer.unpack(&localPtr, 1);

  // return to sender
  comm.setMessageForReturn();
  comm.setMessageType(AckReplyMsg);

  // find the function object and execute the function

  HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(funcID);
  if (!e) {
    cerr << "hpcxx_invokeGFHandler: failed to find " << funcID << endl;
    return;
  }
  e->execute(&comm, NULL, buffer); 
}


// MPI Handler for Remote (global) Function invocation

void hpcxx_invokeGFHandler(HPCxx_Comm& comm)
{
  int funcID;
  unsigned long localPtr;
  HPCxx_ContextID caller;
  HPCxx_Buffer& buffer = comm.messageBuffer();

  // unpack the function id
  buffer.unpack(&funcID, 1);

  // Unpack the global pointer (this is cheating, it should be done by
  // hpcxx_unpack, but we don't know the type yet).  This is done to get
  // the calling context and to advance buffer to the args.

  hpcxx_unpack(buffer, &caller, 1);
  buffer.unpack(&localPtr, 1);

  // return to sender
  comm.setMessageForReturn();
  comm.setMessageType(AckReplyMsg);

  // find the function object and execute the function

  HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(funcID);
  if (!e) {
    cerr << "hpcxx_invokeGFHandler: failed to find " << funcID << endl;
    return;
  }
  e->execute(&comm, NULL, buffer); 
}


void hpcxx_returnHandler(HPCxx_Comm& comm)
{
  int funcID;
  HPCxx_ContextID caller;
  unsigned long localPtr;

  HPCxx_Buffer& buffer = comm.messageBuffer();

  // unpack the function id
  buffer.unpack(&funcID, 1);

  // Unpack the global pointer (this is cheating, it should be done by
  // hpcxx_unpack, but we don't know the type yet).  This is done to get
  // a pointer to the return sync variable.

  hpcxx_unpack(buffer, &caller, 1);
  buffer.unpack(&localPtr, 1);

  HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(funcID);
  if (!e) {
    cerr << "hpcxx_returnHandler: failed to find " << funcID << endl;
    return;
  }
  e->returnValue((void*)localPtr, buffer); 
}


//-----------------------------------------------------------------------------
// Sends the return value back to the calling context.
//-----------------------------------------------------------------------------
void hpcxx_returnGlobalFunction(HPCxx_Buffer& resultBuf, void* commPtr)
{
  HPCxx_Comm& comm =   *(HPCxx_Comm*) commPtr;
  HPCxx_Buffer& commBuf = comm.messageBuffer();

  // Reset the communications buffer so that it points immediately after
  // the function id and the global pointer to the return sync variable.

  commBuf.impl_.reuse(12);

  // Copy the results buffer for final send.

  int size = resultBuf.impl_.size() - sizeof(int);
  resultBuf.impl_.reuse(sizeof(int));	// skip function id
  commBuf.impl_.copyBuffer(resultBuf, size);

  // send the return value back to the calling context

  comm.send(commBuf);

  // release the communication object
  hpcxx_commCoordinator().releaseToPosted((HPCxx_Comm*) commPtr);
}


// returns the local comm coordinator which will build communication objects
HPCxx_CommCoord& hpcxx_commCoordinator(void)
{
  return *hpcxx_commCoord_;
}


//-----------------------------------------------------------------------------
// BARRIER RELATED FUNCTIONS 
//
// Barrier to synchronize operations in pC++. Processors wait at the
// Barrier until all are there.  Processors need to run tulip_Poll() while
// they wait.  tulip_Barrier() must use messages on most message passing
// machines, so that tulip_Poll() can be run periodically while waiting at
// the barrier.  These routines were re-worked by P. Beckman (SP1 port)
// based on earlier designs by D.  Gannon (Paragon port).  Processors do
// a tree fan-in to the root node, then a fan-out and resume computation.
// A fast hardware/OS supported barrier is prefered, but the need to run
// tulip_Poll() while waiting in the barrier excludes most system
// implementations
// 
// The Barrier works in the following way: The leaves start sending
// FanInMsg to their parents, while their parents poll messages from the
// children.  Once a processor receives messages from all of its
// children, it sends a message to its parent.  The fan-in completes when
// processor zero gets the message.  Processor zero then sends FanOutMsg
// down through the tree.  Barrier completes when all processors receive
// FanOutMsg. (SY & PHB)
//
//-----------------------------------------------------------------------------

//---- hpcxx_barrier ----------------------------------------------------------
//-----------------------------------------------------------------------------
void hpcxx_barrier()
{
  // simply return when only a single processor is used
  if(nChildren_ == 0 && parent_ == NULL) return;

  while (remote_execute_flag_) hpcxx_poll();
  
  HPCxx_Comm* comm;
  HPCxx_CommCoord& coord = hpcxx_commCoordinator();

  // wait for all children to report fan in
  while (nBarrierFanIn_ != nChildren_) {
    hpcxx_poll();
    HPCxx_Thread::sleep(1);
  }

  if (parent_) { 
    // pass the fan in up the heirarchy
    comm = coord.communicator(*parent_, BarrierFanInMsg, false);
    comm->setSender(currentIdIndex_);
    comm->send();
    coord.releaseToPosted(comm);
  } else {
    // at the root so begin fan out
    fanOutReceived_ = true;
  }
  
  // wait for parent to report fan out
  while(!fanOutReceived_) {
    hpcxx_poll();
    HPCxx_Thread::sleep(1);
  }
  
  // pass on fan out to children
  if (leftChild_) {
    comm = coord.communicator(*leftChild_,  BarrierFanOutMsg, false);
    comm->setSender(currentIdIndex_);
    comm->send();
    coord.releaseToPosted(comm);
  }
  if (rightChild_) {
    comm = coord.communicator(*rightChild_, BarrierFanOutMsg, false);
    comm->setSender(currentIdIndex_);
    comm->send();
    coord.releaseToPosted(comm);
  }
  
  nBarrierFanIn_ = 0;
  fanOutReceived_ = false;
}


//---- hpcxx_barrierFanIn -----------------------------------------------------
// Handler for a barrier fan in message from child context.
//-----------------------------------------------------------------------------
void hpcxx_barrierFanIn(HPCxx_Comm& comm)
{
  nBarrierFanIn_ += 1;  
}


//---- hpcxx_barrierFanOut ----------------------------------------------------
//-----------------------------------------------------------------------------
void hpcxx_barrierFanOut(HPCxx_Comm& comm)
{
  fanOutReceived_ = true;
}
