/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* This file contains the prototypes and struct/type declarations for
the portable runtime system: Tulip.  Every application linking to
Tulip should include this file */

/* Pete Beckman */

#ifndef TULIP_H
#define TULIP_H

#include <stdlib.h>	/* to get size_t */

/* define basic structs and types.  Programmers do not need to look
into tulip_sys.h, all information about tulip types and prototypes is
here */

#include "tulip_sys.h"

/*********************************************************************/
/** System Macros **/
#define TULIP_WARNING 0
#define TULIP_FATAL   1

/* ********************************************************************
 * Global Pointers: A global pointer is an object that uniquely
 * describes a physical memory location somewhere in the machine
 * *******************************************************************/

/* The machine dependant data type "tulip_GlobalPointer_t" is defined
   in tulip_sys.h */

/* ********************************************************************
 * Local pointer type definition
 * *******************************************************************/
typedef void* tulip_LocalPointer_t;

/* ********************************************************************
 * ACK handles: An ACK (acknowledge) handle lets test or wait on a
 * pending operation (put/get, RSR)
 * *******************************************************************/

/* the machine dependant data type "tulip_ACK_t" is defined in
   tulip_sys.h */

/* ********************************************************************
 * Command line parameters struct: Programs that use the portable
 * run-time system pass command line arguments to the run-time, which
 * extracts the ones that begin with -pcxx_
 * *******************************************************************/

/* the type "pcxx_PARAM_t" is defined in tulip_sys.h */

/* ********************************************************************
 * Context typedefs
 * *******************************************************************/
typedef int tulip_context_t;		/* context identifier */

/* ********************************************************************
 * Remote service request (rsr) types.
 * *******************************************************************/
typedef char* tulip_key_t;		/* type used as keys to rsr table */
typedef int tulip_rsr_index_t;		/* index into a dispatch table */
typedef void* tulip_arg_t;		/* generic argument */
typedef void (*tulip_rsr0_t)(tulip_context_t sender);
typedef void (*tulip_rsr1_t)(tulip_context_t sender,
			     tulip_arg_t arg);
typedef void (*tulip_rsr2_t)(tulip_context_t sender,
			     tulip_arg_t arg1,
			     tulip_arg_t arg2);
typedef void (*tulip_rsrN_t)(tulip_context_t sender,
			     tulip_LocalPointer_t lptr,
			     size_t nBytes);

/*********************************************************************/
/* Symbols and procedures defined in the RTS */

#ifdef __cplusplus
extern "C" {
#endif

/* The blocking factor for messages in bytes.  All "get/put"
   operations on sizes greater than tulip_MsgBlockSize are chopped up
   into smaller messages.  A value of -1 means NO blocking */
extern int  tulip_MsgBlockSize;


/*********************************************************************/
/*********************************************************************/
/****************** INITIALIZATION AND FINISHING *********************/
/*********************************************************************/
/*********************************************************************/

/*****************************************************************************
 * Initializes the Tulip run time system so that it is in an usable/running
 * state. The arguments will be parsed and any tulip specific arguments will
 * be removed.
 *
 * pArgc (inout) : the number of arguments, on output, the new number of args
 * argv  (inout) : array of arguments, on output, the new argument list
 *****************************************************************************/
void tulip_Init(int* pArgc, char* argv[]);

/*****************************************************************************
 * Shutsdown the Tulip run time system by freeing system resources.  After
 * calling this function, the RTS will be in an unusable/nonrunning state.
 *****************************************************************************/
void tulip_Shutdown(void);

/*********************************************************************/
/*********************************************************************/
/************************ PRIMARY FUNCTIONS **************************/
/*********************************************************************/
/*********************************************************************/

/* ********************************************************************
 * Allocate memory from the shared memory pool.  Of course on a
 * distributed memory machine this is a plain old malloc(), however,
 * on a machine such as the T3D or some other cool machine with good
 * memory hardware, this does the right stuff to get memory that can
 * be shared, and be the target of PUT/GET calls.
 * ***************************************************************** */
void *tulip_SharedMalloc(int x);

/* ********************************************************************
 * Free memory from the shared memory pool.
 * *******************************************************************/
void tulip_SharedFree(void *);

/* ********************************************************************
 * Reallocates memory.  Similar to realloc() except that NULL can be 
 * passed as the object to be reallocated.
 * ***************************************************************** */
void* tulip_Realloc(void* p, size_t size);

/* ********************************************************************
 * Free memory from the shared memory pool.
 * *******************************************************************/
void tulip_SharedFree(void *);

/* ********************************************************************
 * The Context Number (zero-based rank) of this context, this requires
 * a mapping, and should things be very dynamic (contexts add and
 * delete themselves) this may require communication.
 * ***************************************************************** */
int tulip_MyContext();

/* ********************************************************************
 * The total number of contexts in this configuration.
 * ***************************************************************** */
int tulip_NumContexts();

/* ********************************************************************
 * Check for incomming messages, RemoteAction requests, etc
 * ***************************************************************** */
void tulip_Poll(); 

/* ********************************************************************
 * Abort the program with and error.
 * ***************************************************************** */
void tulip_Abort();

/* ********************************************************************
 * Send a message to the console
 * ***************************************************************** */
void tulip_Message(char *string, int flag);

/* ********************************************************************
 * Synchronize ALL the contexts before continuing
 * ***************************************************************** */
void tulip_Barrier();

/* ********************************************************************
 * Begin User-defined code:
 * ***************************************************************** */

/* This will eventually be changed... but is very convenient for now */

int tulip_ParallelMain(int argc, char **argv);

/* ********************************************************************
 * ********************************************************************
 * ACKNOWLEDGE FUNCTIONS: "ACK"s.
 * ********************************************************************
 * ***************************************************************** */

/* --------------------------------------------------------------------
 * Set an ACK.  This is an INTERNAL system function, and not needed by
 * general users.
 * ----------------------------------------------------------------- */
void tulip_SetACK(tulip_ACK_t *ackhandle);

/* --------------------------------------------------------------------
 * Clear an ACK. This is an INTERNAL system function, and not needed
 * by general users.
 * ----------------------------------------------------------------- */
void tulip_ClearACK(tulip_ACK_t *ackhandle);

/* --------------------------------------------------------------------
 * Wait on a "handle" until it has completed (blocking)
 * ----------------------------------------------------------------- */
void tulip_WaitACK(tulip_ACK_t *handle);

/* --------------------------------------------------------------------
 * Check if a "handle" has completed (non-blocking)
 * ----------------------------------------------------------------- */
int tulip_TestACK(tulip_ACK_t *handle);

/* --------------------------------------------------------------------
 * Create/Allocate <count> ACK handles.  Returns NULL on error
 * ----------------------------------------------------------------- */
tulip_ACK_t *tulip_ACKCreate(int count);

/* --------------------------------------------------------------------
 * Destroy/Deallocate an ACK handle.
 * ----------------------------------------------------------------- */
void tulip_ACKDestroy(tulip_ACK_t *handle);


/* ********************************************************************
 * ********************************************************************
 * GLOBAL POINTER FUNCTIONS
 * ********************************************************************
 * ***************************************************************** */

/* ----------------------------------------------------------------------------
 * Makes a global pointer.  The calling function is responsible for freeing
 * the global pointer.
 *
 * key    (input) : a registration key used to identify the global pointer
 * nBytes (input) : the size of the buffer to allocate
 * returns : the global pointer (NULL if error)
 * ------------------------------------------------------------------------- */
tulip_GlobalPointer_t* tulip_MakeGlobalPointer(tulip_key_t key, size_t nBytes);

/* ----------------------------------------------------------------------------
 * Returns the global pointer associated with the given key (NULL if
 * not found).
 *
 * key    (input) : the registration key by which the global pointer is known
 * ------------------------------------------------------------------------- */
tulip_GlobalPointer_t* tulip_GetGlobalPointer(tulip_key_t key);

/* ----------------------------------------------------------------------------
 * Registers a global pointer by associating it with the given key/name.
 *
 * key  (input) : a registration key used to identify the global pointer
 * gptr (input) : the global pointer to register
 * ------------------------------------------------------------------------- */
void tulip_RegisterGlobalPointer(tulip_key_t key, tulip_GlobalPointer_t* gptr);

/* --------------------------------------------------------------------
 * Change the LocalPointer portion of a GlobalPointer
 * ----------------------------------------------------------------- */
void tulip_SetLocalPointer(tulip_GlobalPointer_t *, void *);

/* --------------------------------------------------------------------
 * Change the Context portion of a GlobalPointer
 * ----------------------------------------------------------------- */
void tulip_SetContext(tulip_GlobalPointer_t *, int);

/* --------------------------------------------------------------------
 * Extract the Context (0-(n-1))
 * ----------------------------------------------------------------- */
int tulip_ToContext(tulip_GlobalPointer_t *);

/* --------------------------------------------------------------------
 * Create a GlobalPointer from a local one (for this context)
 * ----------------------------------------------------------------- */
tulip_GlobalPointer_t *tulip_ToGlobalPointer(void *);

/* --------------------------------------------------------------------
 * Extract the LocalPointer part of a GlobalPointer
 * ----------------------------------------------------------------- */
void *tulip_ToLocalPointer(tulip_GlobalPointer_t *);


/* ********************************************************************
 * ********************************************************************
 * DATA MOVEMENT
 * ********************************************************************
 * ***************************************************************** */

/* ********************************************************************
 * Store bytes into location designated by tulip_GlobalPointer_t
 * ***************************************************************** */
int tulip_Put(tulip_GlobalPointer_t dst, char *src, int length, 
	      tulip_ACK_t *handle);

/* ********************************************************************
 * Retrieve bytes from location designated by tulip_GlobalPointer_t
 * ***************************************************************** */
int tulip_Get(char *dst, tulip_GlobalPointer_t src, int length,
	      tulip_ACK_t *handle);

/*****************************************************************************
 * Sends RemoteAction request to remote processor.
 *
 * remote_node (input) : the receiving context of the remote action
 * buffer      (input) : data buffer, could be additional parameters
 * length      (input) : the size of the data buffer
 * ackhandle   (input) : acknowlegement handle (no ack sent if NULL)
 *****************************************************************************/
int tulip_RemoteAction(int remote_node, char *buffer, int length, 
		       tulip_ACK_t *ackhandle);

/*****************************************************************************
 * Sends a RemoteAction request to a remote processor with a rmi/rsr key.
 *
 * receiver  (input) : the receiving context of the remote action
 * key       (input) : a key used to select a particular remote action
 * lptr      (input) : local pointer to data buffer or additional parameters
 * nBytes    (input) : the size of the data buffer
 * ackHandle (input) : acknowlegement handle (no ack sent if NULL)
 *****************************************************************************/
void tulip_RemoteActionKeyN(tulip_context_t receiver, tulip_key_t key,
			    tulip_LocalPointer_t lptr, size_t nBytes,
			    tulip_ACK_t* ackHandle);

/* ********************************************************************
 * USER CODE: What to do if I receive a RemoteAction
 * ***************************************************************** */
void tulip_RemoteActionHandler(int from_whom, char *buffer, int length);

/*****************************************************************************
 * Services RemoteAction requests by calling the appropriate function.
 *
 * sender (input) : the sending context of the remote action
 * key    (input) : a key used to select a particular remote action
 * lptr   (input) : a local pointer to data buffer, could be additional params
 * nBytes (input) : the size of the data buffer
 *****************************************************************************/
void tulip_RemoteActionHandlerKeyN(tulip_context_t sending, tulip_key_t key,
				   tulip_LocalPointer_t lptr, size_t nBytes);

/*****************************************************************************
 * Registers a function so that it can be called remotely.
 *
 * key  (input) : a key used to select the function
 * rsr  (input) : the function to register
 *****************************************************************************/
void tulip_RegisterRSRNHandler(tulip_key_t key, tulip_rsrN_t rsr);

/* ********************************************************************
 * ********************************************************************
 * COLLECTIVE OPERATIONS
 * ********************************************************************
 * ***************************************************************** */
void tulip_NativeReduceDoubleAdd(void *,int);
void tulip_NativeReduceFloatAdd (void *,int);
void tulip_NativeReduceIntAdd   (void *,int);
void tulip_NativeReduceDoubleMax(void *,int);
void tulip_NativeReduceFloatMax (void *,int);
void tulip_NativeReduceIntMax   (void *,int);
void tulip_NativeReduceDoubleMin(void *,int);
void tulip_NativeReduceFloatMin (void *,int);
void tulip_NativeReduceIntMin   (void *,int);
void tulip_NativeBroadcastBytes (int,int,void *);

void tulip_AlienReduceDoubleAdd(void *,int);
void tulip_AlienReduceFloatAdd (void *,int);
void tulip_AlienReduceIntAdd   (void *,int);
void tulip_AlienReduceDoubleMax(void *,int);
void tulip_AlienReduceFloatMax (void *,int);
void tulip_AlienReduceIntMax   (void *,int);
void tulip_AlienReduceDoubleMin(void *,int);
void tulip_AlienReduceFloatMin (void *,int);
void tulip_AlienReduceIntMin   (void *,int);
void tulip_AlienBroadcastBytes (int,int,void *);

void tulip_ReduceDoubleAdd(void *,int);
void tulip_ReduceFloatAdd (void *,int);
void tulip_ReduceIntAdd   (void *,int);
void tulip_ReduceDoubleMax(void *,int);
void tulip_ReduceFloatMax (void *,int);
void tulip_ReduceIntMax   (void *,int);
void tulip_ReduceDoubleMin(void *,int);
void tulip_ReduceFloatMin (void *,int);
void tulip_ReduceIntMin   (void *,int);
void tulip_BroadcastBytes (int,int,void *);

#ifdef __cplusplus
}
#endif

/* Define event macros */
#include "pcxx_events_def.h"

/* Define the super-fast timer functions */
#include "pcxx_timers.h"

#endif
