//-----------------------------------------------------------------------------
// Remote.C - Implementation of the Remote category of classes.
//
// Copyright (c) 1997, Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// 
// Craig Rasmussen, Pete Beckman
//
//-----------------------------------------------------------------------------
// $Id: Remote.C,v 1.1 1998/06/23 00:07:11 ejohnson Exp $
//-----------------------------------------------------------------------------

#include "Remote.h"
#include "hpcxx_rts.h"
#include <hpcxx_rts.cc>
#include "tulip.h"
#include "Profile/Profiler.h"

// find out where to include these function prototypes
extern "C" {
int pcxx_InitParam (int *argc, char *argv[]);
}

int SMRTSSetup(void);
int SMRTSFork(int argc, char **argv);
int SMRTSForkFinish(void);

bool ParallelRegistry::fExitPoll = false;

ParallelRegistry::ParallelRegistry(int nContexts)
//-----------------------------------------------------------------------------
// Constructor.  Starts nContexts-1 number of worker threads and has them
// enter a polling loop, waiting for incoming messages.  The current thread is
// also available for work and participates in the barriers, so that the
// total number of threads available for work is nContexts.
//
// nContexts  the total number of contexts to be run
//-----------------------------------------------------------------------------
{
  // make argv strings for pcxx_InitParam() call

  int iArgc = 3;
  char sznContexts[16];
  char* aszArgv[3] = { "app", "-hpcxx_NUMPROC", sznContexts };
  sprintf(sznContexts, "%d", nContexts);

  // initialize and fork workers

  ParallelRegistry::fExitPoll = false;	// ensure that worker threads poll
					// for incoming messages
  pcxx_InitParam(&iArgc, aszArgv);
  SMRTSSetup();
  SMRTSFork(iArgc, aszArgv);
}


ParallelRegistry::~ParallelRegistry(void)
//-----------------------------------------------------------------------------
// Destructor.  
//-----------------------------------------------------------------------------
{
  // need to inform workers that the registry is closing down and they
  // should stop polling.  The default action of the base remote object
  // is used to do this.

  RemoteObject* pro = new (HPCXX_NEW) RemoteObject();
  pro -> remoteAction();
  pro -> destroy();

  SMRTSForkFinish();
}


//void ParallelRegistry::bind(const string& name, Remote* pRemote)
//-----------------------------------------------------------------------------
// Registers the remote object with the registry.
//
// name     the name that the object is to be known by
// pRemote  pointer to the remote object to register
//-----------------------------------------------------------------------------
//{
  // need to implement something useful, just provide the API for now
//}


Registry* LocateParallelRegistry::createRegistry(int iPort, int nContexts)
//-----------------------------------------------------------------------------
// Creates a registry object for running remote objects in a given number of
// contexts.
//
// iPort      the port over which the registry objects are to communicate
// nContexts  the total number of contexts to be run
//-----------------------------------------------------------------------------
{
  return new ParallelRegistry(nContexts);
}


// Global Variables

tulip_ACK_t* h = NULL;
int RemoteObject::Default_CMD  = 0;

RemoteObject::RemoteObject(void)
//-----------------------------------------------------------------------------
// Constructor.
//-----------------------------------------------------------------------------
{
  iOpCode = 0;
}


void RemoteObject::destroy(void)
//-----------------------------------------------------------------------------
// Explicityly calls the destructor on the object and frees the memory.
// This must be used if memory is allocated from the shared area.
//-----------------------------------------------------------------------------
{
  hpcxx_delete(this);
}


int RemoteObject::getContext(void)
//-----------------------------------------------------------------------------
// Returns the context that this object is running in.
//-----------------------------------------------------------------------------
{
  return tulip_MyContext();
}


int RemoteObject::numberOfContexts(void)
//-----------------------------------------------------------------------------
// Returns the total number of contexts.
//-----------------------------------------------------------------------------
{
  return tulip_NumContexts();
}


void RemoteObject::performAction(void)
//-----------------------------------------------------------------------------
// Performs the default action, which for the base class, is to cause
// the registry object to shut down the polling loop.
//-----------------------------------------------------------------------------
{
  ParallelRegistry::exitPolling();
}


void RemoteObject::performAction(int sender, RemoteObject* pro)
//-----------------------------------------------------------------------------
// Performs some specified action, depending on the value of the instance
// variable, pro->iOpCode.
//
// iSender  the context which sent the message
// pro	    a pointer to the remote object to send performAction 
//-----------------------------------------------------------------------------
{
  // nothing specific to do at this level so just perform default action.
  pro -> performAction();
}


void RemoteObject::remoteAction(int iOpCode, bool fSynchronous)
//-----------------------------------------------------------------------------
// Invoke performAction() in every context, including the sending context.
//
// iOpCode        integer coding of the operation to perform
// fSynchronous   synchronize with other contexts before returning
//-----------------------------------------------------------------------------
{
  TAU_TYPE_STRING(taustr, CT(*this) + " void (int)");
  TAU_PROFILE("RemoteObject::remoteAction()", taustr, TAU_MESSAGE);
  TAU_PROFILE_TIMER(timerL, "LocalActionTimer",
		    "in RemoteObject::remoteAction", TAU_MESSAGE);
  TAU_PROFILE_TIMER(timerR, "FinishActionBarrier",
                    "in RemoteObject::remoteAction", TAU_MESSAGE);

  int i;
  RemoteObject* apro[1];
  apro[0] = this;

  // initialize buffer for acknowledgements

  int nContexts = tulip_NumContexts();
  if (h == NULL) {
    // WARNING, make sure that nContexts <= 1024...., size of cache line
    h = (tulip_ACK_t*) tulip_SharedMalloc(sizeof(tulip_ACK_t) * 1024);
  }

  // invoke performAction in every context (other than sending context)

  this->iOpCode = iOpCode;
  int iLen = sizeof(RemoteObject*);
  for (i = 0; i < nContexts; i++) {
    if (i != tulip_MyContext()) {
      if (fSynchronous) {
        tulip_RemoteAction(i, (char*)apro, iLen, &h[i]);
      } else {
        tulip_RemoteAction(i, (char*)apro, iLen, NULL);
      }
    }
  }

  TAU_PROFILE_START(timerL); 
  performAction();	// sending context should execute action also
  TAU_PROFILE_STOP(timerL); 

  // wait for remote actions to complete

  if (fSynchronous) {
    TAU_PROFILE_START(timerR); 
    for (i = 0; i < nContexts; i++) {
      if (i != tulip_MyContext()) {
        tulip_WaitACK(&h[i]);
      }
    }
    TAU_PROFILE_STOP(timerR); 
  }
}


void tulip_RemoteActionHandler(int from_whom, char* buffer, int len)
//-----------------------------------------------------------------------------
// Event handler which services events for this context.
//
// from_whom	the context id of sender
// buffer	a pointer to a shared remote call object which contains
//		  the required parameters for this event, including the
//		  the operation code.
// len		the length of buffer
//-----------------------------------------------------------------------------
{
  TAU_PROFILE("tulip_RemoteActionHandler()", "void (int, char* int)", TAU_MESSAGE);

  // A pointer to a RemoteObject object is stored in buffer, cast it to
  // the RemoteObject class.
  RemoteObject* pro = *((RemoteObject**) buffer);
  RemoteObject::performAction(from_whom, pro);
}


#ifndef SPMD_MODE

int tulip_ParallelMain(int argc, char *argv[])
//-----------------------------------------------------------------------------
// The "main" routine for parallel execution.  At this point the processors are
// executing in SPMD mode.
//
// argc  the number of arguments
// argv  array of arguments
//-----------------------------------------------------------------------------
{ 
  TAU_PROFILE("tulip_ParallelMain()", "int (int, char**)", TAU_DEFAULT);

  TAU_PROFILE_INIT(argc, argv);
  TAU_PROFILE_SET_NODE( tulip_MyContext() );

  // exit if master context
  if ( tulip_MyContext() == tulip_NumContexts() - 1 ) return 0;
  // slave contexts go into idle mode waiting for event
  while ( ParallelRegistry::shouldExitPolling() == false ) {
    tulip_Poll();      // check for event and if exists, handle it
  }

//  TAU_PROFILE_EXIT("exiting tulip_ParallelMain");
  return 0;
}

#endif  // not SPMD_MODE
