/*****************************************************************************
 * hashtable.c
 * 
 * Copyright (c) 1997, Los Alamos National Laboratory
 *****************************************************************************/

/*****************************************************************************
 * Hashtable object for the tulip framework.
 *
 * @author Craig Rasmussen
 * @version 1.1 23 Oct 1997
 *****************************************************************************/

#include <string.h>
#include "hashtable.h"

/*****************************************************************************
 * Compares two keys.  This function assumes that tulip_key_t is a char* 
 * c string.
 *
 * @param  key1		the first key
 * @param  key2		the second key
 *
 * @returns   < 0 if (key1 < key2) : 0 if (key1 == key2) : > 0 if (key1 > key2)
 *****************************************************************************/
static int compareKeys(tulip_key_t key1, tulip_key_t key2)
{
  return strcmp(key1, key2);
}


/*****************************************************************************
 * Makes a copy of a key.  This function assumes that tulip_key_t is a char* 
 * c string.
 *
 * @param key the key to copy
 * @param shared == true (1) if allocated in shared memory
 * @returns a copy of the key
 *****************************************************************************/
static tulip_key_t copyKey(const tulip_key_t key, int shared)
{
  if (shared == 0) {
    return strdup(key);
  } else {
    tulip_key_t newKey = (tulip_key_t) tulip_SharedMalloc(strlen(key) + 1);
    return strcpy(newKey, key);
  }
}


/*****************************************************************************
 * Frees memory for a key.  This function assumes that tulip_key_t is a char* 
 * c string.
 *
 * @param key the key to free
 * @param shared == true (1) if allocated in shared memory
 *****************************************************************************/
static void freeKey(tulip_key_t key, int shared)
{
  if (shared == 0) {
    free(key);
  } else {
    tulip_SharedFree(key);
  }
}


/*****************************************************************************
 * Hash function. This function assumes that tulip_key_t is a char* 
 * c string.
 *
 * @param key the key to hash
 * @returns the hash value of key
 *****************************************************************************/
static unsigned hashKey(tulip_key_t key)
{
  unsigned uHashVal;

  for (uHashVal = 0; *key != '\0'; key++) {
    uHashVal = *key + 31 * uHashVal;
  }
  return uHashVal;
}


/*****************************************************************************
 * Gets the hashtable entry structure associated with the specified key.
 *
 * @param pht a pointer to the hash table
 * @param key the specified key
 * @param iHashVal the hash value of the key
 * @returns the entry structure for the key or NULL if the key
 *          is not defined in the hash table.
 *****************************************************************************/
static struct entry* getEntry(struct Hashtable* pht,
			      tulip_key_t key, int iHashVal)
{
  struct entry* pEntry = NULL;

  for (pEntry = pht->table[iHashVal]; pEntry != NULL; pEntry = pEntry->next) {
    if (compareKeys(key, pEntry->key) == 0) return pEntry;
  }
  return pEntry;
}


/*****************************************************************************
 * Initializes the hash table.
 *
 * @param pht a pointer to the hash table
 * @param iHashSize the size of the hash table
 * @param shared == true (1) if allocated in shared memory
 *****************************************************************************/
void tulip_HashtableInit(struct Hashtable* pht, int iHashSize, int iShared)
{
  int i;

  if (pht == NULL) return;
  if (iHashSize < 8) iHashSize = 8;

  pht->shared   = iShared;
  pht->hashSize = iHashSize;

  if (pht->shared == 0) {
    pht->table = (struct entry**) malloc(iHashSize * sizeof(struct entry*));
  } else {
    pht->table = (struct entry**)
                 tulip_SharedMalloc(iHashSize * sizeof(struct entry*));
  }

  for (i = 0; i < iHashSize; i++) {
    pht->table[i] = NULL;
  }
}


/*****************************************************************************
 * Destroys the hash table.
 *
 * @param pht a pointer to the hash table
 *****************************************************************************/
void tulip_HashtableDestroy(struct Hashtable* pht)
{
  int i;

  if (pht == NULL || pht->table == NULL) return;

  for (i = 0; i < pht->hashSize; i++) {
    struct entry* pEntry = pht->table[i];
    while (pEntry != NULL) {
      struct entry* pNext = pEntry->next;
      freeKey(pEntry->key, pht->shared);
      if (pht->shared == 0) {
        free(pEntry);
      } else {
        tulip_SharedFree(pEntry);
      }
      pEntry = pNext;
    }
  }
  if (pht->table) {
    if (pht->shared == 0) {
      free(pht->table);
    } else {
      tulip_SharedFree(pht->table);
    }
  }
  pht->table = NULL;
}


/*****************************************************************************
 * Gets the object associated with the specified key in the hashtable.
 *
 * @param pht a pointer to the hash table
 * @param key the specified key
 * @returns the element for the key or NULL if the key is not defined in
 *          the hash table.
 *****************************************************************************/
tulip_LocalPointer_t tulip_HashtableGet(struct Hashtable* pht, tulip_key_t key)
{
  struct entry* pEntry;
  int iHashVal;

  if (pht == NULL || key == NULL) return NULL;

  iHashVal = hashKey(key) % pht->hashSize;
  pEntry = getEntry(pht, key, iHashVal);
  if (pEntry == NULL) return NULL;
  return pEntry->value;
}


/*****************************************************************************
 * Puts the object associated with the specified key in the hashtable.
 *
 * @param pht a pointer to the hash table
 * @param key the specified key
 * @param value the specified element to store in the table
 * @returns the old value of the key or NULL if it did not have one
 *****************************************************************************/
tulip_LocalPointer_t tulip_HashtablePut(struct Hashtable* pht, tulip_key_t key,
					tulip_LocalPointer_t value)
{
  int iHashVal;
  struct entry* pEntry = NULL;
  tulip_LocalPointer_t oldValue = NULL;
  
  if (pht == NULL || key == NULL || value == NULL) return NULL;

  iHashVal = hashKey(key) % pht->hashSize;
  pEntry = getEntry(pht, key, iHashVal);

  if (pEntry == NULL) {		/* this is a new entry */
    if (pht->shared == 0) {
      pEntry = (struct entry*) malloc(sizeof(*pEntry));
    } else {
      pEntry = (struct entry*) tulip_SharedMalloc(sizeof(*pEntry));
    }
  } else {
    /* key already present */
    free(pEntry->key);
    oldValue = pEntry->value;
  }
  
  pEntry->next = pht->table[iHashVal];
  pEntry->key = copyKey(key, pht->shared);
  pEntry->value = value;
  pht->table[iHashVal] = pEntry;

  return oldValue;
}
