//-----------------------------------------------------------------------------
// testTulip.C -- code to test the Tulip run time system
//
// Copyright (c) 1997, Los Alamos National Laboratory
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// @author Craig Rasmussen
//-----------------------------------------------------------------------------

#include "tulip.h"
//#include "Profile/Profiler.h"
#include <stdio.h>

int sum_g;

void main(int argc, char* argv[])
{
//  TAU_PROFILE("main()", "int(int,char**)", TAU_DEFAULT);

    tulip_Init(&argc, argv);
    tulip_Shutdown();
}

//-----------------------------------------------------------------------------
// Sets sum_g to provided value.
//
// sender  (input) : the sending context of the remote action
// lptr    (input) : local pointer to data buffer or additional params
// nBytes  (input) : the size of the data buffer
//-----------------------------------------------------------------------------
void setSum(tulip_context_t sender, tulip_LocalPointer_t lptr,
		size_t nBytes)
{
  sum_g = *( (int*) lptr );
}

//-----------------------------------------------------------------------------
// Gets sum_g by calling addSum on provided context.
//
// sender  (input) : the sending context of the remote action
// lptr    (input) : local pointer to data buffer or additional params
// nBytes  (input) : the size of the data buffer
//-----------------------------------------------------------------------------
void getSum(tulip_context_t sender, tulip_LocalPointer_t lptr,
		size_t nBytes)
{
  int receiver = *( (int*) lptr );
  if (receiver != tulip_MyContext()) {
    tulip_RemoteActionKeyN(receiver, "addSum", &sum_g, sizeof(sum_g), NULL);
  }
}

//-----------------------------------------------------------------------------
// Adds provided value to sum_g
//
// sender  (input) : the sending context of the remote action
// lptr    (input) : local pointer to data buffer or additional params
// nBytes  (input) : the size of the data buffer
//-----------------------------------------------------------------------------
void addSum(tulip_context_t sender, tulip_LocalPointer_t lptr,
		size_t nBytes)
{
  sum_g += *( (int*) lptr );
}


//-----------------------------------------------------------------------------
// The "main" routine for parallel execution.  At this point the processors are
// executing in SPMD mode.
//
// argc  the number of arguments
// argv  array of arguments
//-----------------------------------------------------------------------------
int tulip_ParallelMain(int argc, char *argv[])
{
  sum_g = 0;

  tulip_context_t myContext = tulip_MyContext();
  int nContexts = tulip_NumContexts();

  int arrayLength = 1024;
  int procLength = arrayLength/nContexts;
  arrayLength = procLength*nContexts;

  //
  // test global pointer functions and registration
  //

  tulip_GlobalPointer_t* gPtr;
  if (myContext == 0) {
    gPtr = tulip_MakeGlobalPointer("Shared Double Array",
                                   arrayLength*sizeof(double));
    tulip_Barrier();
    if (gPtr == NULL) {
      printf("%d: ERROR couldn't create global pointer\n", myContext);
      return 1;
    }
  } else {
    tulip_Barrier();
    gPtr = tulip_GetGlobalPointer("Shared Double Array");
    if (gPtr == NULL) {
      printf("%d: ERROR couldn't get global pointer\n", myContext);
      return 1;
    }
  }

  // this only works for SMP
  double* array = (double*) tulip_ToLocalPointer(gPtr);

  int i0 = myContext*procLength;
  for (int i = i0; i < i0 + procLength; i++) {
    array[i] = double(i);
  }
  tulip_Barrier();

  // test to see if array has correct values
  i0 += procLength;
  if (i0 == arrayLength) i0 = 0;
  for (i = i0; i < i0 + procLength; i++) {
    if (array[i] != double(i)) {
      printf("%d: ERROR in global pointer test, array[%d] != %lf (=%lf)\n",
             myContext, i, double(i), array[i]);
      return 1;
    }
  }
  tulip_Barrier();

  //
  // test remote action registration
  //

  tulip_RegisterRSRNHandler("setSum", setSum);
  tulip_RegisterRSRNHandler("getSum", getSum);
  tulip_RegisterRSRNHandler("addSum", addSum);

  int receiver = (myContext + 1) % nContexts;
  int sendValue = myContext + receiver;
  tulip_RemoteActionKeyN(receiver, "setSum", &sendValue, sizeof(sendValue),
			 NULL);

  tulip_Barrier();
  tulip_Poll();

  if (myContext != 0) {
    tulip_RemoteActionKeyN(0, "addSum", &sum_g, sizeof(sum_g), NULL);
  }

  tulip_Barrier();
  tulip_Poll();		// make sure no unfinished work to do

  int flag = 0;
  if (myContext == 0) {
    int sum = nContexts * (nContexts - 1);
    if (sum_g != sum) {
      printf("%d: ERROR in accumulated sum %d\n", myContext, sum_g);
    } else {
      printf("%d: Test of Tulip completed successfully\n"
	     "    (unless another context reports an error)\n", myContext);
    }
  } else {
    if (sum_g != (2*myContext - 1)) {
      printf("%d: ERROR in stored sum %d\n", myContext, sum_g);
    }
  }

  return 0;
}
