/*=======================================================================
 $Log: Tulip_ThreadPipe.h,v $
 Revision 1.2  1998/06/23 14:14:21  suvas
 Fixed the problem of the system exiting while there
 are still threads blocked on synchronization primitives.

 Revision 2.1  1998/06/23 00:37:27  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:33  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:14:30  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_ThreadPipe.h,v 1.2 1998/06/23 14:14:21 suvas Exp $
 $Date: 1998/06/23 14:14:21 $
 $Author: suvas $
 ========================================================================*/
//-*-c++-*-
#ifndef _Tulip_ThreadPipe_h_
#define _Tulip_ThreadPipe_h_

#include <vector>
#include <list>
#include "Tulip_Semaphore.h"

template<class Contained, class Container, class SemType>
class Tulip_ThreadPipe
{
  Container container_;
  //
  // Mutex used for atomicity
  //
  Tulip_Mutex mutex_;
  //
  // Semaphore used for get access
  //
  SemType get_access_;

public:
  Tulip_ThreadPipe() : get_access_(0)
  {
    //
    // empty
    //
  }

  void put(const Contained& item)
  {
    mutex_.lock();
    Tulip_Context::GlobalEventsCounter+=1;
    container_.push_back(item);
    mutex_.unlock();
    get_access_.release();
  }

  void get(Contained& item)
  {
    get_access_.acquire();
    mutex_.lock();
    item = container_.front();
    container_.erase(container_.begin());
    Tulip_Context::GlobalEventsCounter-=1;
    mutex_.unlock();
  }

  bool tryGet(Contained& item)
  {
    if ( get_access_.tryAcquire() ) {
      mutex_.lock();
      item = container_.front();
      container_.erase(container_.begin());
      Tulip_Context::GlobalEventsCounter-=1;
      mutex_.unlock();
      return true;
    } else {
      return false;
    }
  }

  long length()
  {
    long retval;
    mutex_.lock();
    retval = container_.size();
    mutex_.unlock();
    return retval;
  }

  //
  // It's not very advisable to use these methods to modify the
  // contents of a container, or certainly not to add or remove items
  // from the container, since the mutex will not remain
  // synchronized. These routines may be useful if you're putting a
  // lot of items into the container, or if you need to access the
  // container to print items.
  //
  void lock() {
    mutex_.lock();
  }

  void unlock() {
    mutex_.unlock();
  }

  const Container& container() {
    return container_;
  }

};

template<class Contained, class Container, class SemType>
class Tulip_BoundedThreadPipe : public Tulip_ThreadPipe<Contained,Container,SemType>
{
  //
  // Semaphore used for get access
  //
  SemType put_access_;
  long max_size_;

public:
  Tulip_BoundedThreadPipe(long max_size) : put_access_(max_size)
  {
    max_size_ = max_size;
  }

  void put(const Contained& item)
  {
    put_access_.acquire();
    Tulip_ThreadPipe<Contained,Container,SemType>::put(item);
  }

  bool tryPut(const Contained& item)
  {
    if ( put_access_.tryAcquire() ) {
      Tulip_ThreadPipe<Contained,Container,SemType>::put(item);
      return true;
    } else {
      return false;
    }
  }

  void get(Contained& item)
  {
    Tulip_ThreadPipe<Contained,Container,SemType>::get(item);
    put_access_.release();
  }

  bool tryGet(Contained& item)
  {
    if ( Tulip_ThreadPipe<Contained,Container,SemType>::tryGet(item) ) {
      put_access_.release();
      return true;
    } else {
      return false;
    }
  }
};

//
// Short cuts
//
template <class Contained>
class Tulip_PipeList : public Tulip_ThreadPipe< Contained, list<Contained>, Tulip_Semaphore>
{
  //
  // Empty
  //
};

template <class Contained>
class Tulip_BoundedPipeList
  : public Tulip_BoundedThreadPipe< Contained, list<Contained>, Tulip_Semaphore>
{

public:
  Tulip_BoundedPipeList(long max_size)
    : Tulip_BoundedThreadPipe< Contained, list<Contained>, Tulip_Semaphore>(max_size)
  {
    //
    // Empty
    //
  }
};



#endif
