// Graph.C  - non-inline part of IDOM package. See Graph.h.

#define LT_SIMPLE 1
// The above selects the L&T non-sophisticated algorithm.
// It runs on g++/Linux and seg faults on the Alpha. Should fix someday.

// HISTORY:
// 6oct94 Don Lindsay <lindsay@cs.colorado.edu> Created.

extern int debug;		// nonzero means do extra work. See DEBUG.
#include "Graph.h"

//---------------------------------------------------------
// "Step 0"
// The nodes don't yet know about edges, or other nodes.
// Traverse edge_list and fill out the nodes.
// Note that really_add_single_exit_node() has similar logic, since it
//	adds more edges, but must be run after this (since it needs to
//	know the indegrees, which are computed here). So, if we need to
//	add more logic to here, we may want to upgrade him, too.

void Graph::propagate_edges_to_nodes()
{
    for( Pix p = edge_list.first(); p; edge_list.next(p) ) {
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	from->successors.append( pe ); from->outdegree++;
	to->predecessors.append( pe ); to->indegree++;
    }
}
//---------------------------------------------------------
// Self-recursive portion of L&T "Step 1".
// Sets node.LT.semi to values obtained via node_upcounter++.
// Add the node.LT.parent arcs that define a spanning tree.
// Build vertex[].
// After this unwinds, any node with .LT.semi==0 is unreachable.
// If post is true, we are doing the postdominators, so use the
// 	predecessors, not the successors, in the calculation.
//	This is achieved with the "list" argument, which is
//	(I believe) actually quite efficient. 
// While we are here anyway, we also compute the rpostorder of each GraphNode,
//	and max_outdegree.

void Graph::depth_first_search( bool post, int v, 
				DLList<GraphEdge*>* list )
{
    node_upcounter++;
    ltarray[v].semi 	  = node_upcounter;
    ltarray[v].label 	  = v;
    ltarray[v].ancestor 	  = 0;
    ltarray[v].child 	  = 0;
    ltarray[v].size	  = 1;

    vertex[ node_upcounter ] = v;	// stamp the array

    // Now visit all unvisited successors (or maybe predecessors).

    for( Pix p = (*list).first(); p; (*list).next(p) ) {
	GraphEdge *pe = (*list)(p);
	GraphNode *pn = post ? (pe->pfrom) : (pe->pto);
	int w = pn->ltindex;
		
	if( ltarray[w].semi == 0 ) {
	    // This successor has not been visited.
	    ltarray[w].parent = v;
	    depth_first_search( post, w,
				post ? &(pn->predecessors):&(pn->successors) );
	}
    }
    ltarray[v].node->rpostorder = node_downcounter;
    node_downcounter--;
    if( max_outdegree < ltarray[v].node->outdegree )
	max_outdegree = ltarray[v].node->outdegree;
}



void Graph::reverse_depth_first_search(GraphNode * node)
{
    assert(node != NULL);

    if (!(node->visited)) {

	node_upcounter++;
        node->visited = TRUE;

 	// Now visit all unvisited successors (or maybe predecessors).

        for( Pix p = node->successors.last(); p; node->successors.prev(p)){  
	    GraphEdge *edge = node->successors(p);
	    reverse_depth_first_search(edge->pto);
        }
	node->lpostorder = node_downcounter;
	node_downcounter--;
      
    }
}

void Graph::reverse_number_the_nodes()
{	
    int count = 0;
    node_upcounter = 0;
    node_downcounter = node_count;	// for lpostorder

    for( Pix p2 = node_map.first(); p2; node_map.next(p2)) {
	GraphNode *n = node_map.data( p2 );
	n->visited = FALSE;
	count ++;
    }
	
    GraphNode *pn = graph_root;

    reverse_depth_first_search(pn);

    if( node_upcounter != node_count ) {
	all_reachable = (bool)FALSE;
	cerr << "REVERSE WARNING: not all nodes reachable: found "
	     << node_upcounter << " of " << node_count << ", " << count
	     << "." << endl;
    }
}

//--------------------------------------------------------- L&T "Step
//1" - Depth First Search.  This may be called twice (postdominator
//and dominator) or once (dominator).  Should be in that order, since
//leaves the rpostorder numbers as a side effect (this would have to
//be a separate DFS if we wanted to reverse the order to dom/postdom).
//While we are here anyway, we also compute the rpostorder of each
//GraphNode and max_outdegree.

void Graph::number_the_nodes( bool post )
{	
    // Create an array for the node numbers to index into.  If this is
    // the nonfirst time, can't re-use the previous array,
    //	because node splitting may have increased node_count.

    if( vertex != 0 ) delete [] vertex;
    vertex = new int[ node_count+1 ];	// L&T name
    if( ltarray != 0 ) delete [] ltarray;
    ltarray = new LT[ node_count+1 ];

    // traverse the node map, cross-stitching ltarray & nodes

    int i = 1;
    for( Pix pix = node_map.first(); pix; node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );
	ltarray[i].node = node;
	ltarray[i].dom = 0;
	node->ltindex = i++;
    }

    vertex[0] = 0;
    node_upcounter = 0;
    node_downcounter = node_count;	// for rpostorder
	
    GraphNode *pn = post ? graph_exit: graph_root;
    depth_first_search( post, pn->ltindex,
			post ? &(pn->predecessors):&(pn->successors) );
		
    node_reachable_count = node_upcounter;
    all_reachable = (bool)TRUE;
    if( node_reachable_count != node_count ) {
	all_reachable = (bool)FALSE;
	cerr << "WARNING: not all nodes reachable: found "
	     << node_reachable_count << " of " << node_count
	     << "." << endl;
    }
}





//---------------------------------------------------------
// L&T's LINK procedure, but that's not a good name in C++.
#ifdef LT_SIMPLE
// This is the first (unsophisticated) of their two flavors.

void Graph::LT_link( int v, int w )
{
    ltarray[w].ancestor = v;
}
#else
//--------
// This is the second (sophisticated) of their two flavors.

void Graph::LT_link( GraphNode *v, GraphNode *w )
{
    GraphNode *s = w;
    while( w->LT.label->LT.semi < s->LT.child->LT.label->LT.semi ) {
	if( s->LT.size + s->LT.child->LT.child->LT.size
	    >= 2 * s->LT.child->LT.size ) {
	    s->LT.child->LT.ancestor = s;
	    s->LT.child = s->LT.child->LT.child;
	}else{
	    s->LT.child->LT.size = s->LT.size;
	    s = s->LT.ancestor = s->LT.child;
	}
    }
    s->LT.label = w->LT.label;
    v->LT.size += w->LT.size;
    if( v->LT.size < 2 * w->LT.size ) {
	GraphNode *swap = s;		// swap s and child(v)
	s = v->LT.child;
	v->LT.child = swap;
    }
    while( s ) {
	s->LT.ancestor = v;
	s = s->LT.child;
    }
}
#endif
//---------------------------------------------------------
void Graph::compress( int v )
{
    int av = ltarray[v].ancestor;
    if( ltarray[av].ancestor ) {
	compress( av );
	if( ltarray[ltarray[ltarray[v].ancestor].label].semi <
	    ltarray[ltarray[v].label].semi ) {

	    ltarray[v].label = ltarray[ltarray[v].ancestor].label;
	}
	ltarray[v].ancestor = ltarray[ltarray[v].ancestor].ancestor;
    }
}
//---------------------------------------------------------
// L&T's EVAL function, but that's not a good name in C++.
#ifdef LT_SIMPLE
// This is the first (unsophisticated) of their two flavors.

int Graph::LT_eval( int v )
{
    if( ltarray[v].ancestor != 0 ) {
	compress( v );
	v = ltarray[v].label;
    }
    return v;
}
#else
//--------
// This is the second (sophisticated) of their two flavors.

int Graph::LT_eval( int v )
{
    if( ltarray[v].ancestor == 0 ) 
	return ltarray[v].label;
    compress( v );
    if( ltarray[ltarray[ltarray[v].ancestor].label].semi >=
	ltarray[ltarray[v].label].semi )
	return ltarray[v].label;
    return ltarray[ltarray[v].ancestor].label;

    // NOTE, multiple return stmts
}
#endif
//---------------------------------------------------------
// Visit all but the root, in order of descending node#
// Computes postdominator, else dominator. No sennsitivity here to order.

void Graph::step23( bool post )
{
    for( int i = node_reachable_count; i >= 2; i-- ) { 
	int w = vertex[i];

	// foreach v in pred(w) do .. od
	// Except that if post, we use succ(w)..

	GraphNode *pw = ltarray[w].node;
	DLList<GraphEdge*>* list = 
	    post ? &(pw->successors): &(pw->predecessors);
	for( Pix p = (*list).first(); p; (*list).next(p) ) {
	    GraphEdge *pe = (*list)(p);
	    GraphNode *pv = post ? (pe->pto) : (pe->pfrom);
	    int v = pv->ltindex;
	    int u = LT_eval( v );
	    if( ltarray[u].semi < ltarray[w].semi )
		ltarray[w].semi = ltarray[u].semi;
	}

	// add w to bucket(vertex(semi(w)))
	ltarray[vertex[ltarray[w].semi]].bucket.append(w);

	LT_link( ltarray[w].parent, w );

	// foreach v in bucket(parent(w)) do
	// 	delete v from bucket(parent(w))
	int wparent = ltarray[w].parent;

	for( ;!(ltarray[wparent].bucket.empty()); ) {
	    int v = ltarray[wparent].bucket.remove_front();
	    int u = LT_eval( v );

	    if( ltarray[u].semi < ltarray[v].semi )
		ltarray[v].dom = u;
	    else
		ltarray[v].dom = ltarray[w].parent;
	}
    }
}
//---------------------------------------------------------
// This may be called twice (postdominator and dominator) or once (dominator).
// No sensitivity to order.

void Graph::step4( bool post )
{
    for( int i = 2; i <= node_reachable_count; i++ ) {
	int w = vertex[i];
	if( ltarray[w].dom != vertex[ltarray[w].semi] ) {
	    ltarray[w].dom = ltarray[ltarray[w].dom].dom;
	}
    }
    GraphNode *root = post ? graph_exit: graph_root;
    ltarray[root->ltindex].dom = 0;
}
//---------------------------------------------------------
// Traverse the DomNode tree, numbering the left_num and right_num field.
// Postorder (children then parent) means the child's left_num is less
//	than the parent's left_num.
// Preorder (parent then children) means the child's right_num is greater
//	than the parent's right_num.

void Graph::order_dom_tree( DomNode *parent )
{
    static int preorder_seed = 0;	// unlikely to overflow.
    static int postorder_seed = 0;	// unlikely to overflow.

    parent->right_num = preorder_seed++;

    for( Pix p = parent->children.first(); p; parent->children.next(p) ) {
	DomNode *child = parent->children(p);
	order_dom_tree( child );
    }
    parent->left_num = postorder_seed++;
}
//---------------------------------------------------------
void Graph::reinit_tree( bool post )
{
    // Choose either the dominator map or postdominator map
    RSTree<NodeKey,DomNode*> *pmap = post ? (&postdom_map) : (&dom_map);

    DLList<DomNode*> empty;

    // Visit all the DomNode's in that map
    for( Pix pix = pmap->first(); pix; pmap->next(pix)) {
	DomNode *node = pmap->data( pix );

	node->outdegree = 0;
	node->parent = 0;
	node->children = empty;
    }
}
//--------------------------------------------------------- Build the
//dominator tree or postdominator tree, filling in the fields of each
//DomNode reachable through Graph.dom_map or Graph.postdom_map.
//Leaves Graph.dom_root or Graph.postdom_root.  No sensitivity to
//order of dom vs postdom call.
//
// Algorithm:
// The L&T algorithm has left us with a way to visit all the nodes
// (vertex[]) and left us the GraphNode.dom fields. The .dom gives the
// dominator/postdominator parent of each GraphNode.  The DomNodes
// were new'ed previously, so that this routine wouldn't need an extra
// pass. They were filed in the dom_map and the postdom_map: we have
// to choose which map to use.

void Graph::build_dom_tree( bool post )
{
    // Choose either the dominator map or postdominator map
    RSTree<NodeKey,DomNode*> *pmap = post ? (&postdom_map) : (&dom_map);

	// Visit all the Graphnodes (except the graph root) and put
	// the matching tree DomNode into the set of his tree-parent's
	// children.

    for( int i = 2; i <= node_reachable_count; i++ ) {
	int n = vertex[i];
	int dom = ltarray[n].dom;
	GraphNode *graphnode = ltarray[n].node;
	graphnode->dom = ltarray[dom].node;
	NodeKey key = graphnode->key;
	NodeKey parent_key = graphnode->dom->key;

	// We have the keys of a node and his parent.
	// We won't visit this node again, except in that he is
	// 	possibly someone's parent.
	DomNode *domnode = (*pmap)[ key ];
	DomNode *parent_domnode = (*pmap)[ parent_key ];

	domnode->parent = parent_domnode;
	parent_domnode->children.append( domnode );
	parent_domnode->outdegree++;
    }

    // Now find dominator tree's root and leave ptr to it in
    // Graph.dom_root

    GraphNode *root = post ? graph_exit: graph_root;
    NodeKey root_key = root->key;
    DomNode *tree_root = (*pmap)[ root_key ];
    if( post )
	postdom_root = tree_root;
    else
	dom_root = tree_root;

    assert( tree_root->parent == 0 );	// assert root is a root
    assert( tree_root->outdegree > 0 );	// assert tree not singular

    order_dom_tree( tree_root );		// traverse, numbering.
}
//---------------------------------------------------------
// Constructor

Graph::Graph(){ 
    vertex = 0;
    ltarray = 0;
    graph_root = graph_exit = 0; 
    dom_root = postdom_root = 0;
    node_count = 0;
    max_outdegree = 0;
    make_single_exit_node = (bool)FALSE;
    make_single_root_node = (bool)FALSE;
    max_key = 0;				// unsigned type.
				// Synthesize keys via ++max.
}

void Graph::graph_delete() {
    Pix pix;
    for( pix = node_map.first(); pix; node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );
        free(node);
    }

    for(pix = dom_map.first(); pix; dom_map.next(pix)) {
	DomNode *node = dom_map.data( pix );
        free(node);
    }

    for(pix = postdom_map.first(); pix; postdom_map.next(pix)) {
	DomNode *node = postdom_map.data( pix );
        free(node);
    }
    for(pix = edge_list.first(); pix; edge_list.next(pix)) {
	GraphEdge *edge = edge_list( pix );
        free(edge);
    }
    free(graph_root);
    free(graph_exit);
    free(dom_root);
    free(postdom_root);

}

//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::add_new_node( NodeKey key )
{
    Pix pix = node_map.lookup( key );
    if( pix == 0 ) {			// not pre-existing
	GraphNode *node = new GraphNode;
	add_node( node, key );
    }
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::add_node( GraphNode *pn, NodeKey key )
{
    if( max_key < key )
	max_key = key;
    node_count++;
    if( graph_root == 0 ) graph_root = pn;

    Pix p = node_map.lookup( key );
    assert( p == 0 );	// assert key not pre-existing

    node_map.append( key, pn );	// remember the node into an RSTree

    // Create two matching nodes to be in the dominator and postdominator
    // trees. Remember them into RStrees.
    DomNode *pdom     = new DomNode;
    DomNode *ppostdom = new DomNode;
    dom_map.append( key, pdom );
    postdom_map.append( key, ppostdom );

    // Stamp all the nodes with their common key value.
    pn->key = key;
    pdom->key = key;
    ppostdom->key = key;
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::add_nodes( NodeKey key1, NodeKey key2 )
{
    GraphNode *node1;
    GraphNode *node2;

    // Can't look in node_map with operator[], it aborts on not-found.
    Pix pix1 = node_map.lookup( key1 );
    Pix pix2 = node_map.lookup( key2 );

    if( pix1 == 0 ) {			// Not pre-existing,
	node1 = new GraphNode;	// 	so create a node,
	add_node( node1, key1 );	// 	and put in graph.
    }else{					// Already in graph
	node1 = node_map(pix1);		//   so finish off the lookup
    }


    if( pix2 == 0 ) {			// Not pre-existing,
	node2 = new GraphNode;	// 	so create a node,
	add_node( node2, key2 );	// 	and put in graph.
    }else{					// Already in graph
	node2 = node_map(pix2);		//   so finish off the lookup
    }


    GraphEdge *edge =  new GraphEdge;	// create edge
    add_edge( edge, node1, node2 );		//        from 1 to 2.
}

//---------------------------------------------------------
// Visit all the GraphNodes and re-init, so that L&T can be run again.
// Since post-dominator is done first if it's done, we save its results.
// Expects vertex[] can be used to visit the reachable nodes.

void Graph::reinit_nodes()
{

    if( vertex != 0 ) delete [] vertex;
    vertex = new int[ node_count+1 ];	// L&T name
    if( ltarray != 0 ) delete [] ltarray;
    ltarray = new LT[ node_count+1 ];

    int i = 1;
    for( Pix pix = node_map.first(); pix; node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );

	node->post_dom = node->dom; // Postdom computed into wrong one
	node->dom = 0;		    //    so preserve his results.

	ltarray[i].node = node;
	ltarray[i].semi = 0;
    }
}


//---------------------------------------------------------
// If post, are computing the postdominator.
// This will be called either once or twice: if two, post should be first,
//	see comments in number_the_nodes.

void Graph::compute_lengauer( bool post )
{


//fprintf(stderr," number the nodess \n");

    number_the_nodes(post);		// L&T Step 1: depth first traversal

//fprintf(stderr," before step 23 \n");

    step23( post ); // L&T Step 2&3: visit by order of decreasing number

//fprintf(stderr," before step 4 \n");

    step4( post );  // L&T Step 4: visit in incr order, cleaning up dom's

//fprintf(stderr," before build dome tree \n");

    build_dom_tree( post );	// DomNode trees	
//fprintf(stderr," after build dom tree \n");

}


//---------------------------------------------------------
// Child is in the loop body (and has an edge to head).
// Generates into head, the set of nodes that are the loop body.
// Head may have been marked as a loop header during a previous
//	invocation, in the case that multiple back edges to a single node
//	are discovered. The 2..n'th calls to here attempt to add any
//	further nodes to the loop body, but we are prepared to not
//	find any.
// This algorithm searces the precessors for things not in the set,
//	and relies on the loop header dominating the loop (hence all
//	predecessor paths eventually fetch up at the loop header, which
//	is in the set, hence that fragment of the search truncates).

void Graph::find_loop_body( GraphNode *child, GraphNode *head )
{
    DLList<GraphNode*> sack;	// just a holding bag

    if( !head->is_loop_header ) {			// first call
	head->is_loop_header = (bool)TRUE;
	head->loop_depth++;
	head->loop_body_map.append( head->key, head );
    }
    sack.append( child );	// If child already in map, it's ok
				// (a self-loop is an example of that)

    for(;!sack.empty();) {
	// any of them would do
	GraphNode *node = sack.remove_front( );

	Pix pixsack = head->loop_body_map.lookup( node->key );
	// Already in the map,
	// so ignore and try another.
	if( pixsack ) {
	    continue;
	}

	head->loop_body_map.append( node->key, node );
	node->loop_depth++;

	// for every pfrom in the set node->predecessors

	for( Pix p = node->predecessors.first(); p; 
						 node->predecessors.next(p) ) {
	    GraphEdge *pe = node->predecessors(p);
	    GraphNode *pfrom = pe->pfrom;
			
	    // is pfrom already in the loop body map?

	    Pix pixnode  = head->loop_body_map.lookup( pfrom->key);
	    if( pixnode == 0 ) {
		sack.append( pfrom );	// not in the map yet.
	    }
	}
    }
    // note, is a return
}
//---------------------------------------------------------
//
// Marks all GraphNodes as to whether they are a loop header.
// For each loop header, finds the loop body.
// We follow the definition used by the .loop_header method: the target
//	of a back edge.
// So, rely on all nodes having been initted FALSE, and scan the edge_list.
// This must be called after the rpostorder field is set, ie after at least
//	the first portion of the L&T algorithm.

void Graph::find_loops()
{
    for( Pix p = edge_list.first(); p; edge_list.next(p) ) {
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	if( to->rpostorder <= from->rpostorder) {
	    to->is_loop_header = (bool)TRUE;
	}

	if( to->lpostorder <= from->lpostorder) {
	    to->is_loop_header = (bool)TRUE;
	}

    }
}

//---------------------------------------------------------
// We know that his indegree > 1.
// Create a new node, with edges out being a duplicate of old's,
//	and with a synthesized key (since duplicate keys cause asserts).
// Partition the in-edges between the old and new nodes.
// Appropriate adjustments to the node and edge lists and counts.

void Graph::split_this( GraphNode *old )
{
    ++max_key;

    GraphNode *newnode = new GraphNode;	// Create a new node and
    add_node( newnode, max_key );		// add it to the graph.
    newnode->split_from = old->key;		// Record where he came from.

    // traverse old's successor set, making new have an equivalent
    // set of successors.

    for( Pix p = old->successors.first(); p; old->successors.next(p)){
	GraphEdge *old_edge = old->successors(p);

	GraphEdge *new_edge =  new GraphEdge;
	add_edge( new_edge, newnode, old_edge->pto );
	newnode->successors.append( new_edge );

	old_edge->pto->predecessors.append( new_edge );
	old_edge->pto->indegree++;
    }
    newnode->outdegree = old->outdegree;

    // Now make half of the old's predecessors point to new
    // (and adjust the indegree's and predecessor sets to agree)

    for(int adjust_count = old->indegree/2; old->indegree > adjust_count;){
	GraphEdge *edge = old->predecessors.remove_front();
	newnode->predecessors.append( edge );
	edge->pto = newnode;			// swing the pointer
	newnode->indegree++;
	old->indegree--;
    }
}
//---------------------------------------------------------
// Checks for irreducibility.
// Cannot be run until after a dominator tree has been built
//	(so that ::dominates() works) and after the rpostorder is
//	available (so that back edge detection works).
// Returns true to mean that splits occurred.
// Algorithm:
// 	Does a pass finding victims, then a pass thru a victim map,
//	splitting those victims.
//	This avoids invalidating the dominator tree (with the early spits)
//	which would make subsequent identifications of victims iffy.
//	Use a victim map, not a list, so can cope with multiple back edges
//	 to a single node causing multiple discovery of same victim.

bool Graph::split_nodes()
{
    bool did_stuff = (bool)FALSE;
    RSTree<NodeKey,GraphNode*> victim_map;

    // Loop through the set of edges

    for( Pix p = edge_list.first(); p; edge_list.next(p) ) {
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	// Check for this edge being a back edge, using the
	// same formula as method .back_edge.
	// Also check dominance. Reducible graphs they're the same.

	if( (to->rpostorder <= from->rpostorder) &&  // back edge
	    (!dominates( to->key, from->key ))) {

	    if( to->indegree == 1 ) {
		cerr << "ERROR: Cant split node " << to->key
		     << " since indegree == 1."	  << endl;
		exit(1);
	    }
	    Pix look = victim_map.lookup( to->key );
	    if( look ) continue; 	// already in map

	    victim_map.append( to->key, to );
	    did_stuff = (bool)TRUE;
	}
    }

    for( Pix pix = victim_map.first(); pix; victim_map.next(pix)) {
	GraphNode *to = victim_map.data( pix );
	split_this( to );
    }
    return did_stuff;
}


void Graph::add_to_nat_loop(GraphNode * loop_head, GraphNode * node) {
    int not_in_nat_loop = TRUE;

    for( Pix p = loop_head->nat_loop.first(); p; loop_head->nat_loop.next(p)){  
	GraphNode * n = loop_head->nat_loop(p);
	if (n == node) {
	    not_in_nat_loop = FALSE;      
	}
    }

    if (not_in_nat_loop) {
	loop_head->nat_loop.append(node);
    }
}

int Graph::make_nat_loop(GraphNode * loop_head, GraphNode * node) {
    int in_loop = FALSE;

//fprintf(stderr,"processing node -> %d \n", node->key);
    if (node->visited) {
	return(node->in_loop);
    }

    node->visited = TRUE;

    for( Pix p = node->successors.first(); p; node->successors.next(p)) {

	GraphEdge *edge = node->successors(p);
	assert(edge->pfrom == node);
	if ((edge->pto == loop_head)
	    & ((edge->pto->rpostorder <= node->rpostorder)
	       | (edge->pto->lpostorder <= node->lpostorder)))  {
	    in_loop = TRUE;
	} else
	    if (make_nat_loop(loop_head, edge->pto)) {
		in_loop = TRUE;
	    }
    }

    node->in_loop = in_loop;

    if (in_loop) {
	add_to_nat_loop(loop_head, node);
    }

    return(in_loop);
}


void Graph::compute_nat_loops() {
    for( Pix pix = node_map.first(); pix; node_map.next(pix)) {
 	GraphNode *node = node_map.data( pix );

        if (node->is_loop_header) {
	    for( Pix p2 = node_map.first(); p2; node_map.next(p2)) {
		GraphNode *n = node_map.data( p2 );
		n->visited = FALSE;
		n->in_loop = FALSE;
	    }
	    node->visited = TRUE;
	    add_to_nat_loop(node, node);
	    for(Pix p = node->successors.first();p; node->successors.next(p)){
		GraphEdge *edge = node->successors(p);
                make_nat_loop(node,edge->pto);
	    }
        }
    }
}

int Graph::loop_exit( NodeKey from_key, NodeKey to_key ) {

    int found_in_nat_loop = FALSE;

    GraphNode *from_node = node_map[ from_key ];	// abort if not found 
    GraphNode *to_node   = node_map[ to_key ];

    for( Pix pix = node_map.first(); pix; node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );
        if (node->is_loop_header) {
	    for( Pix from_p = node->nat_loop.first();
		from_p; node->nat_loop.next(from_p)){  

		GraphNode * from_n = node->nat_loop(from_p);
		if (from_n == from_node) {
		    found_in_nat_loop = FALSE;

		    for ( Pix to_p = node->nat_loop.first();
			 to_p; node->nat_loop.next(to_p)){  
			GraphNode * to_n = node->nat_loop(to_p);
			if (to_n == to_node) {
			    found_in_nat_loop = TRUE;      
			}
		    }
		    if (!(found_in_nat_loop)) {
			return(TRUE);
		    }
		}
	    }
	}
    }
    return(FALSE);
}



//---------------------------------------------------------
// Visible method: see .h for user docn.
//
// NOTE: both compute_lengauer and reinit_nodes expect postdom is 
//	computed before dom. If node splitting turns out to be
//	highly common, this could be fixed, at a mild cost to the
//	non-splitting case.

void Graph::compute_dominators()
{
    //  Traverse the edge set, find outdegrees, predecessors, etc
    propagate_edges_to_nodes();

    if( make_single_exit_node )
	really_add_single_exit_node();
    if( make_single_root_node )
	really_add_single_root_node( pending_base_root );
    exit_indegree = graph_exit ? (graph_exit->indegree) :0;
    if( exit_indegree == 0 ) 
	cout << "WARNING: exit node has indegree 0." << endl;
    root_outdegree = graph_root->outdegree;
    if( root_outdegree == 0 )
	cout << "WARNING: root node has outdegree 0." << endl;

    compute_lengauer( (bool)TRUE );	// find post-dominators
    reinit_nodes();
	
    compute_lengauer( (bool)FALSE );	// find dominators
	  
    reverse_number_the_nodes();		// L&T Step 1: depth first traversal
    find_loops();
    compute_nat_loops();

}

//---------------------------------------------------------
// Visible method: see .h for user docn.

bool Graph::immed_dominates( NodeKey key_parent, NodeKey key_child )
{
    GraphNode *node_parent = node_map[ key_parent ];
    GraphNode *node_child  = node_map[ key_child ];
    return (bool) (node_parent == node_child->dom);
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

bool Graph::dominates( NodeKey key_parent, NodeKey key_child )
{
    DomNode *node_parent = dom_map[ key_parent ];
    DomNode *node_child  = dom_map[ key_child ];

    // We now have two places in the dominator tree, ie the
    // same tree. The predicate now is:
    //	"true if child is in parent's subtree"
    // which is exactly the predicate that the left_num and
    // right_num values can answer.

    return (bool)
	(( node_child->left_num  < node_parent->left_num ) &&
	 ( node_child->right_num > node_parent->right_num ));
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

bool Graph::immed_post_dominates( NodeKey key_parent, NodeKey key_child )
{
    DomNode *node_parent = postdom_map[ key_parent ];
    DomNode *node_child  = postdom_map[ key_child ];
    return (bool) (node_parent == node_child->parent);
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

bool Graph::post_dominates(NodeKey key_parent, NodeKey key_child)
{
    DomNode *node_parent = postdom_map[ key_parent ];
    DomNode *node_child  = postdom_map[ key_child ];

    // We now have two places in the postdominator tree, ie the
    // same tree. The predicate now is:
    //	"true if child is in parent's subtree"
    // which is exactly the predicate that the left_num and
    // right_num values can answer.

    return (bool)
	(( node_child->left_num  < node_parent->left_num ) &&
	 ( node_child->right_num > node_parent->right_num ));
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

bool Graph::back_edge( NodeKey key_from, NodeKey key_to )
{
    GraphNode *node_from = node_map[ key_from ];
    GraphNode *node_to   = node_map[ key_to ];

    // Zima lemma 3.2 (2) p.68

    return (bool) (node_to->rpostorder <= node_from->rpostorder);
}
//---------------------------------------------------------
// Visible method: see .h for user docn.
//
// compute_dominators() leaves a bool in each GraphNode, so, we
//	just return that value.

bool Graph::loop_header( NodeKey key )
{
    GraphNode *graphnode = node_map[ key ];
    return graphnode-> is_loop_header;
}


//	// Loop through the set of edges coming to graphnode
//
//	for( Pix p = graphnode->predecessors.first(); p; 
//		graphnode->predecessors.next(p) ) {
//		GraphEdge *pe = graphnode->predecessors(p);
//		GraphNode *from = pe->pfrom;
//
//		// from is a graphnode with an edge to me.
//
//		if( dominates( key, from->key )) return (bool)TRUE;
//		// We dominate someone who has an arc to us. We're a loop.
//	}
//
//	return (bool)FALSE; // we didn't dominate any predecessors.
//}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::make_root_key( NodeKey key )
{
    GraphNode *graphnode = node_map[ key ];	// abort if not found
    make_root( graphnode );
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::make_exit( GraphNode *pn )
{
    graph_exit = pn;
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::make_exit_key( NodeKey key )
{
    GraphNode *graphnode = node_map[ key ];	// abort if not found
    make_exit( graphnode );
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::add_single_exit_node()
{
    make_single_exit_node = (bool)TRUE;
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::add_single_root_node( NodeKey base_root )
{
    pending_base_root = base_root;
    make_single_root_node = (bool)TRUE;
}
//---------------------------------------------------------
void Graph::really_add_single_exit_node()
{
    ++max_key;				// synthesize a unique key
    NodeKey key = max_key;

    GraphNode *new_exit = new GraphNode;	// create this node
    add_node( new_exit, key );		// and put it in the graph
	

    // Visit all nodes by traversing the node_map
    for( Pix pix = node_map.first(); pix; node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );

	if( node == new_exit ) continue;	// skip myself

	if( node->outdegree == 0 ) {   // Create edge from here to exit

	    cerr << "found outdegree 0\n";

	    GraphEdge *edge =  new GraphEdge;
	    add_edge( edge, node, new_exit );

	    node->successors.append( edge );
	    node->outdegree = 1;
	    new_exit->predecessors.append( edge );
	    new_exit->indegree++;
	}
    } 

    if (new_exit->indegree == 0) {
	cerr << "WARNING: Can't find any nodes with outdegree 0";
	cerr << " while creating EXIT node\n";
	Pix basepix = node_map.lookup( max_key-1 );
	GraphNode *node = node_map.data( basepix );
	GraphEdge *edge =  new GraphEdge;
	add_edge( edge, node, new_exit );
	node->successors.append( edge );
	node->outdegree ++;
	new_exit->predecessors.append( edge );
	new_exit->indegree++;
    }
    make_exit( new_exit );		// do last so indegree is known
}
//---------------------------------------------------------
// The base_root (if it exists ) is to be a successor of the artificial root.

void Graph::really_add_single_root_node( NodeKey base_root_key )
{
    ++max_key;				// synthesize a unique key
    NodeKey key = max_key;

    GraphNode *new_root = new GraphNode;	// create this node
    add_node( new_root, key );		// and put it in the graph

    Pix basepix = node_map.lookup( base_root_key );
    if( basepix ) {
	// The base root node does exist. Complete the lookup
	GraphNode *base_node = node_map.data( basepix );
		
	// If his indegree is zero, just let the loop below find him
	if( base_node->indegree ) {

	    GraphEdge *edge =  new GraphEdge;
	    add_edge( edge, new_root, base_node );

	    base_node->predecessors.append( edge );
	    base_node->indegree += 1;
	    new_root->successors.append( edge );
	    new_root->outdegree++;
	}
    }
	

    // Visit all nodes by traversing the node_map
    for( Pix pix = node_map.first(); pix; node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );
	if( node == new_root ) continue;	// skip myself

	if( node->indegree == 0 ) {   // Create edge from root to here
	    GraphEdge *edge =  new GraphEdge;
	    add_edge( edge, new_root, node );

	    node->predecessors.append( edge );
	    node->indegree = 1;
	    new_root->successors.append( edge );
	    new_root->outdegree++;
	}
    }
    make_root( new_root );		// do last so indegree is known
}
//---------------------------------------------------------
// Visible method: see .h for user docn.

void Graph::show_loop_headers(ostream & out)
{
    out << "Loop header nodes: ";
    for( Pix p = node_map.first(); p; node_map.next(p)){
	GraphNode *node = node_map(p);
	if(node->is_loop_header) {		// this key is a hdr
	    out << node->key << " ";
	}
    }
    out << endl;
}
//-------------------------------------------------------------
// Expects node i is a loop header

void Graph::show_loop_body(ostream & out, GraphNode *node )
{
    out << "Loop body of " << node->key << " is ";

    for( Pix pix = node->nat_loop.last(); pix; node->nat_loop.prev(pix)) {
	GraphNode *node = node->nat_loop( pix );
	out << node->key << " ";
    }
    out << endl;
}

//-------------------------------------------------------------
void Graph::show_loop_bodies(ostream & out)
{

    for( Pix p = node_map.first(); p; node_map.next(p)){
	GraphNode *node = node_map(p);
	if(node->is_loop_header) {		// this key is a hdr
	    show_loop_body(out, node);
	}
    }

}

void Graph::show_nodes(ostream & out)
{
    out << "Nodes:" << endl;
    for( Pix pix = node_map.first(); pix; 
				     node_map.next(pix)) {
	GraphNode *node = node_map.data( pix );

	out 	<< "\t" << node->key
		<< " r=" << node->rpostorder
		<< " l=" << node->lpostorder
		<< " in " << node->indegree
		<< " (";
	for( Pix p=node->predecessors.first();p;
					      node->predecessors.next(p)){
	    GraphEdge *pe = node->predecessors(p);
	    out << pe->pfrom->key << "," << pe->pto->key << " ";
	}
	out << ") out " << node->outdegree << " (";
	for( p = node->successors.first();p;node->successors.next(p)){
	    GraphEdge *pe = node->successors(p);
	    out << pe->pfrom->key << "," << pe->pto->key << " ";
	}
	out << ")" << endl;
    }
}

void Graph::show_loop_exits(ostream & out)
{
    out << "Loop exit edges: ";

    // Loop through the edge list

    for( Pix p2 = edge_list.first(); p2; edge_list.next(p2)){
	GraphEdge *pe = edge_list(p2);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	// print the key pair if the edge is a loop exit

	if( loop_exit( from->key, to->key )) {
	    out << from->key << "," << to->key << " ";
	}
    }
    out << endl;
}
//-------------------------------------------------------------

void Graph::show_back_edge(ostream & out)
{
    out << "Back edges: ";

    // Traverse the Graph's edge list

    for( Pix p = edge_list.first(); p; edge_list.next(p)){
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	// print the key pair if this edge is a back edge

	if(back_edge( from->key, to->key )) {
	    out << from->key << "," << to->key << " ";
	}
    }
    out << endl;
}


void Graph::show_dominance(ostream & out)
{
    out << "Non dominating edges: ";

    // Traverse the Graph's edge list

    for( Pix p = edge_list.first(); p; edge_list.next(p)){
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	// print the key pair if "from" doesn't dominate "to"

	if(!dominates( from->key, to->key )) {
	    out << from->key << "," << to->key << " ";
	}
    }
    out << endl;

    out << "Dominating edges: ";

    // Traverse the Graph's edge list

    for( Pix p2 = edge_list.first(); p2; edge_list.next(p2)){
	GraphEdge *pe = edge_list(p2);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	// print the key pair if "from" doesn't dominate "to"

	if(dominates( from->key, to->key )) {
	    out << from->key << "," << to->key << " ";
	}
    }
    out << endl;
}

//-------------------------------------------------------------
void Graph::show_post_dominance(ostream & out)
{
    int num_printed = 0;
    out << " Non post dominating edges: ";

    // Traverse the Graph's edge list

    for( Pix p = edge_list.first(); p; edge_list.next(p)){
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;


	if( !post_dominates( to->key, from->key )) {
	    out << from->key << "," << to->key << " ";
	    num_printed++;
	}
    }

    out << endl << "Post dominating edges: ";
    for( p = edge_list.first(); p; edge_list.next(p)){
	GraphEdge *pe = edge_list(p);
	GraphNode *from = pe->pfrom;
	GraphNode *to   = pe->pto;

	if(post_dominates( to->key, from->key )) {
	    out << from->key << "," << to->key << " ";
	    num_printed++;
	}
    }
    out << endl;
}



void Graph::show_all(ostream & out, char * name) {
    out << "-------------------------------------------------\n";
    out << "Proc name: " << name << endl;
    show_loop_headers(out);
    show_loop_bodies(out);
    show_back_edge(out);
    show_loop_exits(out);
    show_nodes(out);

    if (!dom_root ) {
	out << "WARNING: No domintator tree." << endl;
    }

    if (!postdom_root ) {
	out << "WARNING: No Postdominator tree." << endl;
    }

    show_dominance(out);
    show_post_dominance(out);

    out << "-------------------------------------------------\n";
  
}

/*

  bool Graph::loop_exit( NodeKey from_key, NodeKey to_key )
  {
	GraphNode *from_node = node_map[ from_key ];	// abort if not found
	GraphNode *to_node   = node_map[ to_key ];

        if( from_node->loop_depth == 0 ) 
		return (bool)FALSE;	// not in a loop to begin with

        if( from_node->loop_depth > to_node->loop_depth )
                return (bool)TRUE;    // crossed out of a loop body

        if( from_node->loop_depth < to_node->loop_depth )
                return (bool)FALSE;   // went into an inferior loop

	// Depths are equal. 

        if( !to_node->is_loop_header ) 
		return (bool)FALSE;	// An exit to a non-header would
					// have made the depth go down, and
					// it didn't go down, so not an exit

	// Either we jumped out of this loop to someone else's loop header,
	// or, we jumped to our own loop header.
	// (Which could be ourself, ie it's a self-loop).

	// Search up the dominance chain for our loop header
	// 	(it isn't necessarily the first loop header we find)
	// 	It can be told by having from_key in its body set.

	GraphNode *header = from_node;	// Might be already at that header
	for(;;) {
		if( header->is_loop_header ) {
			Pix pix  = header->loop_body_map.lookup(from_key);
			if( pix ) break;	// he's there!
		}
		header = header->dom;		// immediate dominator
		assert( header != 0 );		// off the top
	}
	// "header" now is the closest loop header containing "from".
	// Look up "to" in his set of loop body members.
	// 0 means not-found: return TRUE in that case, else FALSE.

	Pix pixnode  = header->loop_body_map.lookup( to_key);
	return (bool)(pixnode==0);
}


*/
