//-*-c++-*-
#ifndef _CUT_FlexiCache_h_
#define _CUT_FlexiCache_h_

#include <iostream.h>
#include <stdio.h>
#include <values.h>

//
// A FlexiCache is similar to the AssocCache class, but
// all parameters are specified when the instance is created,
// not at compile time as in the AssocCache.
//

template <class address>
class CUT_FlexiCache {
public:
  int associativity;

  int lgsets;
  int total_sets;
  unsigned long lgsets_mask;

  int total_blocks;

  int lg_block_size;
  int block_size;

  //
  // Used for LRU replacement aging. You shouldn't
  // use this to determine the total number of references
  // to the cache.
  //
  unsigned long age_references;
  
  //
  // Each cache entry records the following
  //
  struct PerEntry {
    unsigned long age;
    address tags;
    int dirty;
  };

  struct PerSet {
    long stores;
    long loads;
    long store_misses;
    long load_misses;
    long write_backs;
    //
    // This is allocated and initialized when
    // the cache is created.
    //
    PerEntry *entries;
  };

  PerSet *lines;
  PerEntry *entries;

public:

  //
  // Note that the associativity is NOT a logartihm - 
  // this allows you to specify e.g., 3-way associative
  // organizations.

  CUT_FlexiCache(int assoc_,
		 int lg_sets_,
		 int lg_block_size_ = 5) {

    associativity = assoc_;

    lg_block_size = lg_block_size_;
    block_size = (1 << lg_block_size);

    lgsets = lg_sets_;
    total_sets = (1 << lgsets);
    lgsets_mask = ~(-1 << lgsets);

    total_blocks = total_sets * associativity;
      
    lines = new PerSet[ total_sets ];
    entries = new PerEntry[ total_blocks ];

    memset(lines, 0, sizeof(PerSet) * total_sets);
    memset(entries, 0, sizeof(PerEntry) * total_blocks);

    age_references = 0;

    //
    // Make each line point to it's part of the
    // entries array;
    //
    int s = 0;
    for (int i = 0 ; i < total_sets; i++ ) {
      lines[i].entries = &(entries[s]);
      s += associativity;
    }
  }

  virtual ~CUT_FlexiCache() {
    delete entries;
    delete lines;
  }
  
  unsigned int line(address addr) {
    int l = (addr >> (lg_block_size));
    l = l & lgsets_mask;
    return l;
  }
  
  address tag(address addr) {
    address t = (addr >> lg_block_size);
    return t;
  }

  void line_and_tag(int& l, address& t, address addr) {
    t = (addr >> lg_block_size);
    l = t & lgsets_mask;
  }
  
  int is_miss(address addr) {
    int l;
    address t;

    line_and_tag (l, t, addr);

    int found = 0;
    for (int a = 0; a < associativity; a++ ) {
      if ( lines[l].entries[a].age & lines[l].entries[a].tags == t ) {
	return 0;
      }
    }
    return ( 1 );
  };
  
  int is_hit(address addr) {
    return ( ! is_miss(addr) );
  }
  
  //
  // ref_if_hit updates the age on the reference if there's a hit
  // and returns -1 if there's a hit. It returns an INDEX if there's
  // a miss and the line needs to be filled. The INDEX indicates
  // the set that should be replaced.
  //
  int ref_if_hit(address addr, PerSet& line,
		 int l, address t, int make_dirty) {
    //
    // Reference a given cache line, updating the age
    //
    age_references++;

    int oldest = -1;
    unsigned long oldest_age = MAXLONG;

    for (int a = 0; a < associativity; a++ ) {
      PerEntry& entry = line.entries[a];
      if ( entry.age ) {
	//
	// Hit?
	//
	if ( entry.tags == t ) {
	  entry.age = age_references;
	  entry.dirty = entry.dirty | make_dirty;
	  return -1;
	}

	if (entry.age < oldest_age ) {
	  oldest = a;
	  oldest_age = entry.age;
	}
      } else {
	//
	// If this line hasn't been used, we'll
	// fill it before anything else, but
	// we need to keep looking for other
	// references
	//
	oldest = a;
	oldest_age = 0;
      }
    }

    return oldest;
  }
  
  //
  // Put something in the cache
  //
  void fill(PerSet& line, int oldest, address t, int make_dirty) {
    PerEntry& entry = line.entries[oldest];

    if (entry.age & entry.dirty) {
      line.write_backs++;
    }
    entry.tags = t;
    entry.dirty = make_dirty;
    entry.age = age_references;
  }
  
  //
  // Return 1 if it's was a miss, 0 otherwise
  int load(address addr) {
    int l;
    address t;

    line_and_tag(l, t, addr);
    PerSet& line = lines[l];

    line.loads++;
    int index = ref_if_hit( addr, line, l, t, 0 );
    if ( index >= 0 ) {
      line.load_misses ++;
      fill(line, index, t, 0);
      return 1;
    } else {
      return 0;
    }
  }

  //
  // This makes 'references' references to the SAME cache line
  // located at address 'addr'. This is mainly used for faster
  // instruction cache simulation.
  //
  int load_same_line(address addr, int refs_this_line) {
    int l;
    address t;
    line_and_tag(l, t, addr);
    PerSet& line = lines[l];

    line.loads += refs_this_line;

    int index = ref_if_hit( addr, line, l, t, 0 );
    if ( index >= 0 ) {
      line.load_misses ++;
      fill(line, index, t, 0);
      return 1;
    } else {
      return 0;
    }
  }

  int store_allocate(address addr) {
    int l;
    address t;
    line_and_tag(l, t, addr);
    PerSet& line = lines[l];

    line.stores++;

    int index = ref_if_hit( addr, line, l, t, 1 );
    if ( index >= 0 ) {
      line.store_misses++;
      fill(line, index, t, 1);
      return 1;
    }  else {
      return 0;
    }
  }

  int store_no_allocate(address addr) {
    int l;
    address t;
    line_and_tag(l, t, addr);
    PerSet& line = lines[l];

    line.stores++;

    int index = ref_if_hit( addr, line, l, t, 1 );
    if ( index >= 0 ) {
      line.store_misses++;
      return 1;
    } else {
      return 0;
    }
  }

  int flush_dirty() {
    int write_backs = 0;
    for (int l = 0 ; l < total_sets; l++ ) {
      for (int a = 0 ; a < associativity; a++ ) {
	if ( lines[l].dirty[a] ) {
	  lines[l].write_backs++;
	  write_backs++;
	  lines[l].entries[a].dirty = 0;
	}
      }
    }
    return write_backs;
  }

  //
  // Return individual statistics
  //
  long stores() {
    long foo = 0;
    for (long i=0; i < total_sets; i++) {
      foo += lines[i].stores;
    }
    return foo;
  }

  long loads() {
    long foo = 0;
    for (long i=0; i < total_sets; i++) {
      foo += lines[i].loads;
    }
    return foo;
  }
  long store_misses() {
    long foo = 0;
    for (long i=0; i < total_sets; i++) {
      foo += lines[i].store_misses;
    }
    return foo;
  }
  long load_misses() {
    long foo = 0;
    for (long i=0; i < total_sets; i++) {
      foo += lines[i].load_misses;
    }
    return foo;
  }

  long write_backs() {
    long foo = 0;
    for (long i=0; i < total_sets; i++) {
      foo += lines[i].write_backs;
    }
    return foo;
  }
  //
  // Statistics reporting
  //


  void fini(ostream& output) {
    output << "-----------------------------------";
    output << "-----------------------------------\n";
    output << "  ";
    output << associativity << "-way Assoc Cache, ";
    output << (1 << lg_block_size) << " byte lines, ";
    output << (1 << lglines) << " lines, ";
    int size =(1 << (lg_block_size + lglines));
    size /= 1024;
    output << size << "K bytes\n";
    output << "-----------------------------------";
    output << "-----------------------------------\n";


    int totals = loads() + stores();
    int total_misses = load_misses() + store_misses();

    output << "Load references  = " << loads() << "\n";
    output << "Load misses      = " << load_misses() << "\n";
    double load_miss_rate = load_misses();
    if ( loads() > 0 ) {
      load_miss_rate /= loads();
    } else {
      load_miss_rate = 0.0;
    }
    load_miss_rate *= 100.0;
    output << "Load miss rate   = " << load_miss_rate << "%\n";
    output << "\n";

    output << "Store references = " << stores() << "\n";
    output << "Store misses     = " << store_misses() << "\n";
    double store_miss_rate = store_misses();
    if ( stores() > 0 ) {
      store_miss_rate /= stores();
    } else {
      store_miss_rate = 0.0;
    }
    store_miss_rate *= 100.0;
    output << "Store miss rate  = " << store_miss_rate << "%\n";
    output << "\n";

    output << "Total references = " << totals << "\n";
    output << "Total misses     = " << total_misses << "\n";
    double total_miss_rate = total_misses;
    if ( totals > 0 ) {
      total_miss_rate /= totals;
    } else {
      total_miss_rate = 0.0;
    }
    total_miss_rate *= 100.0;
    output << "Total miss rate  = " << total_miss_rate << "%\n";
    output << "\n";

    output << "Write backs      = " << write_backs() << "\n";
  }

  void put_header(char *prefix, ostream& output) {
    output << prefix << '_';
  }


  void digestable_report(ostream& output, char *prefix) {
    int size = block_size * total_sets * associativity;
    put_header(prefix, output);
    output << "block_size -> " << block_size << "\n";
    put_header(prefix, output);
    output << "lines -> " << total_sets << "\n";
    put_header(prefix, output);
    output << "blocks -> " << total_blocks << "\n";
    put_header(prefix, output);
    output << "assoc -> " << associativity << "\n";
    put_header(prefix, output);
    output << "size -> " << size << "\n";

    unsigned long totals = loads() + stores();
    unsigned long total_misses = load_misses() + store_misses();

    put_header(prefix, output);
    output << "load_references  -> " << loads() << "\n";

    put_header(prefix, output);
    output << "load_misses -> " << load_misses() << "\n";
    double load_miss_rate = load_misses();
    if ( loads() > 0 ) {
      load_miss_rate /= loads();
    } else {
      load_miss_rate = 0.0;
    }
    load_miss_rate *= 100.0;

    put_header(prefix, output);
    output << "load_miss_rate -> " << load_miss_rate << "\n";

    put_header(prefix, output);
    output << "store_references -> " << stores() << "\n";
    put_header(prefix, output);
    output << "store_misses     -> " << store_misses() << "\n";
    double store_miss_rate = store_misses();
    if ( stores() > 0 ) {
      store_miss_rate /= stores();
    } else {
      store_miss_rate = 0.0;
    }
    store_miss_rate *= 100.0;
    put_header(prefix, output);
    output << "store_miss_rate  -> " << store_miss_rate << "\n";

    put_header(prefix, output);
    output << "total_references -> " << totals << "\n";
    put_header(prefix, output);
    output << "total_misses     -> " << total_misses << "\n";
    double total_miss_rate = total_misses;
    if ( totals > 0 ) {
      total_miss_rate /= totals;
    } else {
      total_miss_rate = 0.0;
    }
    total_miss_rate *= 100.0;
    put_header(prefix, output);
    output << "total_miss_rate  -> " << total_miss_rate << "\n";

    put_header(prefix, output);
    output << "write_backs      -> " << write_backs() << "\n";
  }

};

#endif
