//-*-c++-*-
#ifndef _CUT_TLBFixed_h_
#define _CUT_TLBFixed_h_

//
// Simulate an entries-entry TLB using LRU replacement
//

#include "functional"
#include "CUT_MTFV.h"


template <
  int entries_per_bank,
  int lg_page_size,
  int lg_num_banks>
class CUT_TLBFixed {
  

  //
  // The following describes a single "bank" of the TLB.
  // For example, a 64-entry, 4-way associative TLB would
  // be configured as 16 "banks" of four entries each.
  //
  class bank {
  public:
    bank() : mtfv(0) {
      refs = 0;
      misses = 0;
      evictions = 0;
    }

    typedef equal_to<unsigned long> mtfv_uleqtype;

    CUT_MTFV<entries_per_bank, unsigned long, mtfv_uleqtype> mtfv;
    unsigned long refs;
    unsigned long misses;
    unsigned long evictions;
  };
  
  //
  // Each bank contains ENTRIES/BANKS entries
  //
  bank tlb[1 << lg_num_banks];
public:

 
  struct decoded {
    int bank_index;
    unsigned long page_offset;
  };

public:
  CUT_TLBFixed() {
    //
    // Empty initialier..
    //
  }

  ~CUT_TLBFixed() {
    //
    // Nothing to do - all static storage
    //
  }

  void map_addr(unsigned long addr, decoded& x) {
    unsigned long page = addr >> lg_page_size;

    x.bank_index = ((1 << lg_num_banks) - 1) & page;
    x.page_offset = page;
  }

  //
  // Takes a page reference
  //
  //
  // returns TRUE if this is a miss, FALSE if it's a hit
  //

  bool check_for_hit(decoded&x ) {
    bank& b = tlb[x.bank_index];
    b.refs++;

    if ( b.mtfv.find_and_move_to_front(x.page_offset) ) {
      //
      // It's a hit - no need for replacement..
      //
      return false;
    }
    b.misses++;
    return true;
  }

  //
  // Update the TLB to include the reference for this page.
  // We assume that the page is not already in the TLB, but
  // we do not check for that condition. You should only
  // enter a page if you got a TRUE from the check_for_hit routine.
  //
  void update_tlb(decoded& x) {
    bank& b = tlb[x.bank_index];
    b.mtfv.lru_replacement(x.page_offset);
    b.evictions++;
  }

  void digestable_report(ostream& out, char* prefix) {

    unsigned long total_refs = 0;
    unsigned long total_misses = 0;
    unsigned long total_evictions = 0;

    const int max_banks = (1 << lg_num_banks);

    for (int i=0; i < max_banks; i++) {
      unsigned long refs = tlb[i].refs;
      unsigned long misses = tlb[i].misses;
      unsigned long evictions = tlb[i].evictions;

      total_refs += refs;
      total_misses += misses;
      total_evictions += evictions;

      out << prefix << "_bank" << i << "_refs -> " << tlb[i].refs << "\n";
      out << prefix << "_bank" << i << "_misses -> " << tlb[i].misses << "\n";
      out << prefix << "_bank" << i << "_evictions -> " << tlb[i].evictions << "\n";
    }

    out << prefix << "_total_refs -> " << total_refs << "\n";
    out << prefix << "_total_misses -> " << total_misses << "\n";
    out << prefix << "_total_evictions -> " << total_evictions << "\n";
  }
};
#endif

