//-----------------------------------------------------------------------------
// GlobalPtr.C
// -----------
// This file contains an example of how to use global pointers to share data
// between contexts.
//-----------------------------------------------------------------------------


#include <assert.h>
#include <iostream.h>
#include <hpcxx_rts.h>
#include <HPCxx_Sync.cc>
#include <hpcxx_invoke.cc>
#include <hpcxx_register.cc>
#include <HPCxx_GlobalPtr.cc>
#include <hpcxx_invokeImpl.cc>
#include <HPCxx_GlobalRefImpl.cc>


// declare function prototypes;

static void printArray(int array[], int size, int context);
static void sendGlobalPointer(HPCxx_GlobalPtr<int> gp, int context);
static HPCxx_GlobalPtr<int> recvGlobalPointer(int context);

// remote function identifiers

//static int bind_ID;
static int resolve_ID;

class NameServiceLite {
public:

  //---------------------------------------------------------------------------
  // Registers the remote functions this object will serve.
  //---------------------------------------------------------------------------
  static void registerMemberFunctions(void)
  {
//    bind_ID    = hpcxx_register(NameServiceLite::bind, 10);
    resolve_ID = hpcxx_register(NameServiceLite::resolve, 11);
  }

  //---------------------------------------------------------------------------
  // Binds the given name with the object.  Returns 0.
  //---------------------------------------------------------------------------
  static int bind(const char* name, HPCxx_GlobalPtr<int> gp)
  {
    name_ = (char*) realloc(name_, strlen(name) + 1);
    strcpy(name_, name);
    gp_ = gp;			// copy the global pointer
    return 0;
  }
  
  //---------------------------------------------------------------------------
  // Returns the object bound to the given name.
  //---------------------------------------------------------------------------
  static HPCxx_GlobalPtr<int> resolve(const char* name)
  {
    assert( strcmp(name, name_) == 0 );	// just make sure names are the same
    return gp_;
  }
  
private:

  static HPCxx_GlobalPtr<int> gp_;	// copy of global pointer
  static char* name_;			// name bound to global pointer

};

HPCxx_GlobalPtr<int> NameServiceLite::gp_;
char* NameServiceLite::name_ = NULL;


int main(int argc, char* argv[])
{
  int array[10];
  int error = 0;
  const char* nameForGP = "integer array of 10";

  //
  // initialize HPC++
  //
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);

  //
  // all processes are executing in SPMD mode at this point
  //

  int nc = group->getNumContexts() - 1;
  int context = group->getLocalIndex();

  //
  // Initially gp is local to all contexts.
  //
  HPCxx_GlobalPtr<int> gp(array);

  //
  // register naming service functions so they can be exported
  //
  NameServiceLite::registerMemberFunctions();

  //
  // Initialize global pointer and register it with the naming service
  // (running on context 0).
  //
  if (context == 0) {

    for (int i = 0; i < 10; i++) {
      array[i] = i;
    }
    printArray(array, 10, 0);

    // Register (bind) the global pointer with the name service so that
    // other contexts can later obtain a reference to it (resolve).
    NameServiceLite::bind(nameForGP, gp);
  }

  hpcxx_barrier();

  //
  // The other context needs to retrieve the global pointer and work
  // with its contents.
  //

  if (context == 1) {

    HPCxx_ContextID* nameServerContext = group->getContextID(0);
    hpcxx_invoke(*nameServerContext, gp, resolve_ID, nameForGP);

    //
    // Now that we have the global pointer, read 10 integers from it.
    //
    gp.read(array, 10);

    //
    // modify the array
    //
    for (int i = 0; i < 10; i++) {
      array[i] = 9 - i;
    }

    //
    // write the modified array back to its owner
    //
    gp.write(array, 10);
  }

  // wait for array to be modified
  hpcxx_barrier();

  if (context == 0) {
    printArray(array, 10, 0);

    // check to see if it has been correctly modifiedd
    for (int i = 0; i < 10; i++) {
      if (array[i] != 9 - i) error = 1;
    }
  }

  //
  // exit HPC++
  //
  hpcxx_barrier();
  hpcxx_exit(group);

  if (context == 0) {
    cout << endl;
    if (error != 0) {
      cout << "ERROR, GlobalPointer example did not complete correctly" <<endl;
    } else {
      cout << "SUCCESSfully completed GlobalPointer example " << endl;
    }
  }

  return error;
}


//-----------------------------------------------------------------------------
// Prints the contents of an integer array to a given context.
//-----------------------------------------------------------------------------
static void printArray(int array[], int size, int context)
{
    cout << endl << "  context[" << context << "]:" << endl;
    for (int i = 0; i < size; i++) {
      cout << "               array[" << i << "] = " << array[i] << endl;
    }
}

size_t hpcxx_sizeof(const char* sz)
{
  return ( sizeof(int) + strlen(sz) );
}


void hpcxx_pack(HPCxx_Buffer& buffer, const char** strArray, int count)
{
  for (int i = 0; i < count; i++) {
    int length = strlen(strArray[i]);
    buffer.pack(&length, 1);
    buffer.pack((char*) strArray[i], length);
  }
}

void hpcxx_unpack(HPCxx_Buffer& buffer, const char** strArray, int count)
{
  for (int i = 0; i < count; i++) {
    int length;
    buffer.unpack(&length, 1);
    //
    // strArray is not really const. It must be declared that way because
    // of the symetry required in the way the pack and unpack functions are
    // used.
    //
    char** str = (char**) strArray;
    str[i] = new char[length + 1];
    str[i][length] = '\0';
    buffer.unpack(str[i], length);
  }
}


// explicit instantiations of required templated classes and functions


template void hpcxx_pack  (HPCxx_Buffer &, HPCxx_GlobalPtr<int> *, int);
template void hpcxx_unpack(HPCxx_Buffer &, HPCxx_GlobalPtr<int> *, int);
