//
// ReaderWriter.C demonstrates the use of reader/writer locks.  This
// type of lock operates as follows:  If a thread tries to get a read
// lock on a resource, it will succeed only if no other thread holds a
// write lock on the resource, in which case the reader must wait until
// the writer releases the lock.  Conversely, if a thread tries to get
// a write lock on the resource, it must wait if any other thread holds
// a read or a write lock.
//
// The main program create three threads.  In the run method of these
// threads, various "simultaneous" read and write requests are made.
//

#include <iostream.h>
#include <hpcxx_rts.h>
#include "src/ReadWriteLock.h"

int               error = 0;		// set to non-zero if error occurs
int	          nDidRead = 0;		// # threads that did read
int	          nDidWrite = 0;	// # threads that did write
int	          maxReaders;		// maximum # of readers
int	          maxWriters;		// maximum # of writers
HPCxx_Mutex       mutex;		// mutex to aid in testing
HPCxx_Barrier     barrier(3);		// rendezvous object
ReadWriteLock     rwLock;		// the R/W lock variable

//
// This thread makes read and write requests to the ReadWriteLock
// in order to read or write to the shared variable.
//
class TestReaderWriter : public HPCxx_Thread 
{ 
  private: int id;

  public:  TestReaderWriter(const char* name, int id_): HPCxx_Thread(name)
    {
      id = id_;
    }

  //
  // The yields in the run method is to mix up the run order of
  // thread execution.
  //
  public:  void run()
    {
      int key = barrier.getKey();

      yield();

      if (id == 0) {
	maxReaders = 0;
	maxWriters = 0;
      }

      //
      // all instances are readers
      //
      rwLock.readLock();
      yield();
      mutex.lock();
      nDidRead += 1;
      if (maxReaders < rwLock.nReaders) maxReaders = rwLock.nReaders;
      if (maxWriters < rwLock.nWriters) maxWriters = rwLock.nWriters;
#ifdef VERBOSE
      cout << "        thread " << id << " received read permission\n";
#endif
      mutex.unlock();
      rwLock.readUnlock();

      barrier(key);

      if (id == 0) {
	if (nDidRead != 3 || maxReaders > 3 || maxWriters != 0) {
	  cout << "ERROR, incorrect number of ReaderWriters" << endl;
	  error = 1;
	}
        cout << "    RW test 1, # readers = "
             << nDidRead << ", # writers = " << nDidWrite << " max readers = "
             << maxReaders << ", max writers = " << maxWriters << endl;
	nDidRead = 0;
	maxReaders = 0;
	maxWriters = 0;
	yield();
      }

      //      barrier(key);

      //
      // id[2] is a writer, other instances are readers
      //
      if (id < 2) {
	rwLock.readLock();
	barrier(key);	// two should be able to read at once
	mutex.lock();
	nDidRead += 1;
	if (maxReaders < rwLock.nReaders) maxReaders = rwLock.nReaders;
	if (maxWriters < rwLock.nWriters) maxWriters = rwLock.nWriters;
	if (rwLock.nReaders > 0 && rwLock.nWriters > 0) {
	  cout << "ERROR, incorrect number of ReaderWriters, # readers = "
               << rwLock.nReaders << ", # writers = " << rwLock.nWriters
	       << endl;
	  error = 1;
	}
#ifdef VERBOSE
	cout << "        thread " << id << " received read permission\n";
#endif
	mutex.unlock();
	rwLock.readUnlock();
      } else {
        barrier(key);
	rwLock.writeLock();
	mutex.lock();
	nDidWrite += 1;
	if (maxReaders < rwLock.nReaders) maxReaders = rwLock.nReaders;
	if (maxWriters < rwLock.nWriters) maxWriters = rwLock.nWriters;
	if (rwLock.nReaders > 0 || rwLock.nWriters != 1) {
	  cout << "ERROR, incorrect number of ReaderWriters, # readers = "
               << rwLock.nReaders << ", # writers = " << rwLock.nWriters
	       << endl;
	  error = 1;
	}
#ifdef VERBOSE
	cout << "        thread " << id << " received write permission\n";
#endif
	mutex.unlock();
	rwLock.writeUnlock();
      }

      barrier(key);

      if (id == 0) {
	if (nDidRead != 2 || nDidWrite != 1
			  || maxReaders != 2 || maxWriters != 1) {
	  cout << "ERROR, incorrect number of ReaderWriters" << endl;
	  error = 1;
	}
        cout << "    RW test 2, # readers = "
             << nDidRead << ", # writers = " << nDidWrite << " max readers = "
             << maxReaders << ", max writers = " << maxWriters << endl;
	nDidRead = 0;
	nDidWrite = 0;
	maxReaders = 0;
	maxWriters = 0;
	yield();
      }

      //      barrier(key);

      //
      // id[0] is a reader, other instances are writers
      //
      if (id == 0) {
	rwLock.readLock();
	yield();
	mutex.lock();
	nDidRead += 1;
	if (maxReaders < rwLock.nReaders) maxReaders = rwLock.nReaders;
	if (maxWriters < rwLock.nWriters) maxWriters = rwLock.nWriters;
	if (rwLock.nReaders != 1 || rwLock.nWriters > 0) {
	  cout << "ERROR, incorrect number of ReaderWriters, # readers = "
               << rwLock.nReaders << ", # writers = " << rwLock.nWriters
	       << endl;
	  error = 1;
	}
#ifdef VERBOSE
	cout << "        thread " << id << " received read permission\n";
#endif
	mutex.unlock();
	rwLock.readUnlock();
      } else {
	rwLock.writeLock();
	yield();
	mutex.lock();
	nDidWrite += 1;
	if (maxReaders < rwLock.nReaders) maxReaders = rwLock.nReaders;
	if (maxWriters < rwLock.nWriters) maxWriters = rwLock.nWriters;
	if (rwLock.nReaders > 0 || rwLock.nWriters != 1) {
	  cout << "ERROR, incorrect number of ReaderWriters, # readers = "
               << rwLock.nReaders << ", # writers = " << rwLock.nWriters
	       << endl;
	  error = 1;
	}
#ifdef VERBOSE
	cout << "        thread " << id << " received write permission\n";
#endif
	mutex.unlock();
	rwLock.writeUnlock();
      }

      barrier(key);

      if (id == 0) {
	if (nDidRead != 1 || nDidWrite != 2
			  || maxReaders != 1 || maxWriters != 1) {
	  cout << "ERROR, incorrect number of ReaderWriters" << endl;
	  error = 1;
	}
        cout << "    RW test 3, # readers = "
             << nDidRead << ", # writers = " << nDidWrite << " max readers = "
             << maxReaders << ", max writers = " << maxWriters << endl;
	nDidRead = 0;
	nDidWrite = 0;
	maxReaders = 0;
	maxWriters = 0;
      }

      barrier(key);

      //
      //  all are writers
      //
      rwLock.writeLock();
      yield();
      mutex.lock();
      nDidWrite += 1;
      if (maxReaders < rwLock.nReaders) maxReaders = rwLock.nReaders;
      if (maxWriters < rwLock.nWriters) maxWriters = rwLock.nWriters;
      if (rwLock.nReaders > 0 || rwLock.nWriters != 1) {
	cout << "ERROR, incorrect number of ReaderWriters, # readers = "
             << rwLock.nReaders << ", # writers = " << rwLock.nWriters
	     << endl;
	error = 1;
      }
#ifdef VERBOSE
      cout << "        thread " << id << " received write permission\n";
#endif
      mutex.unlock();
      rwLock.writeUnlock();

      barrier(key);

      if (id == 0) {
	if (nDidWrite != 3 || maxReaders != 0 || maxWriters != 1) {
	  cout << "ERROR, incorrect number of ReaderWriters" << endl;
	  error = 1;
	}
	cout << "    RW test 4, # readers = "
             << nDidRead << ", # writers = " << nDidWrite << " max readers = "
             << maxReaders << ", max writers = " << maxWriters << endl;
      }

      barrier(key);
    } // end run()
};


int main(int argc, char* argv[])
{
  int i;
  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  int nThreads = 3;
  HPCxx_Thread* thread[3];
  cout << endl;

  //
  // create threads
  //
  thread[0] = new TestReaderWriter("ReaderWriter0", 0);
  thread[1] = new TestReaderWriter("ReaderWriter1", 1);
  thread[2] = new TestReaderWriter("ReaderWriter2", 2);

  //
  // start threads
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->start();
  }

  //
  // wait for threads to finish
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->join();
  }

  cout << endl;
  if (error != 0) {
    cout << "ERROR, reader/writer threads did not complete correctly" << endl;
    return error;
  } else {
    cout << "Successfully completed ReaderWriter" << endl;
  }

  //
  // clean up
  //
  for (i = 0; i < nThreads; i++) {
    delete thread[i];
  }

  return hpcxx_exit(g);
}
