/* 
    this is the C version of the NAS app benchmark 1,
    the "embarassingly parallel" benchmark.
    on 64 bit systems, double precision should be disabled.
 
    based on the fortran version by david h. bailey january 8, 1991
 
    m is the log_2 of the number of complex pairs of uniform (0, 1) random
    numbers.  mk is the log_2 of the size of each batch of uniform random
    numbers.  mk can be set for convenience on a given system, since it does
    not affect the results.  second is a double precision function that
    returns elapsed cpu time in seconds.
*/ 

#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>

#include <nexus.h>
#include <hpcxx_rts.h>
#include <HPCxx_Sync.cc>
#include <HPCxx_Buffer.h>
#include <hpcxx_invoke.cc>
#include <hpcxx_register.cc>
#include <HPCxx_GlobalPtr.h>
#include <HPCxx_GlobalPtr.cc>
#include <hpcxx_invokeImpl.cc>
#include <HPCxx_GlobalRefImpl.cc>

void hpcxx_barrier();

#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )

void* _nx_st_rsh_info(void)
{
  return(NULL);
} 
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/times.h>

#define NODES 8
/*
   mm = m - mk;
   int nn = (int) exp2(mm);
   int nk = (int) exp2(mk);
*/
#ifndef FULLBENCHMARK
#  define m 24
#  define mk 8
#  define mm 16
#  define nn (256*256)
#  define nk 256
#  define nq 10
#else
#  define m 28
#  define mk 8
#  define mm 20
#  define nn (1024*1024)
#  define nk 256
#  define nq 10
#endif

int THREADS;

double max(double x, double y)
{
  if(x > y) return x; else  return y;
}

float second()
{
  struct tms t1;
  times(&t1);
  return( t1.tms_utime/60.0 );
}

double aint( double x)
{
  if (x < 0.0) return -floor(-x);
  else return floor(x);
}

class Gaussian : public HPCxx_Runnable {
public:
  int myId;
  double a;
  double s;
  double an;
  double x[2*nk+1], q[nq];
  double gc;
  double r23, r46, t23, t46;
  HPCxx_CSem* cs;
  
  void init(int id, int base, HPCxx_CSem* cs_);
  double randlc(double* x, double*  a);
  void vranlc(int n, double* x, double a,double* y);
  void compute_pairs(int kk);
  void run(void);
};   

void Gaussian::init(int id, int base, HPCxx_CSem* cs_)
{
  int i;
  double t1, t2;
  cs = cs_;
  myId = id+base;
  
  a = 1220703125.0;
  s = 271828183.0;
  r23 = 1.0;
  r46 = 1.0;
  t23 = 1.0;
  t46 = 1.0;
  for(i = 0; i < 24; i++) {
    r23 = 0.50 * r23;
    t23 = 2.0 * t23;
  }
  for(i = 1; i < 47; i++) {
    r46 = 0.50 * r46;
    t46 = 2.0 * t46;
  }
  for(i = 0; i < nq; i++)
    q[i] = 0.0;
  t1 = a;
  for(i = 1; i <= mk+1; i++)
    t2 = randlc (&t1, &t1);
  
  an = t1;
  gc = 0.0;
  // printf("init done \n");
};

void Gaussian::run(void)
{
  int k;
  for(k = myId+1; k <= nn; k = k+NODES)
    compute_pairs(k);

  cs->incr();
}

void Gaussian::compute_pairs(int k)
{
  int kk, i;
  double t1, t2;
  
  kk = k - 1;
  t1 = s;
  t2 = an;
  
  /* find starting seed t1 for this kk. */
  
  for(i = 1; i <= 100; i++) {
    int ik;
    double t3;
    ik = kk / 2;
    if(2*ik != kk) t3 = randlc(&t1,&t2);
    if(ik == 0) break;
    t3 = randlc(&t2, &t2);
    kk = ik;
  }
  
  /* printf("t1 = %f\n", t1); */
  
  /* compute uniform pseudorandom numbers. */
  
  vranlc(2 * nk, &t1, a, x);
  
  for(i = 1; i <= nk; i++) {
    double x1, x2, t1, t2, t3, t4;
    int l;
    x1 = 2.0 * x[2*i-1] - 1.0;
    x2 = 2.0 * x[2*i] - 1.0;
    t1 = x1*x1 + x2*x2;
    if(t1 <= 1.0) {
      gc = gc + 1.0;
      t2 = sqrt((double) (-2.0 * log (t1) / t1));
      t3 = fabs(x1 * t2);
      t4 = fabs(x2 * t2);
      l = (int) max(t3, t4);
      q[l] = q[l] + 1.0;
    }
  }
} 


/* 
   this routine returns a uniform pseudorandom double precision number in the
   range (0, 1) by using the linear congruential generator
   
   x_{k+1} = a x_k  (mod 2^46)
   
   where 0 < x_k < 2^46 and 0 < a < 2^46.  this scheme generates 2^44 numbers
   before repeating.  the argument a is the same as 'a' in the above formula,
   and x is the same as x_0.  a and x must be odd double precision integers
   in the range (1, 2^46).  the returned value randlc is normalized to be
   between 0 and 1, i.e. randlc = 2^(-46) * x_1.  x is updated to contain
   the new seed x_1, so that subsequent calls to randlc using the same
   arguments will generate a continuous sequence.
   
   this routine should produce the same results on any computer with at least
   48 mantissa bits in double precision floating point data.  on 64 bit
   systems, double precision should be disabled.
   
   david h. bailey	october 26, 1990
   
   if this is the first call to randlc, compute r23 = 2 ^ -23, r46 = 2 ^ -46,
   t23 = 2 ^ 23, and t46 = 2 ^ 46.  these are computed in loops, rather than
   by merely using the ** operator, in order to insure that the results are
   exact on all systems.  this code assumes that 0.50 is represented exactly.
   */

double Gaussian::randlc(double* x, double*  a)
{
  double t1, t2, t3, t4, a1, a2, x1, x2, z; 
  /* break a into two parts such that a = 2^23 * a1 + a2. */
  
  t1 = r23 * *a;
  a1 = aint (t1);
  a2 = *a - t23 * a1;
  
  /* break x into two parts such that x = 2^23 * x1 + x2, compute */
  /* z = a1 * x2 + a2 * x1  (mod 2^23), and then		  */
  /* x = 2^23 * z + a2 * x2  (mod 2^46).	        	  */
  
  t1 = r23 * *x;
  x1 = aint (t1);
  x2 = *x - t23 * x1;
  t1 = a1 * x2 + a2 * x1;
  t2 = aint (r23 * t1);
  z = t1 - t23 * t2;
  t3 = t23 * z + a2 * x2;
  t4 = aint (r46 * t3);
  *x = t3 - t46 * t4;
  return( r46 * *x);
}

/*
  this routine generates n uniform pseudorandom double precision numbers in
  the range (0, 1) by using the linear congruential generator
  
  x_{k+1} = a x_k  (mod 2^46)
  
  where 0 < x_k < 2^46 and 0 < a < 2^46.  this scheme generates 2^44 numbers
  before repeating.  the argument a is the same as 'a' in the above formula,
  and x is the same as x_0.  a and x must be odd double precision integers
  in the range (1, 2^46).  the n results are placed in y and are normalized
  to be between 0 and 1.  x is updated to contain the new seed, so that
  subsequent calls to vranlc using the same arguments will generate a
  continuous sequence.  if n is zero, only initialization is performed, and
  the variables x, a and y are ignored.
  
  this routine is the standard version designed for scalar or risc systems.
  however, it should produce the same results on any single processor
  computer with at least 48 mantissa bits in double precision floating point
  data.  on 64 bit systems, double precision should be disabled.
  
  david h. bailey	october 26, 1990
  
  if this is the first call to vranlc, compute r23 = 2 ^ -23, r46 = 2 ^ -46,
  t23 = 2 ^ 23, and t46 = 2 ^ 46.  see comments in randlc.
  */

void Gaussian::vranlc (int n, double* x, double a,double* y)
{
  int i;
  double a1, a2, t1;
  
  
  if( n == 0) return;
  
  /* break a into two parts such that a = 2^23 * a1 + a2. */
  
  t1 = r23 * a;
  a1 = (double) aint (t1);
  a2 = a - t23 * a1;
  
  /* generate n results.   this loop is not vectorizable. */
  /* 
     break x into two parts such that x = 2^23 * x1 + x2, compute
     z = a1 * x2 + a2 * x1  (mod 2^23), and then
     x = 2^23 * z + a2 * x2  (mod 2^46).
     */
  
  for(i = 1; i <= n; i++) {
    double t1, t2, t3, t4, x1, x2, z;
    t1 = r23 * *x;
    x1 = (double) aint (t1);
    x2 = *x - t23 * x1;
    t1 = a1 * x2 + a2 * x1;
    t2 = (double) aint ((double) (r23 * t1));
    z = t1 - t23 * t2;
    t3 = t23 * z + a2 * x2;
    t4 = (double) aint ((double)(r46 * t3));
    *x = t3 - t46 * t4;
    y[i] = r46 * *x;
  }
}


double q[nq], z[nq];
Gaussian* G;

HPCxx_GlobalPtr<double> fetch_gp(int i)
{
  double* p = &(q[0]);
  HPCxx_GlobalPtr<double> gp(p);
  printf("fetch_int called\n");
  return gp;
}

int main(int argc, char** argv)
{ 
  float tm0, tm1;
  double gc;
  int i, k, j;
  HPCxx_Group* g;
  hpcxx_id_t fetch_id = hpcxx_register(fetch_gp, 33);
  hpcxx_init(argc, argv, g);
  //HPCxx_BarImpl bar(g);
  //bar.synchronousSetUp();
  int NC = g->getNumContexts();
  int MyCID = g->getLocalIndex();
  THREADS = NODES/NC;
  tm0 = second();
  
  printf("num contexts = %d, threads per context = %d\n", NC, THREADS);
  
  G = new Gaussian[THREADS];
  HPCxx_CSem cs(THREADS);
  for(i = 0; i < THREADS; i++)
    G[i].init(i,MyCID*THREADS, &cs);
  
  for(i = 0; i < nq; i++)
    q[i] = 0.0;
  gc = 0.0;
  
  /* each instance of this loop may be performed independently. */
  HPCxx_Thread** th = new HPCxx_Thread*[THREADS];
  for(k = 0; k < THREADS; k++) {
    th[k] = new HPCxx_Thread(&G[k]);
    th[k]->start();
  }
  cs.wait();
  for(k = 0; k < THREADS; k++) {
    for( i = 0; i < nq; i++) {
      q[i] += G[k].q[i];
      gc   += G[k].gc;
    }
    th[k]->join();
  }
    
  hpcxx_barrier();
  if(MyCID == 0) {
    for(i = 1; i < NC; i++) {
      HPCxx_ContextID* othercid = g->getContextID(i);
      HPCxx_GlobalPtr<double> gp;
      int offset = 0;
      hpcxx_invoke(*othercid, gp, fetch_id,  offset);
      //int size = nq;
      //gp.read(z, size);
      for(j = 0; j < nq; j++)
	q[j] += gp[j];
    }
    tm1 = second();
    printf("benchmark 1 results:\n cpu time = %10.4f n = 2^%d\n",
	   tm1-tm0, m);
    printf("no. gaussian pairs = %15.0f\n",gc);
    for( i= 0; i < nq; i++)
      printf("q(%d) = %15.0f\n",i,q[i]);
  }
  //bar.barrier();
  hpcxx_exit(g);
  return(0);
}


// Explicit Template Instatiations

template void hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<HPCxx_GlobalPtr<double> > >*,int);
template void hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<double>*,int);
template void hpcxx_unpack(HPCxx_Buffer&,HPCxx_GlobalPtr<double>*,int);

