#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>
#include <sys/time.h>

#include <nexus.h>
#include <hpcxx_rts.h>

#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )

void* _nx_st_rsh_info(void)
{
  return(NULL);
} 

/**********************************************************************
  This is a simple parallel quicksort that sorts 5 million floats
  first sequentially and then with the parallel method.  The parallel
  version looks at the depth of the tree and the size of the block to
  sort. if the depth is less than 7 or size is greater than 5000, it
  creates a thread to do the computation of that branch.

  the sequence of numbers is random, so you get different speed-up
  results with each run.
***********************************************************************/

#define DATASIZE 500000

template <class T>
void swap(T& x, T& y)
{
  T tmp;
  tmp = x; 
  x = y;
  y = tmp;
}

int th_count = 0;

template <class T>
class SortThread: public HPCxx_Thread {
  HPCxx_Sync<int>* s;
  T* x;
  int low, high;
  int depth;

public:
  SortThread(HPCxx_Sync<int>* s_, T* x_, int low_, int high_, int depth_): 
    s(s_), x(x_), low(low_), high(high_), depth(depth_), 
    HPCxx_Thread(){ th_count++; }

  void run();
};
         
template <class T>
void pqsort( T* x, int low, int high, int depth)
{
  // printf("calling with low= %d high = %d dept = %d\n", low, high, depth);
  HPCxx_Sync<int>* s = NULL;
  SortThread<float>* Th;
  int i = low;
  int j = high;

  int checkflag = 0;
  if (i >= j) return;
  T m = x[i];
  
  while( i < j) {
    while ((x[j] >= m) && (j > low)) j--;
    while ((x[i] < m) && (i < high)) i++;
    if(i < j){ swap(x[i], x[j]);  }
  }

  if(j < high) {
    if( (j-low < 5000) ||  (depth > 6) ) {
      pqsort(x, low, j, depth+1);
    }
    else{
      s = new HPCxx_Sync<int>();
      
      Th = new SortThread<float>(s, x, low, j, depth+1);
      Th->start();
      checkflag = 1;
    }
  }
  if(j+1 > low) pqsort(x, j+1, high, depth);

  int y;
  if(checkflag) {
    s->read(y);
    Th->join();
    delete Th;
  }
}

template <class T>
void qsort( T* x, int low, int high, int depth){
  int i = low;
  int j = high;
  
  if (i >= j) return;
  T m = x[i];
  while( i < j) {
    while ((x[j] >= m) && (j > low)) j--;
    while ((x[i] < m) && (i < high)) i++;
    if(i < j){  swap(x[i], x[j]);  }
  }

  if(j < high)  qsort(x, low, j, depth+1);
  if(j+1 > low) qsort(x, j+1, high, depth+1);
}


template <class T>
void SortThread<T>::run()
{
  int k = 0;
  pqsort(x, low, high, depth);
  s->write(k);
}


int main(int argc, char** argv)
{
   timeval start_time, stop_time;
   float seq_time, work_time;
   HPCxx_Group* group;
   HPCxx_Sync<int> sync;
   HPCxx_Mutex     mutex;
   HPCxx_Cond      cond;

   hpcxx_init(argc, argv, group);

   float *A = new float[DATASIZE];
   float *B = new float[DATASIZE];
   srand(time(NULL)); // initialize the random number generator
   int i;
   for(i = 0; i < DATASIZE; i++) B[i] = A[i] = rand();

   cout << "doing sequential qsort" << endl;
   gettimeofday(&start_time, NULL);
   qsort(B, 0, DATASIZE-1, 0); 
   gettimeofday(&stop_time, NULL);
   seq_time = ElapseTime(start_time, stop_time);
   cout << "time = " << seq_time << endl;

   cout << "start parallel version" << endl;
   gettimeofday(&start_time, NULL);
   pqsort(A, 0, DATASIZE-1, 0);
   gettimeofday(&stop_time, NULL);
   work_time = ElapseTime(start_time, stop_time);
   cout << "done time =" << work_time << ", speed-up = " 
         << seq_time/work_time << ", number of threads =" << th_count << endl;

   int errorcnt = 0;
   for( i = 0; i < DATASIZE-1; i++)
           if(A[i] > A[i+1]) errorcnt++;

  cout << "total errors =" << errorcnt << endl;

   hpcxx_exit(group);
   return 0;
}


  
