#ifndef __READWRITELOCK_H_
#define __READWRITELOCK_H_

//#----------------------------------------------------------------------------
// <summary>Locking allowing multiple readers but only one writer.</summary>
//
// This class takes two synchronization primatives and creates a third -- the
// ReadWriteLock.  This type of lock operates as follows:  If
// a thread tries to get a read lock on a resource, it will succeed
// only if no other thread holds a write lock on the resource, in which
// case the reader must wait until the writer releases the lock.
// Conversely, if a thread tries to get a write lock on the resource,
// it must wait if any other thread holds a read or a write lock.
//
// <note role=caution> This implementation allows new read requests to be
// served even if an outstanding write request exists. </note>
//
// <hr>
//#----------------------------------------------------------------------------

#include <HPCxx_Mutex.h>
#include <HPCxx_Cond.h>

class ReadWriteLock
{
  public:

  //#--------------------------------------------------------------------------
  // Constructs a new ReadWriteLock.
  //#--------------------------------------------------------------------------
  ReadWriteLock(void);

  //#--------------------------------------------------------------------------
  // Obtains a write lock.  If unable the obtain the lock, the calling
  // thread blocks until the lock has been obtained.
  //
  // Returns an implementation-dependent error code.
  //#--------------------------------------------------------------------------
  int readLock(void);

  //#--------------------------------------------------------------------------
  // Obtains a read lock.  If unable the obtain the lock, the calling
  // thread blocks until the lock has been obtained.
  //
  // Returns an implementation-dependent error code.
  //#--------------------------------------------------------------------------
  int writeLock(void);

  //#--------------------------------------------------------------------------
  // Unlocks a read lock.  The scheduling priority determines which blocked
  // thread is resumed.
  //
  // Returns an implementation-dependent error code.
  //#--------------------------------------------------------------------------
  int readUnlock(void);

  //#--------------------------------------------------------------------------
  // Unlocks a write lock.  The scheduling priority determines which blocked
  // thread is resumed.
  //
  // Returns an implementation-dependent error code.
  //#--------------------------------------------------------------------------
  int writeUnlock(void);

  private:  int		nReaders;
  private:  int		nWriters;
  private:  HPCxx_Mutex	mutex;
  private:  HPCxx_Cond	forWriter;
    
  friend class TestReaderWriter;
};


#endif  // __READWRITELOCK_H_

