//
// watch.C demonstrates the use of the HPCxx_Cond class which is used
// to synchronize threads on the value of a condition variable.  The
// main routine creates three threads.  Two of the threads increment a
// counter.  The third thread simply waits until the counter reaches
// a predetermined value and exits.
//

#include <iostream.h>
#include <hpcxx_rts.h>

int            error = -1;	// set to non-zero if error occurs
int            count = 0;	// the counter
const int      watchCount = 6;	// counter value that thread waits upon
HPCxx_Mutex    countMutex;	// mutex object used for synchronization
HPCxx_Cond     watchCondition;	// condition object used for synchronization

//
// This class increments the watch counter
//
class IncThread : public HPCxx_Thread 
{ 
  private: int id;

  public:  IncThread(const char* name, int id_): id(id_), HPCxx_Thread(name)
    {
    }
  public:  void run()
    {
      yield();
      for (int i = 0; i < 2*watchCount; i++) {
        countMutex.lock();	// ensure that only one thread increments count
	count += 1;

	if (count == watchCount) {
	  watchCondition.signal();
	}

	countMutex.unlock();
	yield();
      }
    }
};


//
// This class watches for the counter to reach a threshold
//
class WatchThread : public HPCxx_Thread 
{ 
  public:  WatchThread(const char* name): HPCxx_Thread(name)
    {
    }
  public:  void run()
    {
      countMutex.lock();

      while (count < watchCount) {

	if (count >= watchCount) error = 1;

	cout << endl;
	cout << "    WatchThread::run(): about to start waiting, count is ";
	cout << count << endl << flush;

	watchCondition.wait(countMutex);

	if (count < watchCount) error = 1;
	else			error = 0;
      }

      cout << "    WatchThread::run(): finished waiting, count is ";
      cout << count << endl << flush;

      countMutex.unlock();
    }
};


int main(int argc, char* argv[])
{
  int i;
  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  int nThreads = 3;
  HPCxx_Thread* thread[3];

  //
  // create threads
  //
  thread[0] = new IncThread("inc1", 1);
  thread[1] = new IncThread("inc2", 2);
  thread[2] = new WatchThread("watch");

  //
  // start threads
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->start();
  }

  //
  // wait for threads to finish
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->join();
  }

  cout << endl;
  if (error != 0) {
    cout << "ERROR, watch threads did not complete correctly" << endl;
    return error;
  } else {
    cout << "Successfully completed watch, count = " << count << endl;
  }

  //
  // clean up
  //
  for (i = 0; i < nThreads; i++) {
    delete thread[i];
  }

  hpcxx_exit(g);
  return 0;
}
