//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson, Craig Rasmussen
//
// HPC++_VER: $Id: HPCxx_Barrier.h,v 1.1 1998/06/23 00:55:19 ejohnson Exp $
//##########################################################################

#ifndef __HPCXX_BARRIER_H_
#define __HPCXX_BARRIER_H_

#include <stdio.h>		// printf()
#include <stdlib.h>		// exit()
#include <hpcxx_types.h>
#include <HPCxx_Group.h>
#include <HPCxx_SyncQ.h>
#include <HPCxx_Mutex.h>

//#----------------------------------------------------------------------------
//#CLASS HPCxx_Barrier
//
// <summary> A Barrier synchronization class </summary>
//
// The most basic group synchronization operation is the <em>barrier</em>.
// A group of threads can synchronize at a location in a program by calling a
// barrier, which forces them to halt until the rest of the threads in
// the group have also called the barrier.  The barrier works across multiple
// contexts, with multiple threads per context.
//
// The HPCxx_Barrier class can be used in conjunction with the HPCxx_Group
// class (through an argument to the HPCxx_Barrier constructor) to implement
// a group barrier operation. The barrier() member function will force a
// thread which performs it to block until every other thread in its group
// has also called it.
//
// <note role=caution> The barrier class is not fully implememented and
// currently works only for multiple threads in a single context.  Also note
// that currently the association between a group of threads and a member
// of that group is not enforced.
// </note>
//
// A barrier can be initialized with an integer to represent the number
// of threads participating in the barrier, for example, to construct a
// barrier with three participating threads, code the following:
// <srcblock>
// HPCxx_Barrier barrier(3);
// </srcblock>
// Then, within the run method of the threads, the following could be used
// to force the 3 threads to rendezvous at two particular points in the code.
// <srcblock>
// int key = barrier.getKey();
// //
// // have threads do some work, then synchronize at a barrier
// //
// barrier(key);
// //
// // do more work, then synchronize again at a barrier
// //
// barrier(key);
// </srcblock>
//
// <prerequisite>
//   <li> <a href="HPCxx_Group.html>HPCxx_Group</a>
//   <li> <a href="HPCxx_Thread.html>HPCxx_Thread</a>
// </prerequisite>
// <hr>
//#----------------------------------------------------------------------------
class HPCxx_Barrier {
protected:
  HPCxx_SyncQ<int> *sync;
  HPCxx_Mutex m;
  int key_counter;
  int inside;
public:
  int size;

  HPCxx_Barrier(int height);
  HPCxx_Barrier(HPCxx_Group& group);

  //#--------------------------------------------------------------------------
  // Destroys this barrier.
  //#--------------------------------------------------------------------------
  ~HPCxx_Barrier(void)
  {
    delete [] sync;
  }

  int getKey(void);
  void barrier(int key);

  //#--------------------------------------------------------------------------
  // Causes the calling thread to block until all participating threads have
  // called one of these two barrier functions.
  // <ul>
  // <li><b>key</b> -- the key that has been previously assigned to the
  // calling thread (obtained via the getKey() method).
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>

  void reset(void) { key_counter=0; inside=0;};
  
  void operator()(int key)
  { 
	  barrier(key); 
  }
};


//#----------------------------------------------------------------------------
//#CLASS HPCxx_Bcast
//
// <summary>Provides the abilitity to broadcast data and synchronize on its arrival.</summary>
//#----------------------------------------------------------------------------
template< class T>
class HPCxx_Bcast : public HPCxx_Barrier {
  T data;
public:
  HPCxx_Bcast(HPCxx_Group& g) : HPCxx_Barrier(g) { }
  T  operator()(int key, T* vp){
    if (vp == NULL) barrier(key);
    else{
      data = *vp;
      barrier(key);
    }
  }
};

//#----------------------------------------------------------------------------
//#CLASS HPCxx_Mcast
//
// <summary>Provides the ability to share data and synchronize on its arrival</summary>
// <todo asof="1998/3/4">
//   <li> Needs a destructor.
// </todo>
// <hr>
//#----------------------------------------------------------------------------
template< class T>
class HPCxx_Mcast : public HPCxx_Barrier {
  T* data;
public:
  HPCxx_Mcast(HPCxx_Group& g) : HPCxx_Barrier(g)
  {
    data = new T[size];
  }

  ~HPCxx_Mcast(void)
  {
    delete[] data;
  }

  T* operator()(int key, T &v)
  {
    data[key] = v;
    barrier(key);
    return data;
  }
};


#endif

