//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Buffer.h,v 1.1 1998/06/23 00:55:22 ejohnson Exp $
//##########################################################################

#ifndef __HPCXX_BUFFER_H_
#define __HPCXX_BUFFER_H_

#include <HPCxx_BufferImpl.h>
#include <HPCxx_SyncQ.h>

//#----------------------------------------------------------------------------
//#CLASS HPCxx_Buffer
//
// <summary>Provides simple buffering capabilities</summary>
//
// This class provides a buffer into which one can copy other objects using
// a variety of packing routines.  Subsequently, objects written into the
// buffer can be retrieved using corresponding unpacking routines.  This
// provides a convenient mechanism to use in streaming an object to a disk
// or network interface.  Note that the pack and unpack routines can
// provide marshalling and unmarshalling capabilities between machines with
// different binary architectures.
//
// Note: The objects must be unpacked in the same order that they were packed.
// It is an error to do otherwise.
//
// For example, to allocate a buffer of 16 bytes and copy an integer and a
// float, one writes:
// <srcblock>
// HPCxx_Buffer buffer(16);
// float aFloat = 3.3;
// int anInt = 33;
// //
// // now copy (pack) these objects into the buffer
// //
// buffer.pack(&anInt);
// buffer.pack(&aFloat);
// </srcblock>
// At this point the buffer can be transferred elsewhere (perhaps across the
// network) and then unpacked (in the same order as the packing):
// <srcblock>
// buffer.unpack(&anInt);
// buffer.unpack(&aFloat);
// </srcblock>
// <hr>
//#----------------------------------------------------------------------------
class HPCxx_Buffer {
public:
  HPCxx_BufferImpl impl_;

public:

  //#--------------------------------------------------------------------------
  // Constructs an HPCxx_Buffer of a given size.
  // <ul>
  // <li><b>size</b> -- the size of the buffer to create (in bytes)
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_Buffer(int size);

  //#--------------------------------------------------------------------------
  // Constructs an HPCxx_Buffer by copying another buffer.
  // <ul>
  // <li><b>b</b> -- the buffer from which to copy
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_Buffer(const HPCxx_BufferImpl& b): impl_(b){}

  //#--------------------------------------------------------------------------
  // Destructor.  Frees any memory that was allocated for the buffer.
  //#--------------------------------------------------------------------------
  ~HPCxx_Buffer(void);

  //#--------------------------------------------------------------------------
  // These methods pack (marshall) data into the buffer.  They all take a
  // const pointer to a block of data with an optional count for the number
  // of objects to copy.
  // <ul>
  // <li><b>data</b> -- pointer to data to copy into the buffer
  // <li><b>count</b> -- the number of objects to copy
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  int pack(const float* data, int count = 1);
  int pack(const double* data, int count = 1);
  int pack(const short* data, int count = 1);
  int pack(const u_short* data, int count = 1);
  int pack(const int* data, int count = 1);
  int pack(const u_int* data, int count = 1);
  int pack(const long* data, int count = 1);
  int pack(const u_long* data, int count = 1);
  int pack(const char* data, int count = 1);
  int pack(const u_char* data, int count = 1);
  // </group>

  //#--------------------------------------------------------------------------
  // These methods unpack (unmarshall) data from the buffer.  They all take a
  // pointer to a block of data with an optional count for the number of
  // objects to copy from the buffer.
  // <ul>
  // <li><b>data</b> -- on output, receives the objects from the buffer
  // <li><b>count</b> -- the number of objects to copy
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  void unpack(float* data, int count = 1);
  void unpack(double* data, int count = 1);
  void unpack(short* data, int count = 1);
  void unpack(u_short* data, int count = 1);
  void unpack(int* data, int count = 1);
  void unpack(u_int* data, int count = 1);
  void unpack(long* data, int count = 1);
  void unpack(u_long* data, int count = 1);
  void unpack(char* data, int count = 1);
  void unpack(u_char* data, int count = 1);
  // </group>

#ifndef CONTEXT_ID_AS_INT
  int pack(const HPCxx_ContextID* data, int count = 1);
  void unpack(HPCxx_ContextID* data, int count = 1);
#endif

};


class HPCxx_BufferBlock{
  public:
    HPCxx_Buffer *lcBuffer, *rcBuffer, *pBuffer;
    HPCxx_SyncQ<int> lcLock, rcLock, pLock;
    HPCxx_SyncQ<int> lcLock2, rcLock2, pLock2;
    HPCxx_BufferBlock(){
        pBuffer= lcBuffer= rcBuffer= (HPCxx_Buffer*) 0;
       }
    static int registrationID(){ return -1; }
};



//#--------------------------------------------------------------------------
// These functions are just wrappers to the HPCxx_Buffer pack and unpack
// routines.  See the cooresponding functions descriptions above for
// information regarding the function parameters.
//#--------------------------------------------------------------------------
// <group>
void hpcxx_pack(HPCxx_Buffer &buffer, float *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, double *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, short *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, u_short *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, int *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, u_int *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, long *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, u_long *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, char *data, int count = 1);
void hpcxx_pack(HPCxx_Buffer &buffer, u_char *data, int count = 1);

#ifndef CONTEXT_ID_AS_INT
void hpcxx_pack(HPCxx_Buffer &buffer, HPCxx_ContextID *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, HPCxx_ContextID *data, int count = 1);
#endif

void hpcxx_unpack(HPCxx_Buffer &buffer, float *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, double *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, short *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, u_short *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, int *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, u_int *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, long *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, u_long *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, char *data, int count = 1);
void hpcxx_unpack(HPCxx_Buffer &buffer, u_char *data, int count = 1);
// </group>

#endif  // __HPCXX_BUFFER_H_



