//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson, Craig Rasmussen
//
// HPC++_VER: $Id: HPCxx_Cond.h,v 1.1 1998/06/23 00:55:27 ejohnson Exp $
//##########################################################################

// $Id: HPCxx_Cond.h,v 1.1 1998/06/23 00:55:27 ejohnson Exp $

#ifndef __HPCXX_COND_H_
#define __HPCXX_COND_H_

#include <HPCxx_Mutex.h>
#include <HPCxx_CondImpl.h>

//#----------------------------------------------------------------------------
// <summary> Enables threads to synchronize on the value of a condition variable. </summary>
//
// Whereas the HPCxx_Mutex class allows threads to synchronize by controlling
// their access to data via a locking mechanism, the HPCxx_Cond class
// allows threads to synchronize on the value of the data.  Cooperating threads
// wait until data reaches some particular state or until some particular
// event occurs.  A rule of thumb: use mutex locks to synchronize access
// to shared data and use condition variables to synchronize threads on events.
//
// Before waiting on an event, a mutex must be obtained and passed as an
// argument to the wait method.  For example, the following code segment
// causes a thread to wait for a counter to reach some threshold value:
// <srcblock>
// countMutex.lock();
// while (count < threshold) {
//   countCondition.wait(countMutex);
// }
// //
// // do something now that the condition has been met
// //
// countMutex.unlock();
// </srcblock>
//
// The condition variable must be signaled to release the waiting thread.
// For example
// <srcblock>
// countMutex.lock();
// count = threshold;
// countCondition.signal()
// countMutex.unlock();
// </srcblock>
//
// Notice that both the waiting thread and the signalling thread both lock
// the same mutex.  This may seem strange but it is the proper usage as
// the wait() method unlocks the mutex when it is called.
//
// <prerequisite>
//   <li> <a href="HPCxx_Mutex.html">HPCxx_Mutex</a>
//   <li> <a href="HPCxx_Thread.html">HPCxx_Thread</a>
// </prerequisite>
//
// <hr>
//#----------------------------------------------------------------------------
class HPCxx_Cond {
private:
  HPCxx_CondImpl  impl_;

public:
  HPCxx_Cond(void);
  HPCxx_Cond(const HPCxx_Cond& cond);
  virtual ~HPCxx_Cond(void);

  HPCxx_Cond& operator=(const HPCxx_Cond& cond);

  //#--------------------------------------------------------------------------
  // Suspends the calling thread until another thread releases the wait
  // condition.  The calling thread is required to lock the given mutex
  // variable before calling wait.  The mutex is unlocked when the wait
  // state begins so that another thread can acquire the mutex before
  // releasing the wait condition.
  // <ul>
  // <li><b>mutex</b> -- the mutex object that has been locked
  // </ul>
  //#--------------------------------------------------------------------------
  virtual int wait(HPCxx_Mutex& mutex);

  //#--------------------------------------------------------------------------
  // Releases one of the potentially many threads waiting on the condition.
  // The calling thread must first acquire a lock on the mutex object
  // (that was used in the call to wait()) before calling this method.
  //#--------------------------------------------------------------------------
  virtual int signal(void);

  //#--------------------------------------------------------------------------
  // Releases all of the threads waiting on the condition.  The calling thread
  // must first acquire a lock on the mutex object (that was used in the call
  // to wait()) before calling this method.
  //#--------------------------------------------------------------------------
  virtual int broadcast(void);

};

#endif
