//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_GlobalPtr.h,v 1.4 1998/08/08 00:06:24 rasmussn Exp $
//##########################################################################

#ifndef __HPCXX_GLOBALPTR_H_
#define __HPCXX_GLOBALPTR_H_

//# include files
#include <hpcxx_types.h>
#ifndef HPCXX_NO_THREADS
#include <HPCxx_Context.h>
#endif
#include <HPCxx_GlobalRefImpl.h>

class HPCxx_Buffer;

//#----------------------------------------------------------------------------
//#CLASS HPCxx_GenericPtr
// <summary>The base class for the HPCxx_GlobalPtr class.</summary>
//
// This class is the base class for the templated HPCxx_GlobalPtr<T> set
// of classes.  The HPCxx_GenericPtr class declares virtual functions which
// enable the HPCxx_GlobalPtr classes to display polymorphic behavior.
//
//#----------------------------------------------------------------------------
class HPCxx_GenericPtr {
  public:

  virtual ~HPCxx_GenericPtr(void);

  virtual void* localPtr(void) = 0;
  virtual HPCxx_ContextIDImpl* contextIDImpl(void) = 0;
  virtual int offset(void) = 0;
  virtual hpcxx_type_t type(void) const = 0;
  virtual void initialize(void* localPtr,
			  int offset, HPCxx_ContextIDImpl* contextID) = 0;

  static int registrationID(){ return 99998; }
};


//#----------------------------------------------------------------------------
//#CLASS HPCxx_GlobalPtr
//#---------------------
// <summary>A class which supports get and put operations on remote data</summary>
//
// A global pointer is a generalization of the C pointer type to support
// pointers to objects that may exist in another address space (context).  It
// is closely linked to the idea of a global reference, which is an object
// that is a proxy (local stand-in) for a remote object.
//
// An HPCxx_GlobalPtr object is templated on a generic type T.  An
// actual instantiation can either be on an intrinsic type such as float,
// i.e. HPCxx_GlobalPtr<float>, or on a more complex, user-defined type or
// class.  However, user-defined types must have an hpcxx_pack(), an
// hpcxx_unpack(), and an hpcxx_sizeof() defined, and the type must also
// be registered with hpcxx_register().
//
// The two basic operations of a global pointer are read and write.  A read
// obtains data from a remote memory location specified by the global
// pointer and stores it in a local address.  The write does the opposite,
// it copies data from a local address into the remote address specified by the
// global pointer.
// 
// For example, 
// <srcblock>
// int array[10];
// HPCxx_GlobalPtr<int> gp(array);
// </srcblock>
// creates a global pointer to the integer array in the local context.
// This global pointer can be subsequently copied to another context and 
// either written to or read from in the following ways (where gp is
// a copy of the global pointer that was created above):
// <srcblock>
// int i = 0;
// int j = 48;
// i = gp[3];
// gp[4] = j;
// //
// // or equivalently
// //
// gp.read(&i, 1, 3);
// gp.write(&j, 1, 4); 
// </srcblock>
// A global can also be dereferenced like any other pointer.  For example,
// copy and assignment of floating point number through a global pointer
// is shown below:
// <srcblock>
// HPCxx_GlobalPtr<float> gp;
// *gp = 3.14;
// float r = 2. - *gp;
// </srcblock>
//
// In summary, several examples of the operations supported by the
// GlobalPtr class are shown below.  The form: assign(dest, src) is used to
// describe the action on remote data.
// <srcblock>
// HPCxx_GlobalPtr<int> gp1, gp2;
// int   i;
// int   x;
// int  *q;
// int  &r;
// int *&z;
// HPCxx_ContextID &c;
//  gp1 =  gp2 : pointer copy
// *gp1 = *gp2 : assign(rput, rget)
//  gp1 = &x   : assign(local, local)
// *gp1 =  x   : assign(rput, local)
//  x = *gp    : assign(local, rget)
//  q =  gp    : cast to local pointer
//  gp = q     : assign(local, local) // equiv to gp1 = &x
//  gp[i] = gp[i] : *gp = *gp w/ offset
//  gp[i] = x     : *gp = x w/ offset
//  x = gp[i]     : x = *gp w/ offset
//
// z = &gp : cast to address of local pointer
// c = gp : cast to context id
// r = *gp : illegal?
// </srcblock>
// <hr>
//#----------------------------------------------------------------------------
template <class T> class HPCxx_GlobalPtr : public HPCxx_GenericPtr {

public:

  //#--------------------------------------------------------------------------
  // These methods construct a new global pointer given various parameters
  // described below.
  // <ul>
  // <li><b>ptr</b> -- the local pointer for this object
  // <li><b>id</b>  -- the context in which the local pointer is valid
  // <li><b>gp</b>  -- the global pointer to copy from
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  HPCxx_GlobalPtr();
  HPCxx_GlobalPtr(T* ptr);
  HPCxx_GlobalPtr(T* ptr, HPCxx_ContextID* id);
  HPCxx_GlobalPtr(const HPCxx_GlobalPtr<T> &gp);
  // </group>
  
  //#--------------------------------------------------------------------------
  // Destroys this global pointer.  Does not affect the local memory that this
  // object points to.
  //#--------------------------------------------------------------------------
  virtual ~HPCxx_GlobalPtr();

  //#--------------------------------------------------------------------------
  // Derefences a HPCxx_GlobalPtr to return a HPCxx_GlobalRefImpl
  //#--------------------------------------------------------------------------
  HPCxx_GlobalRefImpl<T>& operator*();

  //#--------------------------------------------------------------------------
  // Conversion operator.  Returns the local pointer if in the local context,
  // NULL otherwise.
  //#--------------------------------------------------------------------------
  operator T *() const; 

  //#--------------------------------------------------------------------------
  // Conversion operators.  Returns the context id in which the local pointer
  // is a valid memory address.
  //#--------------------------------------------------------------------------
  // <group>
  operator HPCxx_ContextID &() const;
  HPCxx_ContextID * getContextID() const;
  // </group>

  //#--------------------------------------------------------------------------
  // Assigns this object to another global pointer.  Both the local pointer
  // and the context id are copied into this object.
  // <ul>
  // <li><b>gp</b> -- the global pointer to assign to this object
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_GlobalPtr<T>& operator=(const HPCxx_GlobalPtr<T>& gp);

  //#--------------------------------------------------------------------------
  // Assigns the local pointer of this object to the given local memory
  // location.  The context id of this object becomes local.
  // <ul>
  // <li><b>ptr</b> -- the local pointer to assign to this object
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_GlobalPtr<T>& operator=(const T* ptr);
  
  //# GlobalPtr Arithmetic

  //#--------------------------------------------------------------------------
  // Prefix ++ operator (increments the local pointer).
  //#--------------------------------------------------------------------------
  HPCxx_GlobalPtr<T>& operator++();

  //#--------------------------------------------------------------------------
  // Postfix ++ operator (increments the local pointer).
  //#--------------------------------------------------------------------------
  HPCxx_GlobalPtr<T> operator++(int);

  //#--------------------------------------------------------------------------
  // Prefix -- operator (decrements the local pointer).
  //#--------------------------------------------------------------------------
  HPCxx_GlobalPtr<T>& operator--();

  //#--------------------------------------------------------------------------
  // Postfix -- operator (decrements the local pointer).
  //#--------------------------------------------------------------------------
  HPCxx_GlobalPtr<T> operator--(int);

  //#--------------------------------------------------------------------------
  // These methods either increment or decrement the local pointer by the given
  // amount.
  // <ul>
  // <li><inc>gp</b> -- the amount to increment by
  // <li><dec>gp</b> -- the amount to decrement by
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  HPCxx_GlobalPtr<T>& operator+=(int inc);
  HPCxx_GlobalPtr<T>& operator-=(int dec);
  HPCxx_GlobalPtr<T> operator+(int inc);
  HPCxx_GlobalPtr<T> operator-(int dec);
  // </group>
  
  //#--------------------------------------------------------------------------
  // Tests the equality of two global pointers.  Two global pointers are equal
  // if their local pointers and context id's are equal.
  // <ul>
  // <li><b>gp</b> -- the global pointer to compare against
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  int operator==(const HPCxx_GlobalPtr<T>& gp);
  int operator==(const HPCxx_GlobalPtr<T>& gp) const;
  int operator!=(const HPCxx_GlobalPtr<T>& gp);
  int operator!=(const HPCxx_GlobalPtr<T>& gp) const;
  // </group>

  //#--------------------------------------------------------------------------
  // Compares this object with the given global pointer.
  // <ul>
  // <li><b>gp</b> -- the global pointer to compare against
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  int operator<(const HPCxx_GlobalPtr<T>& gp);
  int operator<(const HPCxx_GlobalPtr<T>& gp) const;
  int operator>(const HPCxx_GlobalPtr<T>& gp);
  int operator>(const HPCxx_GlobalPtr<T>& gp) const;
  int operator<=(const HPCxx_GlobalPtr<T>& gp);
  int operator<=(const HPCxx_GlobalPtr<T>& gp) const;
  int operator>=(const HPCxx_GlobalPtr<T>& gp);
  int operator>=(const HPCxx_GlobalPtr<T>& gp) const;
  // </group>

  //#--------------------------------------------------------------------------
  // Returns a global reference indexed into this object by the given amount.
  // This is a convenient way to access an array in another context.
  // <ul>
  // <li><b>index</b> -- the amount to index into the array
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_GlobalRefImpl<T> operator[](int index);
  
  //#--------------------------------------------------------------------------
  // Reads a block of data from location specified by this global pointer.
  // <ul>
  // <li><b>ptr</b> -- receives the data from the global pointer
  // <li><b>count</b> -- the number of objects to transfer
  // <li><b>offset</b> -- starting offset into the global pointer array
  // </ul>
  //#--------------------------------------------------------------------------
  void read(T* ptr, int count = 1, int offset = 0);

  //#--------------------------------------------------------------------------
  // Writes a block of data into the location specified by this global pointer.
  // <ul>
  // <li><b>ptr</b> -- source of the data to copy to the global pointer
  // <li><b>count</b>  -- the number of objects to transfer
  // <li><b>offset</b> -- a starting offset into the global pointer array
  // </ul>
  //#--------------------------------------------------------------------------
  void write(T* ptr, int count = 1, int offset = 0);
 
  //#--------------------------------------------------------------------------
  // Returns true if the pointer is local to our context, false otherwise.
  //#--------------------------------------------------------------------------
  bool isLocal() const;

  //#--------------------------------------------------------------------------
  // Copies the local pointer of this object into the given address.  WARNING,
  // this is a valid address only if this function returns true.
  // <ul>
  // <li><b>ptr</b> -- receives the local pointer
  // </ul>
  // Returns true if the pointer is local to our context, false otherwise.
  //#--------------------------------------------------------------------------
  bool isLocal(T*& ptr);

  //#--------------------------------------------------------------------------
  // Returns the registration id for this class.
  //#--------------------------------------------------------------------------
  static int registrationID(void)	{ return 0; }

  //#--------------------------------------------------------------------------
  // Returns the type of this class, e.g., HPCXX_TYPE_INT (for integer type)
  //#--------------------------------------------------------------------------
  virtual hpcxx_type_t type(void) const;

  //#--------------------------------------------------------------------------
  // All local pointers must be allocated with new (HPCXX_NEW)
  // to be portable and hence must call hpcxx_delete()
  // to destroy the local pointers:
  //
  // HPCxx_GlobalPtr<MyClass> gp = new (HPCXX_NEW) MyClass;
  // { some code ...}
  // hpcxx_delete(gp); // deletes gp's localPtr_ *not* gp
  //#--------------------------------------------------------------------------
  friend void hpcxx_delete(const HPCxx_GlobalPtr<T> &gp);

  //#--------------------------------------------------------------------------
  // Copies a given number of objects from the global pointer into the buffer.
  // <ul>
  // <li><b>buf</b> -- receives the data
  // <li><b>r</b> -- the global pointer to copy from
  // <li><count></b> -- the number of objects to transfer
  // </ul>
  //#--------------------------------------------------------------------------
  friend void hpcxx_pack(HPCxx_Buffer& buf, HPCxx_GlobalPtr<T>* r, int count);

  //#--------------------------------------------------------------------------
  // Copies a given number of objects from the buffer to the global pointer.
  // <ul>
  // <li><b>buf</b> -- source of the data
  // <li><b>r</b> -- the global pointer to receive the data
  // <li><count></b> -- the number of objects to transfer
  // </ul>
  //#--------------------------------------------------------------------------
  friend void hpcxx_unpack(HPCxx_Buffer& buf, HPCxx_GlobalPtr<T>* r, int count);

  int offset(void)		   { return globalRef_.offset(); };
  void* localPtr(void)		   { return globalRef_.localPtr(); };
  HPCxx_ContextIDImpl* contextIDImpl(void)
    {
      return globalRef_.contextIDImpl();
    };
  void initialize(void* localPtr, int offset, HPCxx_ContextIDImpl* contextID)
    {
      globalRef_.initialize(localPtr, offset, contextID);
    };

public:
  HPCxx_GlobalRefImpl<T> globalRef_;
  
};


#endif
