//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Mutex.h,v 1.1 1998/06/23 00:55:37 ejohnson Exp $
//##########################################################################

//# $Id: HPCxx_Mutex.h,v 1.1 1998/06/23 00:55:37 ejohnson Exp $

#ifndef __HPCXX_MUTEX_H_
#define __HPCXX_MUTEX_H_

//# include files
#include <HPCxx_MutexImpl.h>

//# forward declarations
class HPCxx_Cond;

//#----------------------------------------------------------------------------
// <summary> This class provides the ability to protect (lock) critical sections of code. </summary>
//
// The HPCxx_Mutex class provides the two basic functions, lock and unlock.
// These functions allow a shared resource to be protected from a
// race condition.  Using a mutex object, one can give threads a turn
// at having exclusive access to data.  When one thread has obtained a
// lock, other threads cannot simultaneously access the same data or
// critical region of code.  For example, the following code segment
// locks a critical region so that only one thread can execute
// within this region at a given time:
// <srcblock>
// HPCxx_Mutex mutex;
// mutex.lock();
//    //
//    // a region of code to be protected...
//    //
// mutex.unlock();
// </srcblock>
//
// Unlike Java, the HPC++ thread API does not support synchronized methods
// or CC++ atomic members; similar functionality can be provided by the
// lock and unlock mechanisms.
//
// <prerequisite>
//   <li> <a href="HPCxx_Thread.html">HPCxx_Thread</a>
// </prerequisite>
// <hr>
//#----------------------------------------------------------------------------

class HPCxx_Mutex {
public:
  //#--------------------------------------------------------------------------
  // These methods construct a new mutex object.
  // <ul>
  // <li><b>mutex</b> -- the mutex object from which to construct this object
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  HPCxx_Mutex(void);
  HPCxx_Mutex(const HPCxx_Mutex& mutex);
  // </group>

  //#--------------------------------------------------------------------------
  // Destroys this mutex object.
  //#--------------------------------------------------------------------------
  virtual ~HPCxx_Mutex(void);

  //#--------------------------------------------------------------------------
  // Overloads the = operator to copy a mutex to this mutex.
  // <note role=warning.gif> What does it mean to have a copy of a mutex? </note>
  //#--------------------------------------------------------------------------
  HPCxx_Mutex& operator=(const HPCxx_Mutex& mutex);

  //#--------------------------------------------------------------------------
  // Locks an unlocked mutex.  If the mutex is already locked, the calling
  // thread blocks until the thread that currently holds the mutex releases it.
  //
  // Returns an implementation-dependent error code.
  //#--------------------------------------------------------------------------
  virtual int lock(void);

  //#--------------------------------------------------------------------------
  // Unlocks a mutex.  The scheduling priority determines which blocked thread
  // is resumed.
  //
  // Returns an implementation-dependent error code.
  //#--------------------------------------------------------------------------
  virtual int unlock(void);

  friend HPCxx_Cond;

private:
  HPCxx_MutexImpl impl_;

};

#endif
