//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Sync.h,v 1.1 1998/06/23 00:55:46 ejohnson Exp $
//##########################################################################

#ifndef __HPCXX_SYNC_H_
#define __HPCXX_SYNC_H_

#include <hpcxx_types.h>
#include <HPCxx_Mutex.h>
#include <HPCxx_Cond.h>

//#----------------------------------------------------------------------------
// <summary>A templated class for synchronization on the write operation to a variable.</summary>
//
// An HPCxx_Sync<T> object is a variable that can be written to once and
// read from as many times as is needed.  However, if a read is attempted prior
// to the write, the reading thread will be blocked.  Many readers can be
// waiting for a single HPCxx_Sync<T> object and when a value is written
// to it, all the readers are released.  Readers that come after the initial
// write see this as a <i>const</i> value.
//
// Given the following declaration:
// <srcblock>
// HPCxx_Sync<int> iSync;
// </srcblock>
// any read operation on the sychronization variable iSync, e.g.,
// <srcblock>
// int i = iSync;
// </srcblock>
// will block until another thread writes to iSync, e.g.,
// <srcblock>
// iSync.write(3);
// </srcblock>
// <hr>
//#----------------------------------------------------------------------------
template<class T> class HPCxx_Sync {
public:
  //#--------------------------------------------------------------------------
  // Constructs an HPCxx_Sync object.
  //#--------------------------------------------------------------------------
  HPCxx_Sync(void);

  //#--------------------------------------------------------------------------
  // Destroys this object.
  //#--------------------------------------------------------------------------
  ~HPCxx_Sync(void);

  //#--------------------------------------------------------------------------
  // Copies the value of this object to the given variable.
  // <ul>
  // <li> val -- reference to the variable to which to copy the value of this
  // object
  // </ul>
  //#--------------------------------------------------------------------------
  void read(T& val);

  //#--------------------------------------------------------------------------
  // Converts this object in place to type T by returning its value.
  // Essentially a read operation on this object.
  //#--------------------------------------------------------------------------
  operator T ();
  
  //#--------------------------------------------------------------------------
  // These methods assign (write) the given value to this object.
  // <ul>
  // <li> val -- value to assign to this object
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>  
  void write(T& val);
  T& operator=(T& val);
  // </group>

  //#--------------------------------------------------------------------------
  // Peeks (checks) to see if the Sync<T> variable has been written to.
  // If the variable has been written to, the T value is copied to val
  // and true is returned.  Otherwise, returns false.
  // <ul>
  // <li> val -- reference to the variable to which to copy the value of this
  // object on return
  // </ul>
  //#--------------------------------------------------------------------------
  bool peek(T& val);

  //#--------------------------------------------------------------------------
  // Returns the address of the value of this object OR the address previously
  // set by setBufferAddr() (see below).  Used for remote reads of blocks.
  //#--------------------------------------------------------------------------
  void* getBufferAddr();

  //#--------------------------------------------------------------------------
  // Sets the buffer address for blocks.
  //#--------------------------------------------------------------------------
  void setBufferAddr(void *);

  //#--------------------------------------------------------------------------
  // Returns the registration id for this class.  Used in Remote Method
  // Invocation (RMI).
  //#--------------------------------------------------------------------------
  static int registrationID() { return 99999; }

private:
  T           val_;
  HPCxx_Mutex mutex_;
  HPCxx_Cond  cond_;
  int         virgin_;
  void*       buffer_;

};

#endif
