//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_SyncQ.h,v 1.1 1998/06/23 00:55:50 ejohnson Exp $
//##########################################################################

// $Id: HPCxx_SyncQ.h,v 1.1 1998/06/23 00:55:50 ejohnson Exp $

#ifndef __HPCXX_SYNCQ_H_
#define __HPCXX_SYNCQ_H_

#include <hpcxx_types.h>
#include <HPCxx_SyncCell.h>
#include <HPCxx_Mutex.h>
#include <HPCxx_Cond.h>

//#----------------------------------------------------------------------------
// <summary>A templated class providing synchronized access to a queue of values.</summary>
//
// HPCxx_SyncQ<T> provides a queue of values of type T.  The queue is FIFO
// ordered in the sense that the first value written to the queue will be the
// first value read from the queue.  Any attempt to read from the queue
// if it is empty will cause the reading thread to block until a value has
// been written.  At some point in time after a value has been written to the
// queue, the waiting thread completes the read operation and continues
// execution.  Note that there is no guarantee in which order threads
// waiting to read from the queue will be granted the request.
//
// Given the following declaration:
// <srcblock>
// HPCxx_SyncQ<int> intQueue;
// </srcblock>
// read operations on the queue, e.g.,
// <srcblock>
// int i = intQueue;
// int j = intQueue;
// </srcblock>
// will block until another thread has written an equivalent number (or more)
// of values into the queue, e.g.,
// <srcblock>
// intQueue.write(1);		// at some point in the future, i = 1
// intQueue.write(2);		// at some point in the future, i = 2
// </srcblock>
// <hr>
//#----------------------------------------------------------------------------
template<class T> class HPCxx_SyncQ {
public:
  //#--------------------------------------------------------------------------
  // Constructs an HPCxx_SyncQ object.
  //#--------------------------------------------------------------------------
  HPCxx_SyncQ(void);

  //#--------------------------------------------------------------------------
  // Destroys this object.
  //#--------------------------------------------------------------------------
  ~HPCxx_SyncQ(void);

  //#--------------------------------------------------------------------------
  // Copies the next value in the queue to the given variable and removes
  // it from the queue.
  // <ul>
  // <li> val -- a reference to the variable to which to copy the next value in
  // the queue
  // </ul>
  //#--------------------------------------------------------------------------
  void read(T& val);

  //#--------------------------------------------------------------------------
  // Converts this object in place to type T by returning the next value (and
  // removing it) from the queue.  Essentially a read operation on this object.
  //#--------------------------------------------------------------------------
  operator T(void);

  //#--------------------------------------------------------------------------
  // These methods copy (write) the given value to this queue.
  // <ul>
  // <li> val -- the value to place in the queue
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>  
  void write(T& val);
  T& operator= (T& val);
  // </group>

  //#--------------------------------------------------------------------------
  // Returns how many items are currently in the queue.
  //#--------------------------------------------------------------------------
  int length(void);

  //#--------------------------------------------------------------------------
  // Peeks (checks) to see if an item is in the queue.  If there are one or 
  // more items currently in the queue, the next T value is copied to val
  // and true is returned.  Otherwise, returns false.  Does <i>not</i> remove
  // the item from the queue.
  // <ul>
  // <li> val -- a reference to the variable to which to copy the next value
  // in the queue
  // </ul>
  //#--------------------------------------------------------------------------
  bool peek(T& val);

  //#--------------------------------------------------------------------------
  // Checks to see if an item is in the queue.  If there are one or 
  // more items currently in the queue, the next T value is copied to val.
  // Otherwise the thread blocks until data becomes available.  Does <i>not</i>
  // remove the item from the queue.
  // <ul>
  // <li> val -- a reference to the variable to which to copy the next value
  // in the queue
  // </ul>
  //#--------------------------------------------------------------------------
  void waitAndCopy(T& val);

private:
  HPCxx_SyncCell<T> *head_;
  HPCxx_SyncCell<T> *tail_;
  int                count_;
  HPCxx_Mutex        mutex_;
  HPCxx_Cond         cond_;

};

#endif
