//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Thread.h,v 1.1 1998/06/23 00:55:52 ejohnson Exp $
//##########################################################################

//# $Id: HPCxx_Thread.h,v 1.1 1998/06/23 00:55:52 ejohnson Exp $

#ifndef __HPCXX_THREAD_H_
#define __HPCXX_THREAD_H_

#include <hpcxx_types.h>
#include <HPCxx_Runnable.h>
#include <HPCxx_ThreadImpl.h>
//# need to put in new profiling package, remove for now
//#include <HPCxx_Profile.h>

//#----------------------------------------------------------------------------
// <summary> A class for managing a thread of execution. </summary>
//
// HPCxx_Thread is a class providing an object-oriented interface to a
// native thread package such as pthreads.  The design of this class is
// similar to the Java thread system.  Normal usage of the class
// is to inherit from HPCxx_Thread in order to override the run method 
// and provide the desired functionality to the thread.  For example:
// <srcblock>
// class Thread: public HPCxx_Thread {
//   public: void run(void)
//           {
//             cout << "Thread " << name() << " is running" << cout;
//           }
// };
// </srcblock>
// Given the Thread class defined above, the following code segment will create
// a thread and cause the thread to run by calling its start method:
// <srcblock>
// Thread thread("demonstration 1");
// thread.start();
// //
// // do something, then wait for the thread to exit
// //
// thread.join();
// </srcblock>
// In this example, the calling thread blocks on the join() call until the run
// method of the created thread completes.
//
// As an alternative, a thread object could be constructed by passing a
// delegate to the HPCxx_Thread constructor.  The delegate class must inherit
// from the Runnable interface and implement the run method.  For example,
// the delegate class could be defined as,
// <srcblock>
// class ThreadDelegate: public Runnable {
//   public: void run(void)
//           {
//             cout << "Thread " << name() << " is running" << cout;
//           }
// };
// </srcblock>
// and constructed and run by,
// <srcblock>
// ThreadDelegate delegate;
// HPCxx_Thread thread(&delegate, "demonstration 2");
// thread.start();
// //
// // do something, then wait for the thread to exit
// //
// thread.join();
// </srcblock>
//
// <note role=caution> Stop and related methods have been deprecated in the
// Java API. </note>
//
// <todo asof="1998/3/4">
//   <li> Implement a container of current threads and provide an id for them
// </todo>
// <hr>
//#----------------------------------------------------------------------------

class HPCxx_Thread {
public:

  //#--------------------------------------------------------------------------
  // These methods construct a new thread object, taking a variety of input
  // parameters described below:
  // <ul>
  // <li><b>name</b> -- the name to be associated with the thread, if NULL, a
  // unique default name is provided
  // <li><b>target</b> -- the thread delegate (a Runnable object whose run
  //method will be called when the thread is started)
  // <li><b>thread</b> -- the thread object used to construct this new thread
  // </ul>
  //#--------------------------------------------------------------------------
  // <group>
  HPCxx_Thread(const char* name = (const char*) 0);
  HPCxx_Thread(HPCxx_Runnable* target, const char* name = (const char*) 0);
  HPCxx_Thread(const HPCxx_Thread& thread);
  // </group>

  //#--------------------------------------------------------------------------
  // Destroys the thread.
  //#--------------------------------------------------------------------------
  virtual ~HPCxx_Thread(void);

  //#--------------------------------------------------------------------------
  // Assigns (copies) the given thread object to this thread.
  // <ul>
  // <li><b>thread</b> -- the thread object to copy to this thread
  // </ul>
  //#--------------------------------------------------------------------------
  HPCxx_Thread& operator=(const HPCxx_Thread& thread);

  //#--------------------------------------------------------------------------
  // If this thread was constructed using a separate Runnable object as
  // a delegate, the delegates run method is called; otherwise, this method
  // does nothing.  Unless a delegate is to be supplied, subclasses of the
  // HPCxx_Thread class should override this method to obtain useful behavior.
  //#--------------------------------------------------------------------------
  virtual void run(void);

  //#--------------------------------------------------------------------------
  // Starts execution of the thread and calls the run() method of the 
  // thread Runnable delegate (or the run() method of the thread if no
  // delegate has been provided).
  //#--------------------------------------------------------------------------
  void start(void);

  //#--------------------------------------------------------------------------
  // Ceases execution of this thread.  DEPRECATED.
  //#--------------------------------------------------------------------------
  void stop(void* status);

  //#--------------------------------------------------------------------------
  // Causes the currently executing thread object to temporarily pause and
  // allow other threads to execute.
  //#--------------------------------------------------------------------------
  static void yield(void);

  //#--------------------------------------------------------------------------
  // Causes the currently executing thread to sleep (cease execution) for the
  // specified number of milliseconds plus the specifed number of nanoseconds.
  // Actual period of time will be rounded up to the resolution of the system
  // timer.
  // <ul>
  // <li><b>milliseconds</b> -- the number of milliseconds to sleep
  // <li><b>nanoseconds</b> -- the number of nanoseconds to sleep
  // </ul>
  //#--------------------------------------------------------------------------
  static void sleep(long milliseconds, long nanoseconds = 0);
  
  //#--------------------------------------------------------------------------
  // Suspends execution of thread until resume is called.  DEPRECATED.
  //#--------------------------------------------------------------------------
  void suspend(void);

  //#--------------------------------------------------------------------------
  // Resumes a suspended thread.  DEPRECATED.
  //#--------------------------------------------------------------------------
  void resume(void);

  //#--------------------------------------------------------------------------
  // Tests if this thread is alive (has been started but has not yet died).
  //
  // Returns TRUE if alive, FALSE otherwise.
  //#--------------------------------------------------------------------------
  bool isAlive(void);
			      
  //#--------------------------------------------------------------------------
  // Returns the currently executing thread object.
  //#--------------------------------------------------------------------------
  static HPCxx_Thread* currentThread(void);

  //#--------------------------------------------------------------------------
  // Returns the numeric ID of the thread.
  //#--------------------------------------------------------------------------
  static int getThreadID(void);

  //#--------------------------------------------------------------------------
  // Waits for the completion of this thread (by definition join returns as
  // soon as the thread is considered not alive).  If a time period is
  // provided, waits for at most the given milliseconds plus nanoseconds for
  // this thread to die.
  // <ul>
  // <li><b>milliseconds</b> -- the number of milliseconds to wait before
  // returning
  // <li><b>nanoseconds</b> -- the number of nanoseconds to wait before
  // returning
  // </ul>
  //#--------------------------------------------------------------------------
  void join(long milliseconds = 0, long nanoseconds = 0);

  //#--------------------------------------------------------------------------
  // Sets the name of this thread to the given name.
  // <ul>
  // <li><b>name</b> -- the name to be associated with this thread
  // </ul>
  //#--------------------------------------------------------------------------
  void setName(const char* name);

  //#--------------------------------------------------------------------------
  // Returns the name of this thread.
  //#--------------------------------------------------------------------------
  const char* getName(void);

  //#--------------------------------------------------------------------------
  // Returns the current priority of this thread, where
  // HPCXX_MIN_THREAD_PRIORITY <= getPriority() <= HPCXX_MAX_THREAD_PRIORITY
  //#--------------------------------------------------------------------------
  int getPriority(void);

  //#--------------------------------------------------------------------------
  // Changes the priority of this thread.  The priority is clamped to
  // (HPCXX_MIN_THREAD_PRIORITY, HPCXX_MAX_THREAD_PRIORITY)
  //#--------------------------------------------------------------------------
  int setPriority(int priority);

  friend void* HPCxx_ThreadImpl::start_(void* rhs);

private:
  // the target for this thread (implementation dependent)
  HPCxx_ThreadImpl  impl_;
  // the name that this thread is to be known by
  char name_[HPCXX_MAX_NAME_LEN];

};

#endif
