//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: hpcxx_rts.h,v 1.2 1998/07/02 16:11:00 rasmussn Exp $
//##########################################################################

#ifndef __HPCXX_RTS_H_
#define __HPCXX_RTS_H_

//
// This header file should be included in every project that
// plans to use HPC++ Lib
//

// HPC++ Library includes
#ifndef HPCXX_NO_THREADS

#include <hpcxx_rts_config.h>
#include <HPCxx_CSem.h>
#include <HPCxx_Context.h>
//#include <HPCxx_GlobalRef.h>
#include <HPCxx_Runnable.h>
#include <HPCxx_Barrier.h>
#include <HPCxx_Thread.h>
#include <HPCxx_Node.h>
#endif

#ifndef HPCXX_NO_RMFI
#include <HPCxx_Buffer.h>
#include <HPCxx_GlobalPtr.h>
#include <hpcxx_invoke.h>
#include <hpcxx_register.h>
#endif

#include <hpcxx_types.h>
#include <HPCxx_Cond.h>
#include <HPCxx_Group.h>
#include <HPCxx_Mutex.h>
#include <HPCxx_Sync.h>
#include <HPCxx_SyncCell.h>
#include <HPCxx_SyncQ.h>
#include <hpcxx_rtsImpl.h>

//
// HPC++ Global Variables
//
#ifndef HPCXX_NO_THREADS
extern HPCxx_Context   *hpcxx_currentContext;
extern HPCxx_Node      *hpcxx_currentNode;
#endif

// This object should be used whenever you need to perform
// and action only in the master context or need a handle
// to the master context
extern HPCxx_ContextID &hpcxx_masterContextID;
extern HPCxx_ContextID &hpcxx_nullContextID;
extern int              hpcxx_debug_level;
extern int              hpcxx_trace_level;

// a counter for keeping track of RMI and RF registration ids
extern hpcxx_id_t hpcxx_curID_s;

// All HPC++ programs must call this function as the very first
// action in main() before any other HPC++ calls are made
int hpcxx_init(int &argc, char **&argv, HPCxx_Group *&group);

// This function should be the last HPC++ call made
// and takes the group you obtained from hpcxx_init
int hpcxx_exit(HPCxx_Group *&group);

// hpcxx_poll() must be called only in non-preemptive
// environments. The constant HPCXX_PREEMPTIVE may be checked
// to determine the preemptiveness of the run time system.
//
// HPCXX_POLL_TIMEOUT determines the max poll timeout in milliseconds
// If a poll is not executed with HPCXX_POLL_TIMOUT milliseonds then
// a timer interrupt will occur and hpcxx_poll will be called.
// (This helps avoid deadlock situations.)
void hpcxx_poll();

// All memory allocated for objects to which HPCxx_GlobalPtr's point
// must be allocated and freed with the following functions
// 
// HPCxx_GlobalPtr's do not manage memory, it is the user's responsibilty
// to allocate and free the memory properly.
void *hpcxx_malloc(size_t size);
void hpcxx_free(void *ptr);

// Allocate an identifier space.  Two spaces are allocated and
// reserved by HPC++ Lib:
//     HPCXX_REGISTER_IDSPACE
//     HPCXX_GROUP_IDSPACE
// Return the next available idspace or HPCXX_INVALID_IDSPACE
// if no spaces are left
//
int hpcxx_createIDSpace();

// After having allocated an id space, id's may be obtained
// via the two functions.
//
// hpcxx_getGlobalID implies a global communication and guarantee's
// synchronization across contexts
//
// hpcxx_getLocalID implies no communication, but it is the user's
// responsibility to insure that keys are synchronized across contexts
// (SPMD mode)
hpcxx_id_t hpcxx_getGlobalID(int idSpace);
hpcxx_id_t hpcxx_getLocalID(int idSpace);

// Sets a synchronization barrier.
void hpcxx_barrier();

// Return the number of nodes in this computation
int hpcxx_getNumNodes();

// Get a list of all nodes
// User must 'delete []' the null terminated node array
// caveat - if getNumNodes is called it may not coincide with
// the number of nodes in this array due to dynamic node creation
// and deletion
#ifndef HPCXX_NO_THREADS
#ifndef HPCXX_NO_RMFI
HPCxx_GlobalPtr<HPCxx_Node> *hpcxx_getNodes();
#endif
#endif

// Print out an error in the master context only
void hpcxx_msg(const char *msg);

// Debug string
void hpcxx_debug(const char *msg, int debugLevel = 1);

// Abort the hpcxx computation.  hpcxx_abort() will call
// hpcxx_exit()
void hpcxx_abort(int status);

// Initialize and register hpcxx_types.
void hpcxx_typeInit(void);

// Overloaded operator new to allocate memory from shared space
void* operator new(size_t s, const hpcxx_shmallocator_t &);

#ifdef HAVE_OP_NEW_ARRAY
// Overloaded operator new[] to allocate memory from shared space
void* operator new[] (size_t s, const hpcxx_shmallocator_t &);
#endif

//#--- hpcxx_sizeof -----------------------------------------------------------
// Returns the packed size of type T in bytes.
//#----------------------------------------------------------------------------
template<class T> size_t hpcxx_sizeof(T& t);


//#--- hpcxx_delete -----------------------------------------------------------
// Deletes an object by explicitly calling its destructor and then freeing
// the memory for the object.  This function must be called if the object
// has been allocated using the (HPCXX_NEW) placement allocator.
//
// pT  pointer to the object that is to be deleted
//#----------------------------------------------------------------------------
template<class T> void hpcxx_delete(T* pT);


//#--- hpcxx_delete -----------------------------------------------------------
// Deletes an array of objects by explicitly calling the destructor for each
// of the objects in the array and then freeing the memory for the array.
// This function must be called if the array of objects has been allocated
// using the (HPCXX_NEW) placement allocator.
//
// aT        the array to delete
// nObjects  the number of objects in the array to delete
//#----------------------------------------------------------------------------
template<class T> void hpcxx_delete(T* aT, int nObjects);


// All local pointers must be allocated with new (HPCXX_NEW)
// to be portable and hence must call hpcxx_delete()
// to destroy the local pointers:
#ifndef HPCXX_NO_RMFI
template<class T> void hpcxx_delete(const HPCxx_GlobalPtr<T> &gp);
#endif


// Other global functions
//
// These prototypes are templated and/or overloaded numerous times
// and are located in their own header files
//
// hpcxx_pack.h
//   hpcxx_pack()
//   hpcxx_unpack()
// hpcxx_register.h
//   hpcxx_register()
// hpcxx_invoke.h
//   hpcxx_invoke()

#endif
