/* TestBuffer.C
 * -------------
 * This file performes a series of tests on the HPCxx_Buffer class.  The 
 * HPCxx_Buffer class supports all the primitive data types as well as arrays
 * of primitive types.  Thus this program tests buffer operations with each of 
 * the primitive types and one user defined type.  The size and content of each
 * array are randomly generated at run time to ensure unbiased testing.
 * 
 * usage: TestBuffer <test_code>
 * test_code: 1 - test primitieves only
 *            2 - test user defined type only
 *            3 - test both 
 */


#include <time.h>
#include <math.h>
#include <stdlib.h>
#include "Complex.C"
#include <hpcxx_rts.h>

/* 
 * RandomInteger
 * -------------
 * Simple random number generator.  Since using random() instead of rand()
 * the lower bits are usable and don't require normalize and scale.
 * Returns integer between lo and hi inclusive.
 */

int RandomInteger(int lo, int hi)
{
  int range = hi - lo + 1;

  return lo + random()% range;
}


int do_user_defined()
{
  ComplexArray test_in(RandomInteger(1, 100));
  int size = test_in.getsize();

  Complex* enter = new Complex[size];
  enter = test_in.getArray_Ptr();

  for (int i=0; i<size; i++) {
    Complex tmp(RandomInteger(1, 10) * 2.39, RandomInteger(1, 10) * 2.71);
    enter[i] = tmp;
  }
  test_in.out();
  delete [] enter;

  int buff_sz = hpcxx_sizeof(test_in);

  HPCxx_Buffer buffer(buff_sz);
  HPCxx_BufferImpl buff_impl = buffer.impl_;
  
  hpcxx_pack(buffer, &test_in);
  ComplexArray test_out(size);
  hpcxx_unpack(buffer, &test_out);
 
  test_out.out();
  
  return test_in == test_out;
}



/* A templated class is used because of the necessity to test all primitive 
 * data types.  The input class generates an array of the specified size and
 * type and has functions which initialize the array and test whether two 
 * instances of this type are equal.
 */

template <class T> 
class input {
private:
  T* X;
  int sz;
  
public:
  input (int size = 1) { X = new T[sz = size]; }
  ~input () { delete [] X; }

  T* getptr(void) { return X; }
  int getsize(void) { return sz; }


  void initializeArray(T val)
  {
    for (int j=0; j<sz; j++)
      X[j] = RandomInteger(1, 10) * val;
    
    /* To accomodate all types the user must pass a base input value to this 
     * function which can then be randomized with a randomly generated 
     * integer multiplier. */
  }

  int operator== (const input& out)
  {
    for (int i=0; i<sz; i++)
      if (X[i] == out.X[i])
	return 1;
    
    return 0;
  }
};



int do_primitives(void)
{
  int buff_sz = (sizeof(float) + sizeof(char) + sizeof(int) + sizeof(short) +
    sizeof(long) + sizeof(u_short) + sizeof(u_char) + sizeof(u_int) +
    sizeof(u_long)) * 100;
  
  HPCxx_Buffer buffer(buff_sz);

  // Create input of each primitive type.

  input<float> one_in(RandomInteger(1, 100));
  one_in.initializeArray(3.14);

  input<char> two_in(RandomInteger(1, 100));
  two_in.initializeArray(17);

  input<int> three_in(RandomInteger(1, 100));
  three_in.initializeArray(83);

  input<short> four_in(RandomInteger(1, 100));
  four_in.initializeArray(-12);

  input<long> five_in(RandomInteger(1, 100));
  five_in.initializeArray(19191);

  input<double> six_in(RandomInteger(1, 100));
  six_in.initializeArray(-63035.2353);

  input<u_short> seven_in(RandomInteger(1, 100));
  seven_in.initializeArray(23);

  input<u_char> eight_in(RandomInteger(1, 100));
  eight_in.initializeArray(21);

  input<u_int> nine_in(RandomInteger(1, 100));
  nine_in.initializeArray(25728);

  input<u_long> ten_in(RandomInteger(1, 100));
  ten_in.initializeArray(595784);

  // Create instances of the class to be used in the unpacking stage.

  input<float> one_out(one_in.getsize());
  input<char> two_out(two_in.getsize());
  input<int> three_out(three_in.getsize());
  input<short> four_out(four_in.getsize());
  input<long> five_out(five_in.getsize());
  input<double> six_out(six_in.getsize());
  input<u_short> seven_out(seven_in.getsize());
  input<u_char> eight_out(eight_in.getsize());
  input<u_int> nine_out(nine_in.getsize());
  input<u_long> ten_out(ten_in.getsize());

  // Pack the *_in objects into the buffer.

  buffer.pack(one_in.getptr(), one_in.getsize());
  buffer.pack(two_in.getptr(), two_in.getsize());
  buffer.pack(three_in.getptr(), three_in.getsize());
  buffer.pack(four_in.getptr(), four_in.getsize());
  buffer.pack(five_in.getptr(), five_in.getsize());
  buffer.pack(six_in.getptr(), six_in.getsize());
  buffer.pack(seven_in.getptr(), seven_in.getsize());
  buffer.pack(eight_in.getptr(), eight_in.getsize());
  buffer.pack(nine_in.getptr(), nine_in.getsize());
  buffer.pack(ten_in.getptr(), ten_in.getsize());

  // Now unpack the buffer using the *_out objects.

  buffer.unpack(one_out.getptr(), one_out.getsize());
  buffer.unpack(two_out.getptr(), two_out.getsize());
  buffer.unpack(three_out.getptr(), three_out.getsize());
  buffer.unpack(four_out.getptr(), four_out.getsize());
  buffer.unpack(five_out.getptr(), five_out.getsize());
  buffer.unpack(six_out.getptr(), six_out.getsize());
  buffer.unpack(seven_out.getptr(), seven_out.getsize());
  buffer.unpack(eight_out.getptr(), eight_out.getsize());
  buffer.unpack(nine_out.getptr(), nine_out.getsize());
  buffer.unpack(ten_out.getptr(), ten_out.getsize());

  // Check whether the *_in versions are equal to the *_out versions.

  int isSame[20];

  isSame[0] = one_in == one_out;
  isSame[1] = two_in == two_out;
  isSame[2] = three_in == three_out;
  isSame[3] = four_in == four_out;
  isSame[4] = five_in == five_out;
  isSame[5] = six_in == six_out;
  isSame[6] = seven_in == seven_out;
  isSame[7] = eight_in == eight_out;
  isSame[8] = nine_in == nine_out;
  isSame[9] = ten_in == ten_out;


  // Now use the hpcxx_pack and unpack functions to perform the 
  // same operations.

  HPCxx_Buffer buffer_2(buff_sz);

  // Pack the second buffer

  hpcxx_pack(buffer_2, one_in.getptr(), one_in.getsize());
  hpcxx_pack(buffer_2, two_in.getptr(), two_in.getsize());
  hpcxx_pack(buffer_2, three_in.getptr(), three_in.getsize());
  hpcxx_pack(buffer_2, four_in.getptr(), four_in.getsize());
  hpcxx_pack(buffer_2, five_in.getptr(), five_in.getsize());
  hpcxx_pack(buffer_2, six_in.getptr(), six_in.getsize());
  hpcxx_pack(buffer_2, seven_in.getptr(), seven_in.getsize());
  hpcxx_pack(buffer_2, eight_in.getptr(), eight_in.getsize());
  hpcxx_pack(buffer_2, nine_in.getptr(), nine_in.getsize());
  hpcxx_pack(buffer_2, ten_in.getptr(), ten_in.getsize());

  // Now unpack it

  hpcxx_unpack(buffer_2, one_out.getptr(), one_out.getsize());
  hpcxx_unpack(buffer_2, two_out.getptr(), two_out.getsize());
  hpcxx_unpack(buffer_2, three_out.getptr(), three_out.getsize());
  hpcxx_unpack(buffer_2, four_out.getptr(), four_out.getsize());
  hpcxx_unpack(buffer_2, five_out.getptr(), five_out.getsize());
  hpcxx_unpack(buffer_2, six_out.getptr(), six_out.getsize());
  hpcxx_unpack(buffer_2, seven_out.getptr(), seven_out.getsize());
  hpcxx_unpack(buffer_2, eight_out.getptr(), eight_out.getsize());
  hpcxx_unpack(buffer_2, nine_out.getptr(), nine_out.getsize());
  hpcxx_unpack(buffer_2, ten_out.getptr(), ten_out.getsize());

  // Check for equality

  isSame[10] = one_in == one_out;
  isSame[11] = two_in == two_out;
  isSame[12] = three_in == three_out;
  isSame[13] = four_in == four_out;
  isSame[14] = five_in == five_out;
  isSame[15] = six_in == six_out;
  isSame[16] = seven_in == seven_out;
  isSame[17] = eight_in == eight_out;
  isSame[18] = nine_in == nine_out;
  isSame[19] = ten_in == ten_out;

  int sum = 0;
  for (int i=0; i<20; i++)
    sum = sum + isSame[i];

  return (sum % 20 == 0) ? 1 : 0;
}


int main(int argc, char* argv[])
{
  srandom(time(0));
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);

  int test_code = atoi(argv[4]);
  int stat;

  if (test_code == 1) {
    stat = do_primitives();
    hpcxx_exit();
    return stat - 1;
  } 
  else if (test_code == 2) {
    stat = do_user_defined();
    hpcxx_exit();
    return stat - 1;
  } 
  else if (test_code == 3) { 
    stat = do_user_defined();
    int stat2 = do_primitives();
    hpcxx_exit();
    return stat + stat2 - 2; 
  } 
  else {
    cerr << "The test code must be in the range 1-3\n";
    hpcxx_exit();
    return 1;
  }
}

