//
// watch.C demonstrates the use of the HPCxx_Cond class which is used
// to synchronize threads on the value of a condition variable.  The
// main routine creates three threads.  Two of the threads increment a
// counter.  The third thread simply waits until the counter reaches
// a predetermined value and exits.
//

#include <iostream.h>
#include <hpcxx_rts.h>

#define START 0
#define PING 1
#define PONG 2
#define GAME_OVER 3
#define SHUT_DOWN 4
#define FINISHED 12

bool           error = false;	// set true if error occurs
int            count = 0;	// count of number of waits finished
int            game = START;	// state of game
HPCxx_Mutex    mutex;		// mutex object used for synchronization
HPCxx_Cond     condition;	// condition object used for synchronization
HPCxx_Mutex    mainMutex;	// mutex object used by main
HPCxx_Cond     mainCond;	// condition object used by main

class Ping : public HPCxx_Thread 
{ 
  public:  Ping(void): HPCxx_Thread("Ping")
    {
    }
  public:  void run(void)
    {
      //
      // play ping-pong
      //
      for (int i = 0; i < 5; i++) {
	mutex.lock();
	while (game == PONG || game == START) {
#ifdef DEBUG
	cerr << "    ping will wait, game = " << game << endl;
#endif
	  condition.wait(mutex);
	}
	if (game != PING) error = true;
	game = PONG;
	count += 1;
#ifdef DEBUG
	cerr << "    ping " << count << ", " << i << endl;
#endif
	condition.signal();
	mutex.unlock();
      }

      //
      // shutdown procedure
      //
      mutex.lock();
      while (game != SHUT_DOWN) {
#ifdef DEBUG
	cerr << "    ping will wait, game = " << game << endl;
#endif
	condition.wait(mutex);
      }

      if (game != SHUT_DOWN) error = true;
      count += 1;
#ifdef DEBUG
	cerr << "    ping finished " << count << endl;
#endif
      if (count == FINISHED) {
#ifdef DEBUG
	cerr << "    ping signalling main " << endl;
#endif
	mainCond.signal();
      }
      mutex.unlock();
    }
};


class Pong : public HPCxx_Thread 
{ 
  public:  Pong(void): HPCxx_Thread("Pong")
    {
    }
  public:  void run(void)
    {
      //
      // play ping-pong
      //
      for (int i = 0; i < 5; i++) {
	mutex.lock();
	while (game == PING || game == START) {
#ifdef DEBUG
	cerr << "    pong will wait, game = " << game << endl;
#endif
	  condition.wait(mutex);
	}
	if (game != PONG) error = true;
	game = PING;
	count += 1;
#ifdef DEBUG
	cerr << "    pong " << count << ", " << i << endl;
#endif
	condition.signal();
	mutex.unlock();
      }

      //
      // shutdown procedure
      //
      mutex.lock();
      game = GAME_OVER;
      mainCond.signal();		// signal main thread
      while (game != SHUT_DOWN) {
#ifdef DEBUG
	cerr << "    pong will wait, game = " << game << endl;
#endif
	condition.wait(mutex);
      }
      if (game != SHUT_DOWN) error = true;
      count += 1;
#ifdef DEBUG
	cerr << "    pong finished " << count << endl;
#endif
      if (count == FINISHED) {
#ifdef DEBUG
	cerr << "    pong signalling main " << endl;
#endif
	mainCond.signal();
      }
      mutex.unlock();
    }
};


int main(int argc, char* argv[])
{
  int i;
  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  int nThreads = 2;
  HPCxx_Thread* thread[2];

  //
  // create threads
  //
  thread[0] = new Ping();
  thread[1] = new Pong();

  //
  // start threads
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->start();
  }

  //
  // start game play
  //
  mutex.lock();
  game = PING;
  condition.signal();
  mutex.unlock();

  //
  // wait for game to finish
  //
  mainMutex.lock();
  while (game != GAME_OVER) {
    mainCond.wait(mainMutex);
#ifdef DEBUG
	cerr << "    main did wait, game = " << game << endl;
#endif
  }
  mainMutex.unlock();
#ifdef DEBUG
  cerr << "    main game over, finished " << count << endl;
#endif
  if (game != GAME_OVER) error = true;
  game = SHUT_DOWN;

  condition.broadcast();

  //
  // sort of do a join
  //
  mainMutex.lock();
  while (count != FINISHED) {
    mainCond.wait(mainMutex);
#ifdef DEBUG
	cerr << "    main did wait, finished = " << count << endl;
#endif
  }
  mainMutex.unlock();

  cerr << "    main did fake join, finished = " << count << endl;
  if (count != FINISHED) error = true;

  //
  // wait for threads to finish
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->join();
  }

  cout << endl;
  if (error) {
    cout << "ERROR, TestCond did not complete correctly, game = "
	 << game << endl;
    return 1;
  } else {
    cout << "Successfully completed TestCond" << endl;
  }

  //
  // clean up
  //
  for (i = 0; i < nThreads; i++) {
    delete thread[i];
  }

  return hpcxx_exit(g);
}
