// TestCond2 -- Implements a mailbox to synchronously send and receive
// messages using condition variables.  Multiple send and receive
// threads are used to test the synchronization (via condition variables)
// of the mailbox.

#include <iostream.h>
#include <hpcxx_rts.h>

static int error = 0;
static int sum = 0;
static int tally = 0;

class MailBox {
  public:

  MailBox(void)
  {
    message = 0;
    messageSent = false;
    messageReceived = false;

    sendPermission  = new HPCxx_Mutex();
    recvPermission  = new HPCxx_Mutex();
    sendRecv        = new HPCxx_Mutex();
    messageDelivery = new HPCxx_Cond();
  }


  ~MailBox(void)
  {
    delete sendPermission;
    delete recvPermission;
    delete sendRecv;
    delete messageDelivery;
  }


  void Send(int message)
  {
    int id = message;

    // Acquire permission to send a message, this will allow only one
    // thread at a time to send a message.  Other threads will block,
    // waiting for permission to send.
    sendPermission->lock();

    // Protect setting of send/receive flags from race conditions
    // between sender and receiver.
    sendRecv->lock();

#ifdef DEBUG
    cerr << "Sender:" << id << " will send " << message << endl;
#endif

    this->message = message;
    messageSent = true;

    // wake up potential receiver
    messageDelivery->signal();

    messageReceived = false;
    while (!messageReceived) {
        messageDelivery->wait(*sendRecv);
    }
#ifdef DEBUG
    cerr << "Sender:" << id << " did send" << endl;
#endif

    // At this point the message has been sent and received.
    sendRecv->unlock();

    // release to other senders
    sendPermission->unlock();
  }


  void Receive(int* message)
  {
    int id = *message;

    recvPermission->lock();

    // Wait for a message to be sent.  The sendRecv lock will be released
    // while the sender is waiting for a reply, so that this receiver
    // can receive the message.

    sendRecv->lock();
#ifdef DEBUG
    cerr << "Receiver:" << id << " will receive" << endl;
#endif
    while (!messageSent) {
        messageDelivery->wait(*sendRecv);
    }
    *message = this->message;

#ifdef DEBUG
    cerr << "Receiver:" << id << " did receive " << *message << endl;
#endif
    if (id != *message) error = 10 + *message;

    messageSent = false;
    messageReceived = true;
    messageDelivery->signal();

    // release so sender can finish
    sendRecv->unlock();

    // release to other receivers
    recvPermission->unlock();
  }

  private:

    int message;		// the message in the mailbox
    bool messageSent;		// notify receiver that message in mailbox
    bool messageReceived;	// notify sender that message has been received
    HPCxx_Mutex* sendPermission;	// permission to send a message
    HPCxx_Mutex* recvPermission;	// permission to receive a message
    HPCxx_Mutex* sendRecv;		// used to synch sender and receiver
    HPCxx_Cond* messageDelivery;	// used to synch sender and receiver

}; // MailBox


static MailBox mail;
HPCxx_Mutex sumMutex;

class Sender : public HPCxx_Thread
{

public: Sender(int id_) : id(id_), HPCxx_Thread("Sender Thread") { }

public: void run(void)
  {

    // try to mess up the order or execution
    if (id % 2) yield();
    mail.Send(id);
    sumMutex.lock();
    sum += id;
    tally += id;
    sumMutex.unlock();
  }

private:
  int id;
};


class Receiver : public HPCxx_Thread
{

public: Receiver(int id_) : id(id_), HPCxx_Thread("Receiver Thread") { }

public: void run(void)
  {
    int message = id;
    if (id % 3) yield();
    mail.Receive(&message);
    sumMutex.lock();
    tally -= message;
    sumMutex.unlock();
  }

private:
  int id;
};




int main(int argc, char* argv[])
{
  int i, total = 0;
  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  int nThreads = (argc > 1) ? atoi(argv[1]) : 3;
  if (nThreads < 2) nThreads = 2;
  cout << endl;

  //
  // create threads
  //
  Sender** sender = new Sender*[nThreads];
  Receiver** receiver = new Receiver*[nThreads];
  for (i = 0; i < nThreads; i++) {
    sender[i] = new Sender(i);
    receiver[i] = new Receiver(i);
    total += i;
  }
  
  //
  // start threads
  // 
  sender[nThreads-1]->start();
  receiver[nThreads-1]->start();
  for (i = 0; i < nThreads-1; i++) {
    receiver[i]->start();
    sender[i]->start();
  }

  //
  // wait for threads to finish
  //
  for (i = 0; i < nThreads; i++) {
    sender[i]->join();
    receiver[i]->join();
  }

  cout << endl;
  if (error != 0 && tally != 0 && sum != total) {
    cout << "ERROR, TestCond1 did not complete correctly" << endl;
    if (error == 0) error = 1;
    return error;
  } else {
    cout << "Successfully completed TestCond1" << endl;
  }

  //
  // clean up
  //
  for (i = 0; i < nThreads; i++) {
    delete sender[i];
    delete receiver[i];
  }
  delete[] sender;
  delete[] receiver;

  return hpcxx_exit(g);
}
