//-----------------------------------------------------------------------------
// GlobalPtr.C
// -----------
// This file contains an example of how to use global pointers to share data
// between contexts.
//-----------------------------------------------------------------------------
// To run the program use the following command line arguments:
//      main -nx -n 1.
// Also, it is possible to print out the values in each array by using the 
// -DDEBUG compiler switch.


#include <math.h>
#include <assert.h>
#include "Complex.C"
#include <iostream.h>
#include <hpcxx_rts.h>
#include <HPCxx_Sync.cc>
#include <hpcxx_invoke.cc>
#include <hpcxx_register.cc>
#include <HPCxx_GlobalPtr.cc>
#include <hpcxx_invokeImpl.cc>
#include <HPCxx_GlobalRefImpl.cc>


// declare function prototypes;

static void printArray(int array[], int size, int context);
static void printArray(float array[], int size, int context);
static void printArray(ComplexArray& array, int size, int context);
//static void sendGlobalPointer(HPCxx_GlobalPtr<int> gp, int context);
//static HPCxx_GlobalPtr<int> recvGlobalPointer(int context);

// remote function identifiers

//static int bind_ID;
static int resolve_ID;
static int resolve1_ID;
static int resolve2_ID;

class NameServiceLite {
public:

  //---------------------------------------------------------------------------
  // Registers the remote functions this object will serve.
  //---------------------------------------------------------------------------
  static void registerMemberFunctions(void)
  {
//    bind_ID    = hpcxx_register(NameServiceLite::bind, 10);
    resolve_ID = hpcxx_register(NameServiceLite::resolve, 11);
    resolve1_ID = hpcxx_register(NameServiceLite::resolve1, 12);
    resolve2_ID = hpcxx_register(NameServiceLite::resolve2, 13);
  }

  //---------------------------------------------------------------------------
  // Binds the given name with the object.  Returns 0.
  //---------------------------------------------------------------------------
  static int bind(const char* name, HPCxx_GlobalPtr<int> gp)
  {
    name_ = (char*) realloc(name_, strlen(name) + 1);
    strcpy(name_, name);
    gp_int = gp;			// copy the global pointer
    return 0;
  }

  //---------------------------------------------------------------------------
  // Binds the given name with the object.  Returns 0.
  //---------------------------------------------------------------------------
  static int bind(const char* name, HPCxx_GlobalPtr<float> gp)
  {
    name_1 = (char*) realloc(name_1, strlen(name) + 1);
    strcpy(name_1, name);
    gp_flt = gp;			// copy the global pointer
    return 0;
  }

  //---------------------------------------------------------------------------
  // Binds the given name with the object.  Returns 0.
  //---------------------------------------------------------------------------
  static int bind(const char* name, HPCxx_GlobalPtr<ComplexArray> gp)
  {
    name_2 = (char*) realloc(name_2, strlen(name) + 1);
    strcpy(name_2, name);
    gp_cpx = gp;			// copy the global pointer
    return 0;
  }
  
  //---------------------------------------------------------------------------
  // Returns the object bound to the given name.
  //---------------------------------------------------------------------------
  static HPCxx_GlobalPtr<int> resolve(const char* name)
  {
    assert( strcmp(name, name_) == 0 );	// just make sure names are the same
    return gp_int;
  }
  
  //---------------------------------------------------------------------------
  // Returns the object bound to the given name.
  //---------------------------------------------------------------------------
  static HPCxx_GlobalPtr<float> resolve1(const char* name)
  {
    assert( strcmp(name, name_1) == 0 ); // just make sure names are the same
    return gp_flt;
  }
  
  //---------------------------------------------------------------------------
  // Returns the object bound to the given name.
  //---------------------------------------------------------------------------
  static HPCxx_GlobalPtr<ComplexArray> resolve2(const char* name)
  {
    assert( strcmp(name, name_2) == 0 ); // just make sure names are the same
    return gp_cpx;
  }
  
private:

  static HPCxx_GlobalPtr<int> gp_int;	// copy of global pointer
  static HPCxx_GlobalPtr<float> gp_flt;	// copy of global pointer
  static HPCxx_GlobalPtr<ComplexArray> gp_cpx;	// copy of global pointer
  static char* name_;			// name bound to global pointer
  static char* name_1;			// name bound to global pointer
  static char* name_2;			// name bound to global pointer

};

HPCxx_GlobalPtr<int> NameServiceLite::gp_int;
HPCxx_GlobalPtr<float> NameServiceLite::gp_flt;
HPCxx_GlobalPtr<ComplexArray> NameServiceLite::gp_cpx;
char* NameServiceLite::name_ = NULL;
char* NameServiceLite::name_1 = NULL;
char* NameServiceLite::name_2 = NULL;


int main(int argc, char* argv[])
{
  int int_array[10];
  float flt_array[10];
  ComplexArray cpx_array(10);


  int error = 0;
  const char* intGP_name = "integer array of 10";
  const char* fltGP_name = "float array of 10";
  const char* cpxGP_name = "complex array of 10";

  //
  // initialize HPC++
  //
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);

  //
  // all processes are executing in SPMD mode at this point
  //

  int nc = group->getNumContexts() - 1;
  int context = group->getLocalIndex();

  //
  // Initially gp is local to all contexts.
  //
  HPCxx_GlobalPtr<int> int_gp(int_array);
  HPCxx_GlobalPtr<float> flt_gp(flt_array);
  HPCxx_GlobalPtr<ComplexArray> cpx_gp(&cpx_array);

  //
  // register naming service functions so they can be exported
  //
  hpcxx_registerClass(&cpx_array);
  NameServiceLite::registerMemberFunctions();
  hpcxx_barrier();

  //
  // Initialize global pointer and register it with the naming service
  // (running on context 0).
  //
  if (context == 0) {

    for (int i = 0; i < 10; i++) {
      Complex tmp(((float) i)/5, (9.0-i)/5);
      int_array[i] = i;
      flt_array[i] = ((float) i) / 5;
      cpx_array[i] = tmp;
    }

    int buf[10];
    int_gp.read(buf, 10, 0);
    printf("buf[3] = %d, buf[7] = %d\n", buf[3], buf[7]);

#ifdef DEBUG    
    printArray(int_array, 10, 0);
    printArray(flt_array, 10, 0);
    printArray(cpx_array, 10, 0);
#endif // DEBUG

    // Register (bind) the global pointer with the name service so that
    // other contexts can later obtain a reference to it (resolve).
    NameServiceLite::bind(intGP_name, int_gp);
    NameServiceLite::bind(fltGP_name, flt_gp);
    NameServiceLite::bind(cpxGP_name, cpx_gp);

  }

  hpcxx_barrier();

  //
  // The other context needs to retrieve the global pointer and work
  // with its contents.
  //

  if (context == 1) {

    HPCxx_ContextID nameServerContext = *group->getContextID(0);
    hpcxx_invoke(nameServerContext, int_gp, resolve_ID, intGP_name);
    hpcxx_invoke(nameServerContext, flt_gp, resolve1_ID, fltGP_name);
    hpcxx_invoke(nameServerContext, cpx_gp, resolve2_ID, cpxGP_name);

    //
    // Now that we have the global pointer, read 10 integers from it.
    //

    int_gp.read(int_array, 10);
#ifdef DEBUG
    printArray(int_array, 10, 1);
#endif // DEBUG

    for (int k = 0; k < 10; k++) {
      flt_array[k] = flt_gp[k];
    }
#ifdef DEBUG
    printArray(flt_array, 10, 1);
#endif // DEBUG

    cpx_gp.read(&cpx_array, 1);
#ifdef DEBUG
    printArray(cpx_array, 10, 1);
#endif // DEBUG
    
    //
    // modify the array
    //
    for (int i = 0; i < 10; i++) {
      Complex tmp(2.0, 2.0);
      int_array[i] = 9 - i;
      flt_array[i] = 2.0 - flt_array[i];
      cpx_array[i] = tmp - cpx_array[i];
    }

    //
    // write the modified array back to its owner
    //
    for (int j = 0; j < 10; j++)
      int_gp[j] = int_array[j];

    flt_gp.write(flt_array, 10);
    cpx_gp.write(&cpx_array);
  }

  // wait for array to be modified
  hpcxx_barrier();

  if (context == 0) {

#ifdef DEBUG
    printArray(int_array, 10, 0);
    printArray(flt_array, 10, 0);
    printArray(cpx_array, 10, 0);
#endif // DEBUG    
    
    // check to see if it has been correctly modified
    for (int i = 0; i < 10; i++) {
      Complex tmp(2 - ((float)i)/5, 2 - (9.0-i)/5);

      if (int_array[i] != 9 - i) error = 1;
      if (fabs(flt_array[i] - .200 * (i + 1)) > .01); else error = 1;
      if (cpx_array[i] == tmp); else error = 1;
    }
  }

  //
  // exit HPC++
  //

  hpcxx_barrier();
  hpcxx_exit();

#ifdef DEBUG
  if (context == 0) {
    cout << endl;
    if (error != 0) {
      cout << "ERROR, GlobalPointer example did not complete correctly" <<endl;
    } else {
      cout << "SUCCESSfully completed GlobalPointer example " << endl;
    }
  }
#endif // DEBUG

  return error;
}


//-----------------------------------------------------------------------------
// Prints the contents of an integer array to a given context.
//-----------------------------------------------------------------------------
static void printArray(int array[], int size, int context)
{
    cout << endl << "  context[" << context << "]:" << endl;
    for (int i = 0; i < size; i++) {
      cout << "\t\tarray[" << i << "] = " << array[i] << flush << endl;
    }
}

//-----------------------------------------------------------------------------
// Prints the contents of a float array to a given context.
//-----------------------------------------------------------------------------
static void printArray(float array[], int size, int context)
{
    cout << endl << "  context[" << context << "]:" << endl;
    for (int i = 0; i < size; i++) {
      cout << "\t\tarray[" << i << "] = " << array[i] << flush << endl;
    }
}


//-----------------------------------------------------------------------------
// Prints the contents of a float array to a given context.
//-----------------------------------------------------------------------------
static void printArray(ComplexArray& array, int size, int context)
{
    cout << endl << "  context[" << context << "]:" << endl;
    for (int i = 0; i < size; i++) {
      cout << "\t\tarray[" << i << "] = " << array[i] << flush << endl;
    }
}

/*
template<> size_t hpcxx_sizeof(const char* sz)
{
  return ( sizeof(int) + strlen(sz) );
}
*/

void hpcxx_pack(HPCxx_Buffer& buffer, const char** strArray, int count)
{
  for (int i = 0; i < count; i++) {
    int length = strlen(strArray[i]);
    buffer.pack(&length, 1);
    buffer.pack((char*) strArray[i], length);
  }
}

void hpcxx_unpack(HPCxx_Buffer& buffer, const char** strArray, int count)
{
  for (int i = 0; i < count; i++) {
    int length;
    buffer.unpack(&length, 1);
    //
    // strArray is not really const. It must be declared that way because
    // of the symetry required in the way the pack and unpack functions are
    // used.
    //
    char** str = (char**) strArray;
    str[i] = new char[length + 1];
    str[i][length] = '\0';
    buffer.unpack(str[i], length);
  }
}


// explicit instantiations of required templated classes and functions


template void hpcxx_pack  (HPCxx_Buffer &, HPCxx_GlobalPtr<int> *, int);
template void hpcxx_unpack(HPCxx_Buffer &, HPCxx_GlobalPtr<int> *, int);
template void hpcxx_pack  (HPCxx_Buffer &, HPCxx_GlobalPtr<float> *, int);
template void hpcxx_unpack(HPCxx_Buffer &, HPCxx_GlobalPtr<float> *, int);
template void hpcxx_pack  (HPCxx_Buffer &, HPCxx_GlobalPtr<ComplexArray> *, int);
template void hpcxx_unpack(HPCxx_Buffer &, HPCxx_GlobalPtr<ComplexArray> *, int);


// first instantiate the invoke function

