//-----------------------------------------------------------------------------
// HPCxx_Prototype.C -- definitions for a category of classes to support
// function prototypes and to provide a mapping between HPC++ function ids
// and function prototypes.
//
// This file implements two classes:
//     HPCxx_Prototype
//     HPCxx_IdProtoMap
//-----------------------------------------------------------------------------

#include "HPCxx_Prototype.h"
#include <assert.h>
#include <stdio.h>


HPCxx_Prototype::HPCxx_Prototype()
               : d_prototype(NULL), d_name(NULL), d_numArgs(-1)
{
}


HPCxx_Prototype::HPCxx_Prototype(const char *prototype)
{
   d_prototype = new char[strlen(prototype) + 1];
   strcpy(d_prototype, prototype);
   d_name = NULL;
   d_numArgs = -1;
}


HPCxx_Prototype::HPCxx_Prototype(const HPCxx_Prototype& s)
{
   if (s.d_prototype == NULL) {
      d_prototype = NULL;
      d_numArgs = -1;
   } else {
      d_prototype = new char[strlen(s.d_prototype) + 1];
      strcpy(d_prototype, s.d_prototype);
      d_numArgs = s.d_numArgs;;
   }
   d_name = NULL;
}


HPCxx_Prototype::~HPCxx_Prototype()
{
   if (d_prototype != NULL) delete [] d_prototype;
   if (d_name != NULL) delete [] d_name;
}


HPCxx_Prototype& HPCxx_Prototype::operator=(const HPCxx_Prototype& s)
{
  if (this != &s) {
    if (d_prototype != NULL) delete[] d_prototype;

    if (s.d_prototype == NULL) {
      d_prototype = NULL;
      d_numArgs = -1;
    } else {
      d_prototype = new char[strlen(s.d_prototype) + 1];
      strcpy(d_prototype, s.d_prototype);
      d_numArgs = s.d_numArgs;;
    }
  }  
  return *this;
}


// Return the number of arguments.
int HPCxx_Prototype::numberOfArgs(void)
{
   if (d_numArgs > -1) {
      return(d_numArgs);
   } else {
      // Assumes always at least one arg, which can be void.
      char *pos = new char[strlen(d_prototype) + 1];
      strcpy(pos, d_prototype);
      char *pos2 = strtok(pos, ",");
   //    cout << "Pos: " << pos2 << endl;
//       if (strcmp(pos2, d_prototype) != 0) {
	 d_numArgs = 1;
	 while (strtok(NULL, ","))
	    d_numArgs++;
//       } else {
// 	 d_numArgs = 0;
//       }
	 delete [] pos;
   }

   return(d_numArgs);
}


// Return the hpcxx_type_t corresponding to the return type of the function.
hpcxx_type_t HPCxx_Prototype::returnType(void)
{
  return(getArgType(d_prototype));   
}

const char* HPCxx_Prototype::name(void)
{
  if (d_name == NULL) {                
    char* begin = strchr(d_prototype, ' ' );
    while (begin != NULL
           && (*begin == ' ' || *begin == '*' || *begin == '&')) begin += 1;

    char* end = strchr(begin, '(');
    if (end == NULL) return "badly_formed_function_prototype";

    int len = end - begin;
    d_name = new char[len + 1];
    strncpy(d_name, begin, len);
    d_name[len] = '\0';
  }
  return d_name;
}


// Return the hpcxx_type_t corresponding to arg number i.
hpcxx_type_t HPCxx_Prototype::argTypeN(int i)
{
   char *cursor = new char[strlen(d_prototype) + 1];
   char *start = cursor;
   strcpy(cursor, d_prototype);
//    char *cursor = d_prototype;
   
   // Skip to the start of the argument list.
   while (*cursor != '(')
      cursor++;
   cursor++;
   
   // Skip to the right arg.
   for (int j = 1; j < i && *cursor != '\0'; j++) {
      while (*cursor != ',' && *cursor != '\0') 
	 cursor++;
      cursor++;
   }
   char *startCursor = cursor;
   while (*cursor != ',' && *cursor != ')') {
      cursor++;
   }
   *cursor = '\0';

   hpcxx_type_t retValue = getArgType(startCursor);
      
   delete [] start;
   return(retValue);
}


// Returns the hpcxx_type_t corresponding to the string.
hpcxx_type_t HPCxx_Prototype::getArgType(char *argType)
{
   hpcxx_type_t type;

   // Skip over any whitespace.
   while (*argType == ' ' && *argType != '\0')
      argType++;
   
   switch (argType[0]) {
    case 'v':
      if (strncmp("void", argType, 4) == 0) type = HPCXX_TYPE_VOID;
      break;
    case 'c':
      if (strncmp("char", argType, 4) == 0) type = HPCXX_TYPE_CHAR;
      break;
    case 'i':
      if (strncmp("int", argType, 3) == 0) type = HPCXX_TYPE_INT;
      break;
    case 'l':
      if (strncmp("long", argType, 4) == 0) type = HPCXX_TYPE_LONG;
      break;
    case 'f':
      if (strncmp("float", argType, 5) == 0) type = HPCXX_TYPE_FLOAT;
      break;
    case 'd':
      if (strncmp("double", argType, 6) == 0) type = HPCXX_TYPE_DOUBLE;
      break;
    case 'G':
      if (strncmp("GP", argType, 2) == 0) type = HPCXX_TYPE_GP;
      break;
    case 's':
      if (strncmp("short", argType, 5) == 0) type = HPCXX_TYPE_SHORT;
      if (strncmp("string", argType, 6) == 0) type = HPCXX_TYPE_STRING;
      break;
    default:
      /* assume a global pointer is used to pass unknown type */
      type = HPCXX_TYPE_GP;
      break;
  }

  return(type);
}


// Print the Prototype.
ostream & operator<<(ostream& o, const HPCxx_Prototype& s)
{
   return(o << s.d_prototype);
}


//-----------------------------------------------------------------------------
// HPCxx_IdProtoMap definitions
//-----------------------------------------------------------------------------


HPCxx_IdProtoMap::HPCxx_IdProtoMap() : freeID(1)
{
}


int HPCxx_IdProtoMap::addPrototype(const HPCxx_Prototype& s, int id)
{
   d_idProtoMap[freeID] = s;
   d_protoIdMap[s] = freeID;
   freeID += 1;
   return(freeID - 1);
}


//---- createRegisterCode -----------------------------------------------------
// Creates code to register functions with HPC++ and create an
// IdProtoMap, given prototypes in a file.
//
// prototypeFile -- path to a file containing prototypes, one per line
// codeFile -- code file to be produced (eg, register.C)
//-----------------------------------------------------------------------------
void HPCxx_IdProtoMap::createRegisterCode( const char* prototypeFile,
					      const char* codeFile )
{
  int i;
  char strBuf[512];
  HPCxx_IdProtoMap pMap;

  FILE* ifp = fopen(prototypeFile, "r");
  FILE* ofp = fopen(codeFile, "w");

  assert(ifp != NULL && ofp != NULL);

  //
  // write out header portion of file
  //
  fprintf(ofp, "//\n// This file is automatically generated.\n//\n\n");
  fprintf(ofp, "#include <HPCxx_Prototype.h>\n");
  fprintf(ofp, "#include <hpcxx_register.h>\n\n");

  //
  // store prototypes in protoMap and write out function prototypes
  //
  int index = 1;
  while (fgets(strBuf, 512, ifp) != NULL) {
    // strip off ';' and/or '\n'
    char* eol = strchr(strBuf, ';');
    if (eol != NULL) {
      *eol = '\0';
    } else {
      eol = strchr(strBuf, '\n');
      if (eol != NULL) *eol = '\0';
    }

    // create prototype and copy to map
    int id = pMap.addPrototype( HPCxx_Prototype(strBuf) );
    assert (id == index);
    
    // write out prototypes
    fprintf(ofp, "%s;\n", strBuf);
    index += 1;
  }

  //
  // write out function body
  //
  fprintf(ofp, "\n\n");
  fprintf(ofp, "void hpcxx_register_functions_(HPCxx_IdProtoMap** map)\n");

  fprintf(ofp, "{\n  HPCxx_IdProtoMap& pMap = hpcxx_prototypeMap;\n\n");

  //
  // add function prototype to the id-prototype map
  //
  for (i = 1; i < index; i++) {
    fprintf(ofp, "  pMap.addPrototype(HPCxx_Prototype(\"%s\"));\n",
	    pMap[i].prototype());
  }
  fprintf(ofp, "\n");

  // write out hpcxx_register... for each prototype
  for (i = 1; i < index; i++) {
    fprintf(ofp, "  hpcxx_register(%s, %d);\n", pMap[i].name(), i);
  }

  // write out end of function
  fprintf(ofp, "\n  *map = &pMap;\n}\n");

  fclose(ifp);
  fclose(ofp);
}


